/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Doe Hyun Yoon
 */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <cstdlib>
#include <algorithm>
#include <vector>
#include <sstream>
#include "common.h"
#include "system.h"
#include "stat.h"
#include "anyoption.h"
#include "drsim_trace.h"

using namespace std;
using namespace drsim;

int main(int argc, char *argv[], char *argp[]) {
  cout << "DrSim trace-driven frontend." << endl;
  print_banner(cout) << endl;
  print_svn_rev(cout) << endl;

  TraceSimArgs args;
  args.parse_input_args(argc, argv);

  // print out parameters before starting simulation
  args.print_tracesim_args(cout);
  DRAMTiming::print_dram_timing(cout);

  // build a system with parameters
  System system(args);
  fstream trace(args.trace_file_name(), ios::in);

  // main loop begins
  TraceReq req;
  req.scheduled_ = true;
  bool trace_left = true;
  bool reqs_pending = false;
  cout << endl << "Simluation starts." << endl;
  do {
    trace_left = read_a_request(req, trace); // this function checks scheduled_ internally

    // send to mem controllers -- if possible
    if (req.scheduled_ == false && (req.timeCurr_ <= clock::get_sys_clk())) {
      bool is_read = (req.reqType_ == REQ_READ ||
                      req.reqType_ == REQ_IFETCH);
      req.scheduled_ = system.sendRequest(is_read,
                                          req.addrBase_,
                                          req.size_,
                                          req.priority_);
    }

    system.tick(); // advance the system clock

    // periodic progress report
#define PROGRESS_REPORT_PERIOD 1000000
    if (clock::get_sys_clk() % PROGRESS_REPORT_PERIOD == 0) {
      cout << "[";
      cout << setfill('0') << setw(8) << clock::get_sys_clk();
      cout << "]";
      cout <<  system  <<  endl;
    }

    reqs_pending = system.isReqPending();
  } while ((trace_left || reqs_pending) &&
           (args.sim_cycles() == 0 || (args.sim_cycles() > clock::get_sys_clk())));

  // report simulation results
  system.wrapUpStats();
  cout << "Simulation finished at master_clk: " <<  clock::get_master_clk()
       << ", sys_clk: " << clock::get_sys_clk()
       << ", dram_clk: " << clock::get_dram_clk() << endl;
  cout << endl << "[Stats]" << endl;
  Stat::printall(cout);
  cout << endl << "[Channel Stats]" << endl;
  MemoryChannel::printChannelStats(cout);

  trace.close();
}

namespace drsim {

#define SET_OPT(longname, shortname, usage)\
  opt->addUsage( (usage) );\
  opt->setOption( (longname), (shortname) )
#define SET_OPTL(longname, usage)\
  opt->addUsage( (usage) );\
  opt->setOption( (longname) )
#define SET_FLAG(longname, shortname, usage)\
  opt->addUsage( (usage) );\
  opt->setFlag( (longname), (shortname))
#define SET_FLAGS(shortname, usage)\
  opt->addUsage( (usage) );\
  opt->setFlag( (shortname) )

#define GET_FLAG( name )\
  opt->getFlag( (name) )
#define GET_STR( name )\
  opt->getValue( (name) )
#define GET_NUM( name )\
  atoi( opt->getValue( (name) ) )

#define NUM_OPTION( target, name )\
  if( GET_STR( (name) ) != NULL ) { \
    (target) = GET_NUM( (name) );\
  }
#define NUM_OPTIONF( function, name )\
  if( GET_STR( (name) ) != NULL ) { \
    (function)( GET_NUM( (name) ) );\
  }

void TraceSimArgs::parse_input_args(int argc, char *argv[]) {
  // the last argument is mandatory: input trace file name
  char *szTraceFileName = argv[argc-1];
  fstream fpTrace(szTraceFileName, ios::in);
  if (fpTrace.fail()) {
    cout << "Can't open the trace file:" << szTraceFileName << endl;
    exit(-1);
  }
  argc--;

  // command-line parser
  AnyOption *opt = new AnyOption(100, 100); // 100 arguments -- probably enough
  opt->setVerbose();
  opt->autoUsagePrint( true );

  // register usage banner
  ostringstream oss(ostringstream::out);
  oss << "Usage: " << argv[0] << " [options] trace_file" << endl;;
  opt->addUsage(oss.str().c_str());

  // register TraceSim-only options
  opt->addUsage( "DrSim Trace Frontend Options:" );
  SET_OPT(    "sim_cycle",'s',    "     -s  --sim_cycle    <n>                Max Simulation Cycles (default:0 -- no limit)");
  SET_OPTL(   "sys_freq",         "         --sys_freq     <n>                System Clock Frequency (default:3000, unit:MHz, min granularity:33.33MHz)" );

  Args::parse_input_args(opt, argc, argv);

  // print usage
  if (argc == 0) {
    opt->printUsage();
    delete opt;
    exit(-1);
  }

  // retrieve TraceSim-only options
  NUM_OPTIONF(set_sim_cycles, "sim_cycle");
  NUM_OPTIONF(set_sim_cycles, 's');
  set_trace_file_name(szTraceFileName);
  NUM_OPTIONF(set_sys_freq,           "sys_freq");

  delete opt;
}

/*
 * trace input file handling
 */
//------------------------------------------------------------------------------------
static char xtod(char c) {
  if (c>='0' && c<='9') return c-'0';
  if (c>='A' && c<='F') return c-'A'+10;
  if (c>='a' && c<='f') return c-'a'+10;
  return c=0;        // not Hex digit
}
static UINT64 HextoDec(char *hex, UINT64 l) {
  if (*hex==0) return(l);
  return HextoDec(hex+1, l*16+xtod(*hex)); // hex+1?
}
static UINT64 xstrtoi(char *hex) {    // hex string to integer
  return HextoDec(hex,0);
}

const char *g_szReqTypes[NUM_REQ_TYPE] = {
  "READ", "WRITE", "IFETCH",
};

void tokenizer(const string& str, vector<string>& tokens, const string& delimiters = " ") {
  // Skip delimiters at beginning.
  string::size_type lastPos = str.find_first_not_of(delimiters, 0);
  // Find first "non-delimiter".
  string::size_type pos     = str.find_first_of(delimiters, lastPos);

  while (string::npos != pos || string::npos != lastPos) {
    // Found a token, add it to the vector.
    tokens.push_back(str.substr(lastPos, pos - lastPos));
    // Skip delimiters.  Note the "not_of"
    lastPos = str.find_first_not_of(delimiters, pos);
    // Find next "non-delimiter"
    pos = str.find_first_of(delimiters, lastPos);
  }
}

/*
 * function: read_a_request
 * return false when there's no request any more
 */
bool read_a_request( TraceReq &rcCurr_req, fstream &rfsTrace ) {
  char szDefaultSize[] = "64";
  if ( rcCurr_req.scheduled_ ) {
    bool bComment = false;
    do {
      if ( rfsTrace.eof() ) {
        return false;
      }
      bComment = false;
      vector <string> tokens;
      char szBuf[1024];
      rfsTrace.getline( szBuf, 1024 );

      string strLine( szBuf );
      tokenizer( strLine, tokens );
      int len = tokens.size();

      if ( szBuf[0] == '#' ) {
        bComment = true;
      }
      if ( len > 0 ) {
        char *temp = (char *) tokens[0].c_str();
        if ( temp[0] == '#' ) {
          bComment = true;
        }
      }

      if ( bComment == false ) {
        if ( len < 3 ) {
          return false;
        }
        char *addr = (char *) tokens[0].c_str();
        char *op = (char *) tokens[1].c_str();
        char *time = (char *) tokens[2].c_str();
        char *req_size;
        if ( len >= 4 && (tokens[3].c_str())[0] != '#') {
          req_size = (char *) tokens[3].c_str();
        } else {
          req_size = szDefaultSize; // 64
        }
        int priority = 2;
        if ( len >= 5 && (tokens[4].c_str())[0] != '#') {
          priority = atoi((char *)tokens[4].c_str());
        }

//cout<<szBuf<<endl;
//cout<<addr<<"----"<<op<<"----"<<time<<"----"<<req_size<<endl;

        /*
         * trace file format
         * one request per line
         * a single request is
         * Address   type    time   size
         * Address is a hexadecimal number starts with "0x"
         * type is one of READ, WRITE, and IFETCH
         * time is a decimal number
         * size is in bytes
         * example)
         * 0x20050280 IFETCH  30  16
         */
        rcCurr_req.addrBase_ = xstrtoi(addr);
        string strReqType(op);
        if ( strReqType.compare( g_szReqTypes[REQ_READ] ) == 0 ) {
          rcCurr_req.reqType_ = REQ_READ;
        } else if ( strReqType.compare( g_szReqTypes[REQ_WRITE] ) == 0 ) {
          rcCurr_req.reqType_ = REQ_WRITE;
        } else if ( strReqType.compare( g_szReqTypes[REQ_IFETCH] ) == 0 ) {
          rcCurr_req.reqType_ = REQ_IFETCH;
        } else {
          // Can't find a valid request anymore!!
          return false; // error? or end of file?
        }
        rcCurr_req.size_ = atoi(req_size); //
        rcCurr_req.timeCurr_ = atoi(time); // is atoi() ok?
        rcCurr_req.scheduled_ = false; // make sure this new req is not scheduled yet
        rcCurr_req.priority_ = priority;

        return true;
      }
    } while ( bComment );
  } else {
    return true; // the current event is not served yet
  }
  return true; // just not to cause a warning in compilation
}

} // namespace drsim
