/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Doe Hyun Yoon
 *          Min Kyu Jeong
 */

#include <cstdlib>
#include "system.h"

using std::cout;
using std::endl;

namespace drsim {

System::System() : num_request_sent_(0) {}

System::System(TraceSimArgs args) : args_(args), num_request_sent_(0) {
  // Translating old parameters 'num_subrank' and 'subrank_map' to
  // subrank_interleaving_granularity and channel_interleaving_granularity.
  if (args.num_subrank() != 0 && args.subrank_map() != INVALID_SUBRANK_MAP) {
    args.set_num_chip(8);
    args.set_subrank_width(8/args.num_subrank());
    switch (args.subrank_map()) {
      case M2:
        args.set_subrank_interleaving_granularity(64);
        args.set_channel_interleaving_granularity(64*args.num_subrank());
        break;

      case M4:
        args.set_subrank_interleaving_granularity(args.module_config().num_column*8/args.num_subrank());
        args.set_channel_interleaving_granularity(64*args.num_subrank());
        break;

      default:
        PARAMCHECK(false, "subrank_map() is neither M2 or M4: " << args.subrank_map());
    }
  }

  decoder_ = new AddrDecoder
      (args.num_channel(), //UINT32 num_channel,
       args.module_config().num_rank, //UINT32 num_rank,
       args.module_config().num_chip, //UINT32 num_chip,
       args.module_config().num_bank, //UINT32 num_bank,
       args.module_config().num_row, //65536, //UINT32 num_row,
       args.module_config().num_column, //1024, //UINT32 num_column,
       args.module_config().chip_width, //1, //UINT32 chip_width,
       args.module_config().burst_length, //8, //UINT32 burst_length,
       args.channel_interleaving_granularity(), //sig*(8/subrank_width[sr_width]), // UINT32 channel_interleaving_granularity
       args.bank_interleave(), //true, // bool rank_interleaving
       args.rank_interleave(), //true, // bool rank_interleaving
       args.subrank_width(),// subrank_width[sr_width], //UINT32 subrank_width,
       args.subrank_interleaving_granularity(),// sig, //UINT32 subrank_interleaving_granularity
       args.bank_xor_hash(),//true,   //bool bank_xor_hashing,    
       0,
       args.subrank_xor_hash(),//true,   //bool subrank_xor_hashing,
       args.channel_xor_hash());//true);  //bool channel_xor_hashing)

  cout << endl << "[Address Decoding Map]" << endl;
  decoder_->PrintMask(cout);

  mc_ = new MemoryChannel*[args.num_channel()];
  for (UINT32 i=0; i < args.num_channel(); i++) {
    mc_[i] = new MemoryChannel(args,
                              args.module_config(),
                              args.read_queue_size(),
                              args.write_queue_size(),
                              i);
  }

  // initialize clock -- DRAM clock and Sys clock all in MHz
  clock::xInitialize(args.dram_freq(), args.sys_freq());
}

System::~System(void) {
  for (UINT32 i=0; i < args_.num_channel(); i++)
    delete mc_[i];
  delete [] mc_;
  delete decoder_;
}

bool System::sendRequest(bool is_read, ADDRESS phy_addr, UINT32 size, int priority) {
  UINT32 channel = decoder_->GetChannel(phy_addr);

  if ((is_read && !mc_[channel]->isReadQueueFull())
      || (!is_read && !mc_[channel]->isWriteQueueFull())) {
    DRAMAddr dram_addr;
    decoder_->Decode(phy_addr, size, &dram_addr);
    Req* req = new Req(num_request_sent_++,
                       is_read,
                       phy_addr,
                       size,
                       dram_addr);

    for (int i = 2; i < priority; i++)
      req->RaisePriority();

    mc_[channel]->sendRequest(req);
    return true;
  }
  
  return false;
}

void System::tick(void) {
  // a system tick -- tick master clock until sys clock ticks
  do {
    clock::xTick();

    if (clock::is_dram_ticked()) {
      for (UINT32 i=0; i < args_.num_channel(); i++)
        mc_[i]->cycle();
    }
  } while (!clock::is_sys_ticked());
}

bool System::isReqPending() {
  bool isReqPending = false;

  for (UINT32 i=0; i < args_.num_channel(); i++)
    isReqPending = (isReqPending || mc_[i]->isReqPending());

  return isReqPending;
}

ostream & operator<<(ostream &os, const System &system) {
#if 0
  os<<endl;
  if (system.decoder_ != NULL)
    system.decoder_->printMask(os);
  for (UINT32 i=0; i < system.args_.num_channel(); i++) {
    system.mc_[i]->printStatus(os);
  }
  os<<endl;
#else
  UINT64 total_reqs_processed = 0;
  for (UINT32 i=0; i < system.args_.num_channel(); i++) {
    total_reqs_processed += system.mc_[i]->numReqProcessed();
  }
  os << " " << total_reqs_processed << " reqs are processed ...";
#endif

  return os;
}

void System::wrapUpStats() {
  for (UINT32 i = 0; i < args_.num_channel(); i++)
    mc_[i]->wrapUpStats();
}

}  // namespace drsim
