/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Doe Hyun Yoon
 */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>

using namespace std;

#include "common.h"
#include "dram.h"
#include "timing_model.h"
#include "clock.h"

/*
 * global variables, if you need
 */

// hard-coded configuration
const UINT32 num_ch = 1; // don't care, but should be 1
const UINT32 num_rk = 2;
const UINT32 num_bk = 8; // set to 8 for FAW testing
const UINT32 num_sr = 2; // support only 2 subranks here -- need more subranks here???

// The timing model to be tested
cTimingModel *theTM; // num_ch, num_rk, num_bk, num_sr );

/*
 * issue command until it gets granted
 */
void issue( eMemCmd cmd, UINT32 rk, UINT32 bk, UINT32 row, UINT32 col, bool s0, bool s1 ) {
  // construct a DRAMAddr class
  DRAMAddr daddr;
  daddr.channel = 0;
  daddr.rank = rk;
  daddr.bank = bk;
  daddr.row = row;
  daddr.column = col;
  daddr.subrank_mask[0] = s0;
  daddr.subrank_mask[1] = s1;
  daddr.num_cmd_repeat = 1;

  bool ok;
  do {
    // wait until it's ok to issue the command
    ok = theTM->checkCommand( cmd, daddr );
    if ( !ok ) {
      clock::xTick();
      if ( clock::is_dram_ticked() ) {
        theTM->xCycle();
      }
    }
  } while ( !ok );
  // then issue the command
  theTM->xIssueCommand( cmd, daddr );
  clock::xTick();
  if ( clock::is_dram_ticked() ) {
    theTM->xCycle();
  }
}

/*
 * main
 * note: let's make the main as simple as possible -- put every dirty stuffs in dsim_trace_aux.cpp
 */
#ifndef LOG_ENABLE // for release version build -- why would I need release version of TM test?
bool DBG_LEVEL[20] = {false};
#endif

int main(int argc, char *argv[], char *argp[]) {
  DBG_LEVEL[TRACE_TM] = true; // debug level set sup
  // then create timing model
  theTM = new cTimingModel( num_ch, num_rk, num_bk, num_sr, 2 , 800);
  DRAMAddr::num_subrank = num_sr; // it's for proper address print out

  // declare variables here
  //**********************************************************************
  // each test case starts with ACTs and PREs
  bool bRRD = false;
  bool bACT = false;
  bool bActRdPre = false;
  bool bRdBkRkSwitching = false;
  bool bActWrPre = false;
  bool bRdWrRdWr = false;
  bool bRandomRdWr_closepage = false;
  bool bFAW = true;

  // initialize
  //**********************************************************************
  clock::xInitialize( 3000, 3000); // same clock for sys and dram -- driving sys or master will directly drive dram

  // start DRAM command sequence
  //**********************************************************************
  //   	  	Cmd				rk		bk		row			col			sr0		sr1
  if ( bRRD ) {
    issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	0,		1, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	1,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	1,		1, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_PRE,	0,		0,		1234,		5678,		1,		1 );
    issue( 	Cmd_PRE,	0,		1,		1234,		5678,		1,		1 );
    issue( 	Cmd_PRE,	1,		0,		1234,		5678,		1,		1 );
    issue( 	Cmd_PRE,	1,		1,		1234,		5678,		1,		1 );
  }
  if ( bACT ) {
    issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_PRE,	0,		0,		1234,		5678,		1,		1 );
    issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_PRE,	0,		0,		1234,		5678,		1,		1 );
    issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_PRE,	0,		0,		1234,		5678,		1,		1 );
    issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_PRE,	0,		0,		1234,		5678,		1,		1 );
    issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_PRE,	0,		0,		1234,		5678,		1,		1 );
  }
  if ( bActRdPre ) {
    issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		0	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		0,		1	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		0	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		0,		1	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		0	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		0,		1	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		0	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		0,		1	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		0	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		0,		1	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		0	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		0,		1	);
    issue( 	Cmd_PRE,	0,		0,		1234,		5678,		1,		1 );
  }
  if ( bRdBkRkSwitching ) {
    issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	0,		1, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	1,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	1,		1, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
    issue( 	Cmd_RD,		0,		1,		1234,		5678,		1,		1	);
    issue( 	Cmd_RD,		1,		0,		1234,		5678,		1,		1	);
    issue( 	Cmd_RD,		1,		1,		1234,		5678,		1,		1	);
    issue( 	Cmd_PRE,	0,		0,		1234,		5678,		1,		1 );
    issue( 	Cmd_PRE,	0,		1,		1234,		5678,		1,		1 );
    issue( 	Cmd_PRE,	1,		0,		1234,		5678,		1,		1 );
    issue( 	Cmd_PRE,	1,		1,		1234,		5678,		1,		1 );
  }
  if ( bActWrPre ) {
    issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		0	);
    issue( 	Cmd_WR,		0,		0,		1234,		5678,		0,		1	);
    issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		0	);
    issue( 	Cmd_WR,		0,		0,		1234,		5678,		0,		1	);
    issue( 	Cmd_PRE,	0,		0,		1234,		5678,		1,		1 );
  }
  if ( bRdWrRdWr ) {
    issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
    issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
    issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
    issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
    issue( 	Cmd_PRE,	0,		0,		1234,		5678,		1,		1 );
  }
  if ( bRandomRdWr_closepage ) {
    if ( 0 ) {
      issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RDAP,	0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WRAP,	0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RD,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_RDAP,	0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WR,		0,		0,		1234,		5678,		1,		1	);
      issue( 	Cmd_WRAP,	0,		0,		1234,		5678,		1,		1	);
    } else {
      issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_ACT, 	1,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	1,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_ACT, 	0,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	0,		1, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_ACT, 	1,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	1,		1, 		1234, 	5678, 	0, 		1 );

      issue( 	Cmd_RDAP,	0,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_RDAP,	0,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_RDAP,	0,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_RDAP,	0,		1, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_RDAP,	1,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_RDAP,	1,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_RDAP,	1,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_RDAP,	1,		1, 		1234, 	5678, 	0, 		1 );

      issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_ACT, 	1,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	1,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_ACT, 	0,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	0,		1, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_ACT, 	1,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	1,		1, 		1234, 	5678, 	0, 		1 );

      issue( 	Cmd_WRAP,	0,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_WRAP,	0,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_WRAP,	0,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_WRAP,	0,		1, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_WRAP,	1,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_WRAP,	1,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_WRAP,	1,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_WRAP,	1,		1, 		1234, 	5678, 	0, 		1 );

      issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_ACT, 	1,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	1,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_ACT, 	0,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	0,		1, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_ACT, 	1,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_ACT, 	1,		1, 		1234, 	5678, 	0, 		1 );

      issue( 	Cmd_RDAP,	0,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_RDAP,	0,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_RDAP,	0,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_RDAP,	0,		1, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_RDAP,	1,		0, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_RDAP,	1,		0, 		1234, 	5678, 	0, 		1 );
      issue( 	Cmd_RDAP,	1,		1, 		1234, 	5678, 	1, 		0 );
      issue( 	Cmd_RDAP,	1,		1, 		1234, 	5678, 	0, 		1 );
    }
  }
  if ( bFAW ) {
    issue( 	Cmd_ACT, 	0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	0,		1, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	0,		2, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	0,		3, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	0,		4, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	0,		5, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	0,		6, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_ACT, 	0,		7, 		1234, 	5678, 	1, 		1 );

    issue( 	Cmd_RDAP, 0,		0, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_RDAP, 0,		1, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_RDAP, 0,		2, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_RDAP, 0,		3, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_RDAP, 0,		4, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_RDAP, 0,		5, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_RDAP, 0,		6, 		1234, 	5678, 	1, 		1 );
    issue( 	Cmd_RDAP, 0,		7, 		1234, 	5678, 	1, 		1 );
  }
  //**********************************************************************
}
// main ends here!


