#ifndef __MEM_DRSIM_HH__
#define __MEM_DRSIM_HH__

#include <string>
#include <vector>
#include "mem/physical.hh"
#include "params/DrSim.hh"
#include "debug/DrSim.hh"
#include "drsiminc/memorychannel.h"
#include "drsiminc/addrdecoder.h"
#include "drsiminc/args.h"
#include "sim/stats.hh"
#include "base/callback.hh"
#include "mem_drsim/events.hh"

using std::string;
using std::vector;
using drsim::ADDRESS;
using drsim::DRAMAddr;

class DrSim;

class Gem5Req : public drsim::Req
{
  public:
    Gem5Req(DrSim* drsim,
            uint64_t id,
            bool is_read,
            ADDRESS phy_addr,
            int size,
            DRAMAddr dram_addr,
            int page_owner_id,
            PacketPtr pkt,
            int port_id)
        : Req(id, is_read, phy_addr, size, dram_addr, false, page_owner_id),
              drsim_(drsim), pkt_(pkt), port_id_(port_id), etHandle_()
    {}

    void completed_callback();
    void critical_word_callback(int cw_size);
    virtual void PrintExtra(ostream &os) const;

    ET::TransHandle ETHandle() { return etHandle_; }
    void ETHandle(ET::TransHandle val) { etHandle_ = val; }

  protected:
    DrSim* drsim_;
    PacketPtr pkt_;
    int port_id_;

    ET::TransHandle etHandle_;
};

class DrSim : public PhysicalMemory
{
    friend class Gem5Req;

    class DrSimPort : public SimpleTimingPort
    {
        friend class DrSim;

      public:
        DrSimPort(string name, DrSim* d, int i);
        ~DrSimPort() {}

        void recvFunctional(PacketPtr pkt);
        Tick recvAtomic(PacketPtr pkt) { return drsim->doAtomicAccess(pkt); }
        bool recvTiming(PacketPtr pkt);
        void recvStatusChange(Status stat) { drsim->recvStatusChange(stat); }
        void getDeviceAddressRanges(AddrRangeList &resp, bool &snoop) {
            drsim->getAddressRanges(resp, snoop);
        }
      protected:
        int id;
        DrSim* drsim;
    };

    class ExitCallbackEvent : public Callback
    {
      public:
        ExitCallbackEvent(DrSim* d, string ofn) : drsim(d), output_fn(ofn) {}
        virtual void process();
      protected:
        DrSim* drsim;
        string output_fn;
    };

  public:
    typedef DrSimParams Params;
    DrSim(const Params* p);
    ~DrSim();

    void init();
    Port* getPort(const string &if_name, int idx);

    virtual void regStats();
    void wrapUpStats();
    void dumpStates();
    void printStatus();

  protected:
    void startup();
    void sendResponse(PacketPtr pkt, int port_id);

    drsim::MemoryChannel** mc;
    drsim::AddrDecoder** decoder;

    drsim::Args args;
    uint64_t numReqSent;
    Tick responseDelay;
    int32_t cpuPriorityMod;

    Stats::Scalar totalAccesses;
    Stats::Scalar readwriteAccesses;
    Stats::Scalar otherAccesses;

    void printDecoderMask(ostream &os);

    // zesto-MC defined interfaces from here
    void tick();
    bool enqueueable(PacketPtr pkt);
    void enqueue(PacketPtr pkt, int port_id);
    void printStat(ostream &os);
    void periodicStatPrint();

    // gem5 dramsim port inherited
    vector<DrSimPort*> ports;
    Tick ticks_per_dram_cycle;
    Tick stat_print_period;
    EventWrapper<DrSim, &DrSim::tick> tickEvent;
    EventWrapper<DrSim, &DrSim::periodicStatPrint> statEvent;
    ExitCallbackEvent exitEvent;

    // Event Tracer
    bool etEnabled;
    Tick etStartTick;
    Tick etEndTick;
    ET* globalTrace;
    EventTrace* events;
    void etStartStop();
    EventWrapper<DrSim, &DrSim::etStartStop> etStartStopEvent;
};

#endif // __MEM_DRSIM_HH__
