/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#ifndef __MEM_DRSIM_HH__
#define __MEM_DRSIM_HH__

#include <string>
#include <vector>
// Gem5 headers
#include "mem/physical.hh"
#include "params/DrSim.hh"
#include "debug/DrSim.hh"
#include "sim/stats.hh"
#include "base/callback.hh"
// DrSim headers
#include "drsiminc/memorychannel.h"
#include "drsiminc/addrdecoder.h"
#include "drsiminc/args.h"

using std::string;
using std::vector;
using drsim::ADDRESS;
using drsim::DRAMAddr;

class DrSim;

class Gem5Req : public drsim::Req
{
  public:
    Gem5Req(DrSim* drsim,
            uint64_t id,
            bool is_read,
            ADDRESS phy_addr,
            int size,
            DRAMAddr dram_addr,
            int page_owner_id,
            PacketPtr pkt,
            int port_id)
        : Req(id, is_read, phy_addr, size, dram_addr, false, page_owner_id),
              drsim_(drsim), pkt_(pkt), port_id_(port_id)
    {}

    void completed_callback();
    void critical_word_callback(int cw_size);
    virtual void PrintExtra(ostream &os) const;

  protected:
    DrSim* drsim_;
    PacketPtr pkt_;
    int port_id_;
};

class DrSim : public PhysicalMemory
{
    friend class Gem5Req;

    class DrSimPort : public MemoryPort
    {
      public:
        DrSimPort(string name, DrSim* d, int i);
        ~DrSimPort() {}
        bool recvTiming(PacketPtr pkt);

      protected:
        int id;
        DrSim* drsim;

        friend class DrSim;
    };

    class ExitCallbackEvent : public Callback
    {
      public:
        ExitCallbackEvent(DrSim* d, string ofn) : drsim(d), output_fn(ofn) {}
        virtual void process();

      protected:
        DrSim* drsim;
        string output_fn;
    };

  public:
    typedef DrSimParams Params;
    DrSim(const Params* p);
    ~DrSim();

    void init();
    Port* getPort(const string &if_name, int idx);

    virtual void regStats();
    void wrapUpStats();
    void dumpStates();
    void printStatus();

  protected:
    void startup();
    void sendResponse(PacketPtr pkt, int port_id);

    drsim::MemoryChannel** mc;
    drsim::AddrDecoder* decoder;

    drsim::Args args;
    uint64_t numReqSent;
    Tick responseDelay;
    int32_t cpuPriorityMod;
    bool use_ingress_q;

    /*
    Stats::Scalar totalAccesses;
    Stats::Scalar readwriteAccesses;
    Stats::Scalar otherAccesses;
    */

    // zesto-MC defined interfaces from here
    void tick();
    bool enqueueable(PacketPtr pkt);
    void enqueue(PacketPtr pkt, int port_id);
    void printStat(ostream &os);
    void periodicStatPrint();

    // gem5 dramsim port inherited
    vector<DrSimPort*> ports;
    Tick ticks_per_dram_cycle;
    Tick stat_print_period;
    EventWrapper<DrSim, &DrSim::tick> tickEvent;
    EventWrapper<DrSim, &DrSim::periodicStatPrint> statEvent;
    ExitCallbackEvent exitEvent;
};

#endif // __MEM_DRSIM_HH__
