/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#ifndef ADDRDECODER_H_
#define ADDRDECODER_H_

#include <string>
#include "dram.h"

using std::string;

namespace drsim {

// DEPRECATED
enum SubrankMap {
  M0 = 0,   // interleave every IW_S (interface width of a subrank)
  M1,       // interleave every BL*IW_S
  M2,       // interleave every 64B (whole cacheline comes from a subrank)
  M3,
  M4,       // row interleaving
  M5,       // reserved for future use
  M6,       // reserved for future use
  MCUSTOM,
  INVALID_SUBRANK_MAP
};
SubrankMap fromString(const string &str);

class AddrDecoder {
 public:
  AddrDecoder(UINT32 num_channel,
              UINT32 num_rank,
              UINT32 num_chip,
              UINT32 num_bank,
              UINT32 num_row,
              UINT32 num_column,
              UINT32 chip_width,
              UINT32 burst_length,
              UINT32 channel_interleaving_granularity,
              bool bank_interleaving,
              bool rank_interleaving,
              UINT32 subrank_width,
              UINT32 subrank_interleaving_granularity,
              bool bank_xor_hashing,
              UINT32 bank_xor_hashing_offset,
              bool subrank_xor_hashing,
              bool channel_xor_hashing);

  // DEPRECATED
//  AddrDecoder(UINT32 num_channel,
//              DRAMModule module_config,
//              bool subrank_xor_hashing,
//              bool channel_xor_hashing,
//              SubrankMap mapping,
//              UINT32 bit_offset_subrank = 0,
//              UINT32 bit_offst_channel = 0,
//              UINT32 rank_interleave = 0);

  UINT32 GetChannel(ADDRESS paddr);
  void Decode(ADDRESS phy_addr, UINT32 size, DRAMAddr* dram_addr);
  ostream& PrintMask(ostream &os);
  UINT32 GetCriticalWordSize() { return critical_word_size_; }
  UINT32 bank_offset() { return bank_offset_; }

 protected:
  void CalculateMask();
  ostream& PrintMask(ostream &os, ADDRESS mask, char c);

  ADDRESS SetBits(UINT32 num_one, UINT32 offset);
  ADDRESS ApplyMask(ADDRESS a, ADDRESS mask);

  UINT32 GetRank(ADDRESS paddr);
  UINT32 GetSubrank(ADDRESS paddr);
  UINT32 GetRow(ADDRESS paddr);
  UINT32 GetColumn(ADDRESS paddr);
  UINT32 GetBank(ADDRESS paddr);
  UINT32 GetNumOne(ADDRESS value);

  INT32 FirstNonZeroIndex(ADDRESS addr);

  // Memory channel parameters
  UINT32 num_channel_;
  UINT32 num_rank_;
  UINT32 num_chip_;
  UINT32 num_channel_log2_;
  UINT32 num_rank_log2_;
  UINT32 num_chip_log2_;

  // DRAM chip parameters
  UINT32 num_bank_;
  UINT32 num_row_;
  UINT32 num_column_;
  UINT32 chip_width_;
  UINT32 burst_length_;
  UINT32 num_bank_log2_;
  UINT32 num_row_log2_;
  UINT32 num_column_log2_;
  UINT32 chip_width_log2_;
  UINT32 burst_length_log2_;

  // Mapping policy parameters
  UINT32 channel_interleaving_granularity_;
  bool bank_interleaving_;
  bool rank_interleaving_;
  UINT32 subrank_width_;
  UINT32 subrank_interleaving_granularity_;
  bool bank_xor_hashing_;
  UINT32 bank_xor_hashing_offset_;
  bool subrank_xor_hashing_;
  bool channel_xor_hashing_;

  UINT32 subrank_width_log2_;
  UINT32 subrank_interleaving_granularity_log2_;
  UINT32 channel_interleaving_granularity_log2_;

  // Decoding masks
  ADDRESS mask_channel_;
  ADDRESS mask_channel_xor_;
  ADDRESS mask_rank_;
  ADDRESS mask_bank_;
  ADDRESS mask_bank_xor_;
  ADDRESS mask_row_;
  ADDRESS mask_column_;
  ADDRESS mask_subrank_;
  ADDRESS mask_subrank_xor_;

  // Cache
  UINT32 critical_word_size_;
  UINT32 bank_offset_;
  UINT64 capacity_;

  UINT32 id_;
  static UINT32 next_id;
};

}  // namespace drsim

#endif
