/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 *          Doe Hyun Yoon
 */

#ifndef ARGS_H_
#define ARGS_H_

#include <boost/format.hpp>
#include "common.h"
#include "dram.h"
#include "addrdecoder.h"
#include "scheduler.h"

namespace drsim {

using std::endl;

class AnyOption;

/**
 * Defines all the simulation parmeters.
 * When adding a new parameter, also add getter/setter, and a default value
 * initilization code in the constructor.
 */
class Args {
  // System Configuration
  UINT64 dram_freq_;
  UINT32 num_channel_;

  // Channel Configuration
  DRAMModule module_config_;
  DRAMIddSpec module_idd_spec_;
//  double module_vdd_; // need for a power calc, which isn't working for now.

  // Scheduler Configuration
  UINT32 read_queue_size_;
  UINT32 write_queue_size_;
  string scheduler_;
  bool closed_page_;
  bool wmode_allow_read_CAS_;
  bool rmode_allow_write_ACTPRE_;
  UINT32 wmode_enter_threshold_;
  UINT32 wmode_exit_threshold_;
  UINT32 age_promo_threshold_;
  UINT32 powerdown_;

  bool per_req_policy_;
  bool split_;
  bool cw_priority_;
  bool use_repeat_queue_;
  UINT32 rot_q_search_off_;
  UINT32 multi_cas_;
  bool highp8_;
  bool split_cwf_;
  bool cg_priority_;
  UINT32 sched_lat_;

  // Decoder Configuration
  bool bank_interleave_;
  bool rank_interleave_;
  UINT32 subrank_width_;
  bool bank_xor_hash_;
  bool subrank_xor_hash_;
  bool channel_xor_hash_;
  UINT32 subrank_interleaving_granularity_;
  UINT32 channel_interleaving_granularity_;
  // old interface
  SubrankMap subrank_map_;
  UINT32 num_subrank_;
//  string addr_mask_;
//  string xor_mask_;

 public:
  // Default values
  Args() {
    // System Configuration
    dram_freq_ = 800;    // 800MHz
    num_channel_ = 1;

    // Channel Configuration
    // 1*8*8192*1024*64/8 = 512MB
    module_config_.num_rank = 1;
    module_config_.num_chip = 8;
    module_config_.num_bank = 8;
    module_config_.num_row = 32768;
    module_config_.num_column = 1024;
    module_config_.chip_width = 1;
    module_config_.burst_length = 8;
    module_config_.num_ABUS = 1;

    // Micron DDR3-1600 x8
    module_idd_spec_.IDD0    = 95;   // ACT-PRE
    module_idd_spec_.IDD1    = 105;  // ACT-RD-PRE
    module_idd_spec_.IDD2P0  = 12;   // Precharged PDN (slow, locked DLL)
    module_idd_spec_.IDD2P1  = 35;   // Precharged PDN (fast)
//    module_idd_spec_.IDD2Q   = 40;
    module_idd_spec_.IDD2N   = 42;    // Precharged standby
//    module_idd_spec_.IDD2NT  = 50;
    module_idd_spec_.IDD3P   = 40;    // Active PDN
    module_idd_spec_.IDD3N   = 45;    // Active standby
    module_idd_spec_.IDD4R   = 180;   // Read
    module_idd_spec_.IDD4W   = 185;   // Write
//    module_idd_spec_.IDD5B   = 220; // Self-refresh
    module_idd_spec_.IDD6    = 12;    // Self-refresh extende temperature
    module_idd_spec_.IDD6ET  = 15;
//    module_idd_spec_.IDD7    = 485;

    // Micron DDR3-1866 x8
//    module_idd_spec_.IDD0    = 105;
//    module_idd_spec_.IDD1    = 110;
//    module_idd_spec_.IDD2P0  = 12;
//    module_idd_spec_.IDD2P1  = 40;
////    module_idd_spec_.IDD2Q   = 45;
//    module_idd_spec_.IDD2N   = 47;
////    module_idd_spec_.IDD2NT  = 55;
//    module_idd_spec_.IDD3P   = 45;
//    module_idd_spec_.IDD3N   = 50;
//    module_idd_spec_.IDD4R   = 200;
//    module_idd_spec_.IDD4W   = 205;
////    module_idd_spec_.IDD5B   = 220;
//    module_idd_spec_.IDD6    = 12;
//    module_idd_spec_.IDD6ET  = 15;
////    module_idd_spec_.IDD7    = 485;

//    module_vdd_ = 1.5;

    // Scheduler Configuration
    read_queue_size_ = 64;
    write_queue_size_ = 64;
    scheduler_ = "FR_FCFS";
    closed_page_ = false;
    wmode_allow_read_CAS_ = false;
    rmode_allow_write_ACTPRE_ = false;
    wmode_enter_threshold_ = 48;
    wmode_exit_threshold_ = 16;
    age_promo_threshold_ = 1000;
    powerdown_ = 0;

    per_req_policy_ = false;
    split_ = false;
    cw_priority_ = false;
    use_repeat_queue_ = false;
    rot_q_search_off_ = 0;
    multi_cas_ = 0;
    highp8_ = false;
    split_cwf_ = false;
    sched_lat_ = 0;
    cg_priority_ = false;

    // Decoder Configuration
    bank_interleave_ = false;
    rank_interleave_ = true;
    subrank_width_ = 8;
    bank_xor_hash_ = true;
    subrank_xor_hash_ = true;
    channel_xor_hash_ = false;
    subrank_interleaving_granularity_ = 64;
    channel_interleaving_granularity_ = 64;
    subrank_map_ = INVALID_SUBRANK_MAP;
    num_subrank_ = 0;
//    addr_mask_ = "";
//    xor_mask_ = "";
  };

  void parse_input_args(AnyOption* opt, int argc, char *argv[]);
  void parse_input_args(int argc, char *argv[]);

  virtual ~Args() {}

  GETSET(UINT32, num_channel);
  GETSET(UINT64, dram_freq);

  // old interface
  GETSET(UINT32, num_subrank);
  GETSET(SubrankMap, subrank_map);

  GETSET(bool, bank_interleave);
  GETSET(bool, rank_interleave);
//  GETSET(string, addr_mask);
//  GETSET(string, xor_mask);
  GETSET(bool, channel_xor_hash);
  GETSET(bool, subrank_xor_hash);
  GETSET(bool, bank_xor_hash);
  GETSET(UINT32, subrank_width);
  GETSET(UINT32, subrank_interleaving_granularity);
  GETSET(UINT32, channel_interleaving_granularity);

  GETSET(UINT32, sched_lat);
  GETSET(UINT32, read_queue_size);
  GETSET(UINT32, write_queue_size);
  GETSET(UINT32, age_promo_threshold);

  GETSET(string, scheduler);
  GETSET(UINT32, wmode_enter_threshold);
  GETSET(UINT32, wmode_exit_threshold);
  GETSET(bool, closed_page);
  GETSET(bool, wmode_allow_read_CAS);
  GETSET(bool, rmode_allow_write_ACTPRE);
  GETSET(bool, per_req_policy);
  GETSET(bool, split);
  GETSET(bool, cw_priority);
  GETSET(bool, use_repeat_queue);
  GETSET(UINT32, powerdown);
  GETSET(bool, cg_priority);
  GETSET(bool, rot_q_search_off);
  GETSET(UINT32, multi_cas);
  GETSET(bool, highp8);
  GETSET(bool, split_cwf);

  GETSET(DRAMModule, module_config);
  GETSETMEMBER(UINT32, module_config, num_rank);
  GETSETMEMBER(UINT32, module_config, num_chip);
  GETSETMEMBER(UINT32, module_config, num_bank);
  GETSETMEMBER(UINT32, module_config, num_row);
  GETSETMEMBER(UINT32, module_config, num_column);
  GETSETMEMBER(UINT32, module_config, burst_length);
  GETSETMEMBER(UINT32, module_config, num_ABUS);
  GETSETMEMBER(UINT32, module_config, chip_width);

//  GETSET(double, module_vdd);
  GETSET(DRAMIddSpec, module_idd_spec);
  GETSETMEMBER(UINT32, module_idd_spec, IDD0);
  GETSETMEMBER(UINT32, module_idd_spec, IDD1);
  GETSETMEMBER(UINT32, module_idd_spec, IDD2P0);
  GETSETMEMBER(UINT32, module_idd_spec, IDD2P1);
  GETSETMEMBER(UINT32, module_idd_spec, IDD2N);
  GETSETMEMBER(UINT32, module_idd_spec, IDD3P);
  GETSETMEMBER(UINT32, module_idd_spec, IDD3N);
  GETSETMEMBER(UINT32, module_idd_spec, IDD4R);
  GETSETMEMBER(UINT32, module_idd_spec, IDD4W);
  GETSETMEMBER(UINT32, module_idd_spec, IDD6);
  GETSETMEMBER(UINT32, module_idd_spec, IDD6ET);

  virtual ostream & print(ostream &os) const {
    using boost::format;
    os << "[System Configuration]" << endl;
    os << format("%-30s : %d") % "DRAM Clock Freq (CK, MHz)" % dram_freq() << endl;
    os << format("%-30s : %d") % "Number of DRAM channels" % num_channel() << endl;
    os << endl;

    os << "[Channel Configuration]" << endl;
    os << module_config();
//    os << module_idd_spec_ << endl;
    os << endl;

    os << "[Scheduler Configuration]" << endl;
    os << format("%-30s : %d") % "Read Queue Size" % read_queue_size() << endl;
    os << format("%-30s : %d") % "Write Queue Size" % write_queue_size() << endl;
    os << format("%-30s : %s") % "Scheduler" % scheduler() << endl;
    os << format("%-30s : %s") % "Closed Page" % closed_page() << endl;
    os << format("%-30s : %s") % "WMode Allow READ CAS" % wmode_allow_read_CAS() << endl;
    os << format("%-30s : %s") % "RMode Allow WRITE ACT/PRE" % rmode_allow_write_ACTPRE() << endl;
    os << format("%-30s : %s") % "WMode Enter Threshold" % wmode_enter_threshold() << endl;
    os << format("%-30s : %s") % "WMode Exit Threshold" % wmode_exit_threshold() << endl;
    os << format("%-30s : %s") % "Starvation Threshold" % age_promo_threshold() << endl;
    os << format("%-30s : %s") % "Powerdown" % powerdown_ << endl;
    // Scheduler Experimental
    os << format("%-30s : %s") % "per_req_policy:" % per_req_policy_ << endl;
    os << format("%-30s : %s") % "split:" % split_ << endl;
    os << format("%-30s : %s") % "cw_priority:" % cw_priority_ << endl;
    os << format("%-30s : %s") % "use_repeat_queue:" % use_repeat_queue_ << endl;
    os << format("%-30s : %s") % "rot_q_search_off:" % rot_q_search_off_ << endl;
    os << format("%-30s : %s") % "multi_cas:" % multi_cas_ << endl;
    os << format("%-30s : %s") % "highp8:" % highp8_ << endl;
    os << format("%-30s : %s") % "split_cwf:" % split_cwf_ << endl;
    os << endl;

    os << "[Decoder Configuration]" << endl;
    // os << "channel_interleave:" % channel_interleave_ << endl;
    os << format("%-30s : %s") % "bank_interleave:" % bank_interleave_ << endl;
    os << format("%-30s : %s") % "rank_interleave:" % rank_interleave_ << endl;
    os << format("%-30s : %s") % "subrank width :" % subrank_width_ << endl;
    // os << sig
    os << format("%-30s : %s") % "bank xor_hash:" % bank_xor_hash_ << endl;
    // os << llc_tag_shift
    os << format("%-30s : %s") % "subrank xor_hash:" % subrank_xor_hash_ << endl;
    os << format("%-30s : %s") % "channel xor_hash:" % channel_xor_hash_ << endl;
    os << endl;

    return os;
  }
};

}  // namespace drsim
#endif
