/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Doe Hyun Yoon
 */

#ifndef CLOCK_H_
#define CLOCK_H_

#include <cmath>
#include "common.h"

namespace drsim {

/*
 * clock object:
 *   dram_clk: DRAM address bus clock
 *   sys_clk: system (processor, cache, or on-chip network) clock
 *   master_clk: fake clock with frequency of integer multiple (L.C.M) of both dram_clk and sys_clk
 * NOTE: this is a static object -- there's only ONE clock; no need to instantiate an object
 * ASSUMPTIONS:
 *   (i) freqeuncy granularity is 33MHz, though we accept clock in MHz unit
 *       e.g., 1001MHz, 1003MHz, 997MHz all treated as 1000MHz
 *   (ii) xSysClkTick() is vaild only when sys_clk is faster than dram_clk
 */

class clock {
  // configurations
  static UINT64 s_freqDRAM;  // dram_clk frequency in MHz
  static UINT64 s_freqSys;  // sys_clk frequency in MHz
  static UINT64 s_freqMaster;  // master_clk frequency in MHz

  static TIME s_clkMaster_to_Sys;  // how many master cycles tick a sys_clk
  static TIME s_clkMaster_to_DRAM;  // how many dram cycles tick a dram_clk

  // current clock time in dram_clk, sys_clk, and master_clk
  static TIME s_timeDRAM;
  static TIME s_timeSys;
  static TIME s_timeMaster;

  // to check whether dram_clk or sys_clk changes upon master_clk ticks
  static bool s_bTickDRAM;
  static bool s_bTickSys;

  // it's used only in xInitialize
  static int gcd(const UINT64 aa, const UINT64 bb);
 public:
  static void xInitialize(const UINT64 freqDRAM, const UINT64 freqSys);

  // advance sys_clk AND master_clk/dram_clk accordingly:
  // use only when sys_clk is faster than dram_clk
  static void xSysClkTick(void);

  // advance master_clk AND sys_clk/dram_clk accordingly
  static void xTick(void);

  // functions that return frequency setting
  static UINT64 get_dram_freq(void) {
    return s_freqDRAM;
  }
  static UINT64 get_sys_freq(void) {
    return s_freqSys;
  }
  static UINT64 get_master_freq(void) {
    return s_freqMaster;
  }

  // functions that return current time in dram_clk, sys_clk, and master_clk
  static TIME get_dram_clk(void) {
    return s_timeDRAM;
  }
  static TIME get_sys_clk(void) {
    return s_timeSys;
  }
  static TIME get_master_clk(void) {
    return s_timeMaster;
  }

  // check whether dram_clk or sys_clk changed
  static bool is_dram_ticked(void) {
    return s_bTickDRAM;
  }
  static bool is_sys_ticked(void) {
    return s_bTickSys;
  }

  // convert clock numbers of a domain to another domain
  // note: return values are double precision floating point
  static double dram_clk_to_sys_clk(TIME clocks_in_dram_clk) {
    return ((double) s_clkMaster_to_DRAM ) * ((double) clocks_in_dram_clk)
           / ((double) s_clkMaster_to_Sys);
  }
  static double dram_clk_to_master_clk(TIME clocks_in_dram_clk) {
    return ((double) s_clkMaster_to_DRAM) * ((double) clocks_in_dram_clk);
  }
  static double sys_clk_to_dram_clk(TIME clocks_in_sys_clk) {
    return ((double) s_clkMaster_to_Sys) * ((double) clocks_in_sys_clk)
           / ((double) s_clkMaster_to_DRAM);
  }
  static double sys_clk_to_master_clk(TIME clocks_in_sys_clk) {
    return ((double) s_clkMaster_to_Sys) * ((double) clocks_in_sys_clk);
  }
  static double master_clk_to_dram_clk(TIME clocks_in_master_clk) {
    return ((double) clocks_in_master_clk) / ((double) s_clkMaster_to_DRAM);
  }
  static double master_clk_to_sys_clk(TIME clocks_in_master_clk) {
    return ((double) clocks_in_master_clk) / ((double) s_clkMaster_to_Sys);
  }
};

}  // namespace drsim

#endif /*CLOCK_H_*/
