/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 *          Doe Hyun Yoon
 */

#ifndef COMMON_H_
#define COMMON_H_

#define __STDC_LIMIT_MACROS 1

#include <stdint.h>
#include <assert.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <sstream>

namespace drsim {

typedef char INT8;
typedef unsigned char UINT8;
typedef short INT16;
typedef unsigned short UINT16;
typedef int INT32;
typedef unsigned int UINT32;
typedef long long int INT64;
typedef unsigned long long int UINT64;
typedef UINT64 ADDRESS;
typedef INT64 TIME;

#ifndef INT64_MAX
# if __WORDSIZE == 64
#  define __INT64_C(c)	c ## L
#  define __UINT64_C(c)	c ## UL
# else
#  define __INT64_C(c)	c ## LL
#  define __UINT64_C(c)	c ## ULL
# endif

/* Limits of integral types.  */

/* Minimum of signed integral types.  */
# define INT8_MIN		(-128)
# define INT16_MIN		(-32767-1)
# define INT32_MIN		(-2147483647-1)
# define INT64_MIN		(-__INT64_C(9223372036854775807)-1)
/* Maximum of signed integral types.  */
# define INT8_MAX		(127)
# define INT16_MAX		(32767)
# define INT32_MAX		(2147483647)
# define INT64_MAX		(__INT64_C(9223372036854775807))
#endif
const TIME TIME_MAX = INT64_MAX;

// define this to ignore DRAM REFRESH
#define IGNORE_REFRESH

// for readability in stat print out
inline std::string fmt_str_ch(std::string name, UINT32 channel_id) {
  std::ostringstream oss(std::ostringstream::out);
  oss << name << "[" << channel_id << "]";
  return oss.str();
}
inline std::string fmt_str_ch(std::string name, UINT32 channel_id, UINT32 x) {
  std::ostringstream oss(std::ostringstream::out);
  oss << name << "[" << channel_id << "]" << "[" << x << "]";
  return oss.str();
}
inline std::string fmt_str_ch(std::string name, UINT32 channel_id, std::string suffix) {
  std::ostringstream oss(std::ostringstream::out);
  oss << name << "[" << channel_id << "] ";
  oss << suffix;
  return oss.str();
}

inline std::string formatNumberBase(UINT64 v, UINT32 base) {
  std::string digits = "0123456789abcdef";
  std::string result;
  if ((base < 2) || (base > 16)) {
    result = "Error: base out of range.";
  }
  else {
    do {
      result = digits[v % base] + result;
      v /= base;
    }
    while (v);
  }
  return result;
}

/**
 * Simulator state assertion. abort() to generate core dump.
 * $ ulimit -c unlimited
 * enables core dumps.
 */
#define ASSERT(exp) \
  if (!(exp)) { std::cerr << "Assertion failed. " \
    << __FILE__ << ":"  << __LINE__ << " (" << __STRING(exp) << "). " \
    << std::endl; abort(); \
  }

/**
 * Corrupt simulator state. abort() to generate core dump.
 */
#define PANIC(...) \
  { std::cerr << "Panic. Should never reach here." \
    << __FILE__ << ":"  << __LINE__ << " " << __VA_ARGS__ \
    << std::endl; abort(); \
  }

/**
 * User input assertion.
 */
#define PARAMCHECK(exp, ...) \
  if (!(exp)) { std::cerr << "User input & param sanity check failed. " \
    << __FILE__ << ":"  << __LINE__ << " (" << __STRING(exp) << "). " \
    << __VA_ARGS__ << std::endl; exit(-1); \
  }

/**
 * These macros generate getter and setter funcs.
 */
#define GETSETMEMBER(type, member, var) \
  type get_##var() const \
  {\
     return member##_.var; \
  }\
  void set_##var(type val) \
  {\
     member##_.var = val; \
  }

#define GETSET(type, var) \
  type var() const \
  {\
     return var##_; \
  }\
  void set_##var(type val) \
  {\
     var##_ = val; \
  }

std::ostream & print_svn_rev(std::ostream &os);
std::ostream & print_banner(std::ostream &os);

}  // namespace drsim

#endif /*COMMON_H_*/
