/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 *          Doe Hyun Yoon
 */

#ifndef DRAM_H_
#define DRAM_H_

#include "common.h"

using std::ostream;
using std::string;

#define MAX_NUM_SUBRANK 8

namespace drsim {

class ChipMask {
 public:
  ChipMask();
  UINT32 & operator[] (UINT32 i) { return chip_i[i]; }
  const UINT32 & operator[] (UINT32 i) const { return chip_i[i]; }
  ChipMask operator&& (ChipMask b);
  ChipMask operator|| (ChipMask b);
  ChipMask operator* (UINT32 b);
  ChipMask operator- (ChipMask b);
  ChipMask operator-= (ChipMask b);
  void SetAll(UINT32 i);
  bool Empty() const;
  bool AndNotEmpty(ChipMask const& b);
  bool IsSet(INT32 i, INT32 n);
  void Mask(ChipMask const& b);
  void Normalize();

  friend ostream & operator<<(ostream &, const ChipMask);

  static INT32 num_chip;  // a cache of DramModule::num_chip

 protected:
  UINT32 chip_i[MAX_NUM_SUBRANK];
  INT32 first_nz_i;
};

// A pair of <physical_address, size> maps to one DRAMAddr
struct DRAMAddr {
  UINT32 channel;
  UINT32 rank;
  UINT32 bank;
  UINT32 row;
  UINT32 column;

  // ex)
  // chip_cmd_count = 00000088
  // num_chip_accessed = 2
  // num_cmd_repeat = 8
  // chip_offset = 6
  ChipMask chip_cmd_count;
  INT32 critical_word_chip;
  UINT32 num_chip_accessed;
  UINT32 num_cmd_repeat;
//  UINT32 chip_offset;

  DRAMAddr();
  void reset();
  bool isSamePerChipAddr(const DRAMAddr &b) const;
  bool isSrConflict(const DRAMAddr &b) const;
  bool isSameRow(const DRAMAddr &b) const;
  bool isSameBank(const DRAMAddr &b) const;
  bool isBankConflict(const DRAMAddr &b) const;
};

// [channel:rank:bank:{subranks}:row:column]
ostream & operator<<(ostream & os, const DRAMAddr & dram_addr);
ostream & operator<<(ostream & os, const bool* dram_addr);

void boolArrayOR(bool* dest_src1, bool* src2, INT32 length);
INT32 boolArrayAND(bool* dest_src1, bool* src2, INT32 length);
void boolArrayNeg(bool* dest_src1, INT32 length);

enum eMemCmd {
  Cmd_ACT = 0,  // activate
  Cmd_RD,       // Column Read
  Cmd_RDAP,     // Column Read with precharge
  Cmd_WR,       // Column Write
  Cmd_WRAP,     // Column Write with precharge
  Cmd_PRE,      // Precharge (a bank)
  Cmd_PREA,     // Precharge All Banks
  Cmd_REF,      // Refresh All banks
  Cmd_PDN,      // Power down
  Cmd_PUP,      // Power up
  Cmd_NOP,      // NO operation
  NUM_Cmd,
};
// name of commands in std::string -- for debugging
extern string g_stringMemCmd[NUM_Cmd];
ostream & operator<<(ostream & os, eMemCmd dram_cmd);

enum PowerState {
  ACTIVE = 0,      // at least one bank activated
  ACTIVE_PDN,      // active power-down
  PRE,             // all bank precharged
  PRE_PDN,         // precharged power-down
  PRE_PDN_DLLOFF,  // PRE_PDN with DLL off
  NUM_PowerState,
  INVALID_PowerState,
};
extern string g_stringPowerState[NUM_PowerState];
ostream & operator<<(ostream & os, PowerState state);
bool isPowerDown(PowerState state);

struct DRAMModule {
  // Module
  UINT32 num_rank;
  UINT32 num_chip;
  UINT32 num_ABUS;
  // Chip
  UINT32 num_bank;
  UINT32 num_row;
  UINT32 num_column;
  UINT32 chip_width;
  UINT32 burst_length;
};
ostream & operator<<(ostream & os, const DRAMModule & module);

struct DRAMIddSpec {
  UINT32 IDD0;    // ACT-PRE
  UINT32 IDD1;    // ACT-RD-PRE
  UINT32 IDD2P0;  // Precharged PDN (slow, locked DLL)
  UINT32 IDD2P1;  // Precharged PDN (fast)
//  UINT32 IDD2Q;
  UINT32 IDD2N;   // Precharged standby
//  UINT32 IDD2NT;
  UINT32 IDD3P;   // Active PDN
  UINT32 IDD3N;   // Active standby
  UINT32 IDD4R;   // Read
  UINT32 IDD4W;   // Write
//  UINT32 IDD5B;
  UINT32 IDD6;    // Self-refresh
  UINT32 IDD6ET;  // Self-refresh extende temperature
//  UINT32 IDD7;
//  UINT32 IDD8;
};
ostream & operator<<(ostream &, const DRAMIddSpec &);

namespace DRAMTiming {
  extern const UINT32 DataClockRate;  // 2 for double data rate
  extern UINT32 ABUS_delay;  // ABUS additional delay to model ABUS register

  // timing in unit of DRAM clock cycles (ABUS)
  extern UINT32 AL;      // Added Latency to column access for posted cas
  extern UINT32 BURST;   // data BURST duration in ABUS cycle: 2 for DDR2, 4 for DDR3
  extern UINT32 CAS;     // Column Access Strobe latency: column read to data burst
  extern UINT32 CCD;     // Column-to-Column Delay: intra-device column to column command
  extern UINT32 CMD;     // CoMmanD transport duration: IT SHALL BE 1
  extern UINT32 CWD;     // Column Write Delay: time btw column write and data to DRAM
  extern UINT32 FAW;     // Four Activation Window: a rolling time-frame of 4 ACTs in a rank
  extern UINT32 OST;     // ODT Switching Time: ODT on-time - ODT off-time
  extern UINT32 RAS;     // Row Access Strobe: time to access a row and data restoration
  extern UINT32 RC;      // Row Cycle: ACT to ACT in a bank
  extern UINT32 RCD;     // Row to Column command Delay: ACT to column command
  extern UINT32 RFC;     // ReFresh Cycle: time btw Refresh and ACT
  extern UINT32 RP;      // Row Precharge: time to precharge a bank
  extern UINT32 RRD;     // Row activation to Row activation Delay: ACT to ACT on a rank
  extern UINT32 RTP;     // Read To Precharge: time btw Column Rd and Precharge
  extern UINT32 RTRS;    // Rank-To-Rank Switching:
  extern UINT32 WR;      // Write Recovery: time btw end of write data burst and Precharg
  extern UINT32 WTR;     // Write to Read: time btw end of write data burst and column read
  extern UINT32 CKEMIN;  // =PD(MIN) CKE MIN pulse width: minimum PDN to PUP
  extern UINT32 XP;      // DLL on, any valid cmd, or DLL off to cmds not requiring locked DLL
  extern UINT32 XPDLL;   // Precharge power-down with DLL off to commands requiring a locked DLL
  extern UINT32 REFI;    // REFresh Interval; refresh period (64ms) / 8k = 7.8us in ABUS clock
  extern UINT32 refresh_defer_allowance;  // up to 8 refreshes can be deferred

  extern void print_dram_timing(ostream &os);
};  // namespace DRAMTiming

}  // namespace drsim

#endif
