/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 *          Doe Hyun Yoon
 */

#ifndef MEMORYCHANNEL_H_
#define MEMORYCHANNEL_H_

#include <list>
#include <sstream>
#include "args.h"
#include "reqbuffer.h"
#include "scheduler.h"
#include "timing_model.h"
#include "clock.h"
#include "stat.h"
#include "pdncontroller.h"

using std::stringstream;

namespace drsim {

class MemoryChannel {
 public:
  MemoryChannel(const Args& args,
                DRAMModule module_cfg,
                UINT32 readq_size,
                UINT32 writeq_size,
                UINT32 chl_id);
  ~MemoryChannel();

  bool isIngressQueueFull();
  bool isReadQueueFull();
  bool isWriteQueueFull();
  bool isReqPending();
  void sendRequest(Req* req);
  void sendRequestInQ(Req* req);
  void cycle();
  void wrapUpStats();

  UINT32 num_rank() const           { return num_rank_; }
  UINT32 num_chip() const           { return num_chip_; }
  UINT32 num_bank() const           { return num_bank_; }
  UINT32 num_row() const            { return num_row_; }
  UINT32 num_column() const         { return num_column_; }
  UINT32 num_interface_bits() const { return num_interface_bits_; }
  UINT32 BL() const                 { return BL_; }
  UINT32 num_ABUS() const           { return num_ABUS_; }
  UINT32 id() const                 { return id_; }
  UINT32 numReqProcessed() const;
  ostream& printStatus(ostream &os);
  static ostream& printChannelStats(ostream &os);

 protected:
  bool CheckWriteBufferHit(Req* req);
  void CheckReqsFinished();
  void UpdatePriorities();
  void ScheduleDRAMCommand();
  void UpdateRowStatusInQueues(const DRAMAddr& daddr, eMemCmd cmd, Req* req);
  void UpdateRowStatusInQueues(const DRAMAddr& daddr, eMemCmd cmd, Req* req, ReqBuffer* buffer);

  void PreIssueStatUpdate(const DRAMAddr& daddr, eMemCmd cmd, Req* req);
  void IssueDRAMCommand(DRAMAddr addr, eMemCmd cmd, Req* req);
  void PostIssueStatUpdate(const DRAMAddr& daddr, eMemCmd cmd, Req* req);

  void IncrementSrStat(ArrayCounterStat& stat, const DRAMAddr&);
  void CheckRedundantActPre(ReqRowStatus prev_status, Req* req, Req* act_pre_issuer);
  bool FindRepeater(DRAMAddr* daddr, eMemCmd* cmd, Req** req);
  void ReceiveReqs();

  UINT32 id_;
  static list<MemoryChannel*> memChannels;

  // channel configuration
  UINT32 num_rank_;
  UINT32 num_chip_;              // per rank
  UINT32 num_bank_;              // per rank
  UINT32 num_row_;               // per bank
  UINT32 num_column_;            // per row
  UINT32 num_interface_bits_;
  UINT32 BL_;
  UINT32 chip_width_;
  UINT32 num_ABUS_;

  // policy
  UINT32 age_promo_threshold_;
  UINT32 use_powerdown_mode_;
  bool cg_priority_;
  bool split_;
  bool cw_priority_;
  UINT32 multi_cas_;
  bool highp8_;
//  bool use_repeat_queue_;
  bool split_cwf_;

  TIME sched_lat_;

  ReqBuffer* in_queue_;
  ReqBuffer* read_queue_;
  ReqBuffer* write_queue_;
//  RepeatReqBuffer* repeat_queue_;
  InFlightReqBuffer* inflight_queue_;
  cTimingModel* timing_logic_;
  Scheduler* scheduler_;
  PDNController* pdn_control_;

  // power
  DRAMIddSpec dram_idd_spec_;
  double vdd_;
  UINT64 dram_freq_;

  // stats
  struct MCStat {
    #define MAX_SOURCE_ID 8
    explicit MCStat(MemoryChannel* ch);
    void SampleTimeToFinish(Req* req);
    void SampleTimeToPH(Req* req, const DRAMAddr& daddr);
    void SampleTimeToPH(Req* req, const ChipMask& mask, TIME PH_time);
    void SampleTimeToRD(Req* req, const DRAMAddr& daddr);
    void IncrementSrStat(ArrayCounterStat& stat, const DRAMAddr& daddr);

    MemoryChannel* ch_;

    CounterStat num_req_received;
    CounterStat num_ACT_issued;
    CounterStat num_PRE_issued;
    CounterStat num_RD_issued;
    CounterStat num_RDAP_issued;
    CounterStat num_WR_issued;
    CounterStat num_WRAP_issued;
    CounterStat num_REF_issued;
    CounterStat num_req_finished;
    CounterStat num_req_starved;
    CounterStat num_req_promote_large_size;

    // 2D rk*num_sr + sr
    ArrayCounterStat num_cmds_in_ABUS;
    ArrayCounterStat num_req_received_in_sr;
    ArrayCounterStat num_ACT_issued_in_sr;
    ArrayCounterStat num_PRE_issued_in_sr;
    ArrayCounterStat num_RD_issued_in_sr;
    ArrayCounterStat num_RDAP_issued_in_sr;
    ArrayCounterStat num_WR_issued_in_sr;
    ArrayCounterStat num_WRAP_issued_in_sr;
    ArrayCounterStat num_REF_issued_in_sr;
    ArrayCounterStat num_PUP_issued_in_sr;
    ArrayCounterStat num_PDN_issued_in_sr;

    // to keep track all-bank precharged, and powerdown periods
    TIME* last_precharged_in_sr_;
    ArrayCounterStat num_cycles_precharged_in_sr;  // 2D rk*num_sr + sr

    ArrayCounterStat num_rw_cmd_in_sr;  // 1D array counter idx w/ subranks
    ArrayCounterStat num_req_in_rk_bk;  // 2D array counter idx w/ rk*(NUM_BK)+bk
    ArrayCounterStat num_rd_in_rk_bk;  // 2D array counter idx w/ rk*(NUM_BK)+bk
    ArrayCounterStat num_wr_in_rk_bk;  // 2D array counter idx w/ rk*(NUM_BK)+bk

    // per-source stat
    CounterStat* num_req_received_core[MAX_SOURCE_ID];
    CounterStat* num_ACT_issued_core[MAX_SOURCE_ID];
    CounterStat* num_PRE_issued_core[MAX_SOURCE_ID];
    CounterStat* num_ACT_RD_issued_core[MAX_SOURCE_ID];
    CounterStat* num_PRE_RD_issued_core[MAX_SOURCE_ID];
    CounterStat* num_ACT_WR_issued_core[MAX_SOURCE_ID];
    CounterStat* num_PRE_WR_issued_core[MAX_SOURCE_ID];
    CounterStat* num_RD_issued_core[MAX_SOURCE_ID];
    CounterStat* num_WR_issued_core[MAX_SOURCE_ID];

    SSHistogram* dist_read_latency[MAX_SOURCE_ID + 1];
    SSHistogram* dist_time_to_PH[2 * (MAX_SOURCE_ID + 1)];
    SSHistogram* dist_time_to_RD[2 * (MAX_SOURCE_ID + 1)];

    CounterStat num_Redundant_ACT_PRE;
  };

  MCStat stat_;
};

inline bool MemoryChannel::isIngressQueueFull() {
  return (in_queue_->size() == in_queue_->max_size());
}

inline bool MemoryChannel::isReadQueueFull() {
  return (read_queue_->size() == read_queue_->max_size());
}

inline bool MemoryChannel::isWriteQueueFull() {
  return (write_queue_->size() == write_queue_->max_size());
}

inline UINT32 MemoryChannel::numReqProcessed() const {
  return stat_.num_RD_issued.get_counter() + stat_.num_WR_issued.get_counter();
}

}  // namespace drsim

#endif
