/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#ifndef REQBUFFER_H_
#define REQBUFFER_H_

#include <iostream>
#include <list>
#include <vector>
#include <algorithm>
#include "common.h"
#include "dram.h"

using std::ostream;
using std::min;
using std::max;

namespace drsim {

enum ReqRowStatus {
  ROW_HIT,
  ROW_MISS,
  ROW_EMPTY,
  ROW_MIX,
  INVALID_ROW_STATUS
};
ostream& operator<<(ostream& os, const ReqRowStatus reqRowStatus);

enum ReqStatus {
  REQ_WAIT,
  REQ_CAS_STARTED,
  REQ_LAST_CAS_ISSUED,
  INVALID_REQ_STATUS
};
ostream& operator<<(ostream& os, const ReqStatus status);

typedef INT32 Priority;
static Priority const kInvalidLow = 0;
static Priority const kLow = 1;
static Priority const kNormal = 2;
static Priority const kHigh = 3;
static Priority const kPreempting = 4;
static Priority const kStarved = 5;
static Priority const kInvalidHigh = 6;

class Req {
 public:
  Req() {}
  Req(UINT64 id,
      bool is_read,
      ADDRESS phy_addr,
      UINT32 size,
      DRAMAddr dram_addr,
      bool is_prefetch = false,
      INT32 source_id = 0);
  virtual ~Req() {}

  GETSET(UINT32, source_id);
  DRAMAddr dram_addr() const                { return dram_addr_; }
  ChipMask cw_block()                       { return cw_block_; }
  ReqRowStatus row_status() const           { return row_status_; }
  bool subrank_powerdown() const            { return subrank_powerdown_; }
  TIME time_received() const                { return time_received_; }
  TIME time_page_hit() const                { return time_page_hit_;}
  TIME time_issued() const                  { return time_issued_; }
  TIME time_cw_completed() const            { return time_cw_completed_; }
  TIME time_completed() const               { return time_completed_; }
  Priority priority() const                 { return priority_; }
  UINT64 id() const                         { return id_; }
  ReqStatus req_status() const              { return req_status_; }
  ADDRESS phy_addr() const                  { return phy_addr_; }
  UINT32 size() const                       { return size_; }
  bool is_read() const                      { return is_read_; }
  bool is_closed_page() const               { return is_closed_page_; }
  bool allow_rmode_write_ACTPRE() const     { return allow_rmode_write_ACTPRE_; }

  void set_dram_addr(DRAMAddr dram_addr)        { dram_addr_ = dram_addr; }
  void set_cw_block(ChipMask cw_block)          { cw_block_ = cw_block; }
  void set_row_status(ReqRowStatus row_status)  { row_status_ = row_status; }
  void set_subrank_powerdown(bool sr_pdn)       { subrank_powerdown_ = sr_pdn; }
  void set_time_received(TIME time)             { time_received_ = time; }
  void set_time_page_hit(TIME time)             { time_page_hit_ = time; }
  void set_time_issued(TIME time)               { time_issued_ = time; }
  void set_time_cw_completed(TIME time)         { time_cw_completed_ = time; }
  void set_time_completed(TIME time)            { time_completed_ = time; }
  void set_req_status(ReqStatus req_status)     { req_status_ = req_status; }
  void set_phy_addr(ADDRESS phy_addr)           { phy_addr_ = phy_addr; }
  void set_size(UINT32 size)                    { size_ = size; }
  void set_is_closed_page(bool is_closed_page)  { is_closed_page_ = is_closed_page; }
  void set_allow_rmode_write_ACTPRE(bool v)     { allow_rmode_write_ACTPRE_ = v; }

  void RaisePriority();
  void LowerPriority();
  bool CanPreempt() const   { return (priority_ == kPreempting || priority_ == kStarved); }
  void ColumnCmdIssued(ChipMask m)          { dram_addr_.chip_cmd_count -= m; }
  bool IsColumnCmdAllIssued() const         { return dram_addr_.chip_cmd_count.Empty(); }

  virtual void critical_word_callback(INT32 cw_size) {}
  virtual void completed_callback() {}
  virtual void PrintExtra(ostream &os) const {}

  bool ACTissued_;
  TIME time_page_hit_chip_[8];  // this is per-chip

 protected:
  UINT64 id_;           // request id
  bool is_read_;        // false for write
  ADDRESS phy_addr_;    // physical address
  UINT32 size_;         // request size in bytes
  DRAMAddr dram_addr_;  // decoded dram address
  Priority priority_;
  bool subrank_powerdown_;

  ChipMask cw_block_;   // snapshot of (1-dram_addr.chip_cmd_count) when cw was issued

  // debug/stat
  INT32 source_id_;
  bool is_prefetch_;
  ReqRowStatus row_status_;
  ReqStatus req_status_;
  TIME time_received_;
  TIME time_page_hit_;
  TIME time_issued_;
  TIME time_cw_completed_;
  TIME time_completed_;

  // apdmap
  bool is_closed_page_;
  bool allow_rmode_write_ACTPRE_;

 private:
  friend ostream & operator<<(ostream & os, const Req & r);
};

inline void Req::RaisePriority() {
  priority_ = min(priority_ + 1, kInvalidHigh - 1);
};

inline void Req::LowerPriority() {
  priority_ = max(priority_ - 1, kInvalidLow + 1);
};

class ReqBuffer {
 public:
  typedef std::vector<Req*>::iterator iterator;

  explicit ReqBuffer(UINT32 max_size) : max_size_(max_size) {}

  UINT32 size() const { return entries_.size(); }
  UINT32 max_size() const { return max_size_; }
  iterator begin() { return entries_.begin(); }
  iterator end() { return entries_.end(); }
//  Req* & operator[] (UINT32 i) { return entries_[i]; }
  Req* const& operator[] (UINT32 i) const { return entries_[i]; }

  void Insert(Req* req);
  void Erase(Req* req);
  iterator Erase(iterator it);
  UINT32 NumReqsToRow(const DRAMAddr& addr);
  UINT32 GetIdleChipMask(UINT32 rank, ChipMask* chip_idle);

  friend ostream & operator<<(ostream &os, ReqBuffer &rb);

 protected:
  std::vector<Req*> entries_;
  UINT32 max_size_;
};

class InFlightReqBuffer : public ReqBuffer {
 public:
  explicit InFlightReqBuffer(UINT32 num_chip);
  void Insert(Req* req);
//  bool isDqFree(UINT32 subrank, TIME from, TIME until) const;

 private:
  UINT32 num_chip_;
//  TIME* dq_busy_until_;
};

class RepeatReqBuffer : public ReqBuffer {
 public:
  explicit RepeatReqBuffer(UINT32 max_size) : ReqBuffer(max_size) {}
  void Insert(Req* req, DRAMAddr daddr);
  void Erase(Req* req);
  void Erase(UINT32 i);
  DRAMAddr& Daddr(UINT32 i) { return daddrs_[i]; }

 private:
  std::vector<DRAMAddr> daddrs_;
};

}  // namespace drsim

#endif
