/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#ifndef SCHEDULER_H_
#define SCHEDULER_H_

#include "reqbuffer.h"
#include "timing_model.h"
#include "stat.h"

namespace drsim {

class Scheduler {
 public:
  Scheduler(UINT32 id,
            ReqBuffer* read_queue,
            ReqBuffer* write_queue,
            InFlightReqBuffer* inflight_read_queue,
            cTimingModel* timing_logic,
            bool closed_page);
  virtual ~Scheduler();
  virtual bool FindCandidate(DRAMAddr* daddr, eMemCmd* dram_cmd, Req** cand) = 0;
  virtual bool IsReadMode() = 0;  // NOTE(mjeong): ugly hack!
  virtual void DumpLog(ostream &os) {}

 protected:
  UINT32 id_;

  ReqBuffer* read_queue_;
  ReqBuffer* write_queue_;
  InFlightReqBuffer* inflight_read_queue_;
  cTimingModel* timing_logic_;

  bool keep_search_offset_;
  UINT32 rq_search_offset_;
  UINT32 wq_search_offset_;
  bool closed_page_;
};

class Scheduler_FIFO : public Scheduler {
 public:
  Scheduler_FIFO(UINT32 id,
                 ReqBuffer* read_queue,
                 ReqBuffer* write_queue,
                 InFlightReqBuffer* inflight_read_queue,
                 cTimingModel* timing_logic,
                 bool closed_page);

  virtual bool FindCandidate(DRAMAddr* daddr, eMemCmd* dram_cmd, Req** cand);
  bool IsReadMode() { return false; }  // NOTE(mjeong): ugly hack!
};

class Scheduler_FRFCFS : public Scheduler {
 public:
  Scheduler_FRFCFS(UINT32 id,
                   ReqBuffer* read_queue,
                   ReqBuffer* write_queue,
                   InFlightReqBuffer* inflight_read_queue,
                   cTimingModel* timing_logic,
                   bool closed_page,
                   UINT32 wmode_enter_threshold,
                   UINT32 wmode_exit_threadhold,
                   bool wmode_allow_read_CAS,
                   bool rmode_allow_write_ACTPRE,
                   bool per_req_policy,
                   bool split,
                   UINT32 rot_q_search_off,
                   UINT32 multi_cas);

  virtual bool FindCandidate(DRAMAddr* daddr, eMemCmd* dram_cmd, Req** cand);
  bool IsReadMode() { return (rwmode_ == READ_MODE); }
  virtual void DumpLog(ostream &os);

 protected:
  enum RWMode { READ_MODE = 0, WRITE_MODE, NUM_RWMode };

  void UpdateRWMode();
  bool FindCASCandidate(Priority p, DRAMAddr* daddr, eMemCmd* dram_cmd, Req** cand);
  bool FindACTCandidate(Priority p, DRAMAddr* daddr, eMemCmd* dram_cmd, Req** cand);
  bool FindPRECandidate(Priority p, DRAMAddr* daddr, eMemCmd* dram_cmd, Req** cand);
  Priority HasBankConflictHit(ReqBuffer* q, const DRAMAddr& addr);
  bool HasHighpReqDelay(DRAMAddr addr, Priority p);
  bool IsCASIssuable(Req const* req, Priority p, DRAMAddr* daddr);
  bool AndPrecharge(Req const* req, DRAMAddr const* daddr);
  bool SearchCASIssuable(ReqBuffer* queue, UINT32 offset, Priority p, DRAMAddr* daddr,
                         eMemCmd* dram_cmd, Req** cand);
  bool IsACTIssuable(Req const* req, Priority p, DRAMAddr* daddr);
  bool IsPREIssuable(Req const* req, Priority p, DRAMAddr* daddr);

  RWMode rwmode_;
  UINT32 wmode_enter_threshold_;
  UINT32 wmode_exit_threshold_;
  bool wmode_allow_read_CAS_;
  bool rmode_allow_write_ACTPRE_;
  bool per_req_policy_;
  bool split_;
  UINT32 rot_q_search_off_;  // 0: no, 1:rotate when the req finished, 2:rotate always
  UINT32 multi_cas_;

  CounterStat num_RM_to_WM_switch_;
  CounterStat num_WM_to_RM_switch_;

  std::stringstream log_buffer_;

 private:
  friend ostream & operator<<(ostream & os, Scheduler_FRFCFS::RWMode mode);
};

inline
ostream & operator<<(ostream & os, Scheduler_FRFCFS::RWMode mode) {
  return os << (mode == Scheduler_FRFCFS::READ_MODE ? "READ_MODE" : "WRITE_MODE");
};

}  // namespace drsim

#endif
