/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#ifndef STAT_H_
#define STAT_H_

#include <string>
#include <iostream>
#include <list>
#include <vector>
#include "common.h"

using std::list;
using std::vector;
using std::string;
using std::ostream;

// Forward declaration
namespace SimpleScalar {
struct stat_stat_t;
}

namespace drsim {

class Stat {
 public:
  explicit Stat(const string &name);

  virtual ~Stat();
  void set_print_default(bool v) {
    print_default_ = v;
  }
  void set_print_periodic(bool v) {
    print_periodic_ = v;
  }
  static void printall(ostream& os);
  static void printall_periodic(ostream &os);

 protected:
  string name_;
  bool print_default_;
  bool print_periodic_;

 private:
  virtual ostream & print(ostream &os) const = 0;
  virtual ostream & print_periodic(ostream &os);

  static list<Stat*> allstat;   // list of all Stat instances

  friend ostream & operator<<(ostream &os, const Stat &stat);
};

class CounterStat : public Stat {
 public:
  explicit CounterStat(const string &name);
  CounterStat();
  ~CounterStat();
  void increment() {
    counter_++;
    counter_periodic_++;
  }
  void increment(INT32 n) {
    counter_ += n;
    counter_periodic_ += n;
  }
  void reset() {
    counter_ = 0;
    counter_periodic_ = 0;
  }
  UINT64 get_counter() const { return counter_; }

 private:
  virtual ostream & print(ostream &os) const;
  virtual ostream & print_periodic(ostream &os);
  UINT64 counter_;
  UINT64 counter_periodic_;
};

class ArrayCounterStat : public Stat {
 public:
  ArrayCounterStat(const string &name, const UINT64 length,
                   const UINT64 num_per_line, bool printPercent = true);
  ~ArrayCounterStat();
  void increment(UINT64 id, UINT64 delta = 1) { counters_[id] += delta; }
  void reset();
  void set(UINT64 id, UINT64 value) { counters_[id] = value; }
  UINT64 get_counter(UINT64 id) const { return counters_[id]; }
  UINT64 get_sum() const;

 private:
  virtual ostream& print(ostream &os) const;
  UINT64 len_;
  UINT64 *counters_;
  // print() parameters
  bool printPercent_;    // print in % of sum across
  UINT64 uiNumPerLine_;
};

// SimpleScalar Histogram Wrapper
class SSHistogram : public Stat {
 public:
  SSHistogram(const char* name, const char *desc);
  ~SSHistogram();
  void add_sample(UINT32 index);

 protected:
  virtual ostream & print(ostream &os) const;
  virtual ostream & print_periodic(ostream &os);

  SimpleScalar::stat_stat_t* hist_;
  SimpleScalar::stat_stat_t* hist_periodic_;
};

}  // namespace drsim

#endif
