/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Doe Hyun Yoon
 *          Min Kyu Jeong
 */

#ifndef TIMING_MODEL_H_
#define TIMING_MODEL_H_

#include <vector>
#include "reqbuffer.h"  // need this for the return type of getBankStatus functions in cTimingModle

namespace drsim {

// TODO(dhyoon): this also can be a static class -- we use a single protocol!
class cProtocol {
    enum eSameDiffAll {
        SAME = 0,
        DIFF,
        ALL,
        NUM_SAME_DIFF_ALL,
    };

    // TODO(dhyoon): can we do verification coverage analysis
    // using which entries are used and which are not?
    UINT32 m_iMinTime[NUM_Cmd][NUM_Cmd][2][2];  // idx w/ [PrevCmd][CurrCmd][same_rank][same_bank]
    UINT32 m_iCompleteTime[NUM_Cmd];

    void set(eMemCmd prevCmd,
             eMemCmd currCmd,
             eSameDiffAll same_rank,
             eSameDiffAll same_bank,
             UINT32 min_time);
    void initializeTable(void);
 public:
    explicit cProtocol(UINT64 dram_freq);
    ~cProtocol(void);

    void xInitializeProtocol(UINT64 dram_freq);

    UINT32 getMinTime(const eMemCmd prevCmd,
                      const eMemCmd currCmd,
                      const bool same_rank,
                      const bool same_bank) const;
    UINT32 getCompleteTime(const eMemCmd currCmd) const;
};

// DRAM timing model
// DRAM model has ch-rk-bk-sr

class cMemReq {
    eMemCmd m_eMemCmd;
    UINT32 m_uiRank;  // here, only bk and rk are important
    UINT32 m_uiBank;
    TIME m_time;
 public:
    cMemReq(void);
    cMemReq(eMemCmd mem_cmd, UINT32 rk, UINT32 bk, TIME time);
    ~cMemReq(void);
    // need a copy constructor ??

    bool sameRank(const cMemReq &otherCmd) const;
    bool sameBank(const cMemReq &otherCmd) const;
    TIME getTime(void) const;
    eMemCmd getCmd(void) const;
};

inline bool cMemReq::sameRank(const cMemReq &otherReq) const {
    return (m_uiRank == otherReq.m_uiRank);
}
inline bool cMemReq::sameBank(const cMemReq &otherReq) const {
    return (m_uiBank == otherReq.m_uiBank);
}
inline TIME cMemReq::getTime(void) const {
    return m_time;
}
inline eMemCmd cMemReq::getCmd(void) const {
    return m_eMemCmd;
}

// to track individual DRAM bank states
class cBank {
    enum eBankState {
        BANK_CLOSE = 0,
        BANK_OPEN,
        NUM_BANK_STATE,
    };
    eBankState m_eBankState;
    ADDRESS m_addrRow;

    cMemReq m_cPrevACT;  // last ACT on this bank
    cMemReq m_cPrevPRE;  // last PRE on this bank
//  cMemReq m_cPrevREF;  // last REF on this bank
    cMemReq m_cPrevRW;   // last RW on this bank

 public:
    cBank(void);
    ~cBank(void);

    bool isBankOpen(void) const;
    ADDRESS getRow(void) const;
    void xClose(const bool bDontCheckOpen = false);
    void xOpen(const ADDRESS row);
    void xIssueReq(const cMemReq &curr_cmd);
    const cMemReq& getPrevACT(void) const;
    const cMemReq& getPrevPRE(void) const;
//  const cMemReq& getPrevREF(void) const;
    const cMemReq& getPrevRW(void) const;
};

class cRank {
    UINT32 m_uiTotalBanks;
    UINT32 m_uiNumOpenBanks;
    TIME m_atimeFourActs[4];  // to keep track of last 4 ACT on this rank
    // maybe power state later!
    TIME m_timeProjectedFinalActivity;  // Not using this at all now

// not quite sure: do we need to manage all ACT/PRE/REF ?
    cMemReq m_cPrevACT;  // last ACT on this rank
    cMemReq m_cPrevPRE;  // last PRE on this rank
    cMemReq m_cPrevPDN;  // last PDN on this (sub)rank
    cMemReq m_cPrevPUP;  // last PUP on this (sub)rank
//  cMemReq m_cPrevREF;  // last REF on this rank

    UINT32 m_uiRefCounter;  // counter for refresh

 public:
    cRank(void);
    explicit cRank(const UINT32 total_banks);
    ~cRank(void);

    void xIssueReq(const cMemReq &curr_req);
    void xOpen(const UINT32 bk);
    void xClose(const UINT32 bk);
    bool violateFAW(const TIME curr_time) const;

    const cMemReq& getPrevACT(void) const;
    const cMemReq& getPrevPRE(void) const;
    const cMemReq& getPrevPDN(void) const;
    const cMemReq& getPrevPUP(void) const;
//  const cMemReq& getPrevREF(void) const;

    UINT32 getRefCounter(void) const;
    void setRefCounter(UINT32 refCounter);
    void incrementRefCounter(void);
};

class cDBUS {
    cMemReq m_cPrevRW;

 public:
    cDBUS(void);
    ~cDBUS(void);

    void xIssueReq(const cMemReq &curr_req);

    // pretty sure that only RW commands that use DBUS need to be tracked
    const cMemReq& getPrevRW(void) const;
};

// top level timing model
// TODO(dhyoon): add an optional Register: adds latency, but may have custom commands?
class cTimingModel {
    static bool s_bInitialized;
    const UINT32 m_uiChannelID;  // channel ID might be needed later!
    const UINT32 m_uiNumRanksPerChannel;
    const UINT32 m_uiNumBanksPerRank;
    const UINT32 m_uiNumSubranksPerRank;
    const UINT32 m_uiNumABUS;

    bool *m_bpSubRankAvailability;  // tracks ABUSes to which reqs are sent THIS cycle

    // states -- indexing: c stands for channel, r for rank, b for bank, and s for subrank
    cDBUS *m_sDBUS;  // 1D array of pointers of cDBus; indexed as [sr]
    cRank **m_rsRank;  // 2D array of pointers of cRank; indexed as [rk][sr]
    cBank ***m_rbsBank;  // 3D array of pointers of cBank; indexed as [rk][bk][sr]

    // a pointer to the config structure to retrieve timing info

    // protocol table
    cProtocol m_cProtocol;

    bool checkViolation(const cMemReq &prev_req, const cMemReq &curr_req) const;

 public:
    cTimingModel(const UINT32 channel_id,
                 const UINT32 num_ranks,
                 const UINT32 num_banks,
                 const UINT32 num_subranks,
                 const UINT32 num_ABUS,
                 const UINT64 dram_freq);
    ~cTimingModel(void);

    // call this function upon a new DRAM cycle -- per cycle initialization
    void xCycle(void);

    // check whether a command is ok to be issued at this cycle
    // this function returns true if it's ok
    bool checkCommand(const eMemCmd cmd, const DRAMAddr daddr) const;
    bool checkCommand(const eMemCmd cmd,
                      const DRAMAddr daddr,
                      bool split,
                      ChipMask* chip_issuable) const;

    // issue a command to update the timing model
    // this function returns the projected completion time of this command
    // valid only for read/write commands
//  void xReserveCommand(const eMemCmd cmd, const DRAMAddr daddr);
    TIME xIssueCommand(const eMemCmd cmd, const DRAMAddr daddr);
    TIME xIssueCommand(const eMemCmd cmd, const DRAMAddr daddr, UINT32 n);

    UINT32 canIssueMoreCommand(void) const;

    // return bank open status: true for open
    bool isBankOpen(const UINT32 rank, const UINT32 bank, const UINT32 subrank) const;
    bool isBankOpen(const UINT32 rank, const ChipMask sr_vector, const UINT32 bank) const;
    bool isBankOpen(const DRAMAddr daddr) const;

    bool getRowStatusChipMask(DRAMAddr* daddr, ReqRowStatus status, bool split) const;

    ReqRowStatus getBankStatus(const UINT32 rank,
                               const UINT32 subrank,
                               const UINT32 bank,
                               const UINT32 row) const;
    ReqRowStatus getBankStatus(const DRAMAddr daddr) const;

//  TIME getLastBankRWTime(UINT32 rank, UINT32 bank, UINT32 subrank) const {
//    return m_rbsBank[rank][bank][subrank].getPrevRW().getTime();
//  }
    TIME getLatency(const eMemCmd cmd) {
        return (m_cProtocol.getCompleteTime(cmd) + DRAMTiming::ABUS_delay );
    }
    TIME getLastActTime(UINT32 rank, UINT32 bank, UINT32 chip) {
        return m_rbsBank[rank][bank][chip].getPrevACT().getTime();
    }

    // return rank open status: true for open
    // TODO(mjeong): we could accelerate this, using open bank counter per sr
    bool isRankOpen(UINT32 rank, UINT32 subrank);

    TIME GetRepeatInterval(eMemCmd cmd) {
        return m_cProtocol.getMinTime(cmd, cmd, true, true);
    }

    // TODO(dhyoon): need Refresh interface!!!!
};

}  // namespace drsim

#endif /*TIMING_MODEL_H_*/
