/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include "addrdecoder.h"
#include "clock.h"
#include "log.h"

using std::cout;
using std::cerr;
using std::endl;
using std::setw;
using std::dec;
using std::hex;
using std::vector;

namespace drsim {

bool isPowerOfTwo(UINT64 x) {
  return ((x != 0) && (x & (x-1)) == 0);
}

UINT32 AddrDecoder::next_id = 0;

AddrDecoder::AddrDecoder(UINT32 num_channel,
                         UINT32 num_rank,
                         UINT32 num_chip,
                         UINT32 num_bank,
                         UINT32 num_row,
                         UINT32 num_column,
                         UINT32 chip_width,
                         UINT32 burst_length,
                         UINT32 channel_interleaving_granularity,
                         bool bank_interleaving,
                         bool rank_interleaving,
                         UINT32 subrank_width,
                         UINT32 subrank_interleaving_granularity,
                         bool bank_xor_hashing,
                         UINT32 bank_xor_hashing_offset,
                         bool subrank_xor_hashing,
                         bool channel_xor_hashing)
    : num_channel_(num_channel),
      num_rank_(num_rank),
      num_chip_(num_chip),
      num_bank_(num_bank),
      num_row_(num_row),
      num_column_(num_column),
      chip_width_(chip_width),
      burst_length_(burst_length),
      channel_interleaving_granularity_(channel_interleaving_granularity),
      bank_interleaving_(bank_interleaving),
      rank_interleaving_(rank_interleaving),
      subrank_width_(subrank_width),
      subrank_interleaving_granularity_(subrank_interleaving_granularity),
      bank_xor_hashing_(bank_xor_hashing),
      bank_xor_hashing_offset_(bank_xor_hashing_offset),
      subrank_xor_hashing_(subrank_xor_hashing),
      channel_xor_hashing_(channel_xor_hashing),
      id_(next_id++) {
  num_channel_log2_ = log2(num_channel_);
  num_rank_log2_ = log2(num_rank_);
  num_chip_log2_ = log2(num_chip_);
  num_bank_log2_ = log2(num_bank_);
  num_row_log2_ = log2(num_row_);
  num_column_log2_ = log2(num_column_);
  chip_width_log2_ = log2(chip_width_);
  burst_length_log2_ = log2(burst_length_);
  subrank_width_log2_ = log2(subrank_width_);
  channel_interleaving_granularity_log2_ = log2(channel_interleaving_granularity_);
  capacity_ = static_cast<UINT64>(num_channel) * num_rank * num_chip * num_bank
              * num_row * num_column * chip_width;

  // TODO(mjeong): These are to unburden the user from calculating interleaving granularity
  //               manually. I haven't figure out the right 'sugar' interface yet.
  //               It will have to embrace channel and subrank together.
  if (subrank_interleaving_granularity == 0) {
    subrank_interleaving_granularity_log2_ = num_column_log2_ + subrank_width_log2_;
    subrank_interleaving_granularity_ = 1 << subrank_interleaving_granularity_log2_;
  } else {
    subrank_interleaving_granularity_log2_ = log2(subrank_interleaving_granularity_);
  }

  PARAMCHECK((num_chip_ >= subrank_width_),
             "Subrank " << subrank_width << " is wider than rank " << num_chip_);
  PARAMCHECK(subrank_width_ * chip_width_ * burst_length_ <= subrank_interleaving_granularity_,
             "Subrank interleaving granularity " << subrank_interleaving_granularity_
             << "should be larger than minimum access granularity");
  // NOTE(mjeong): For now, I am assuming cig to be greater than sig. Using narrower subrank
  //               while interleaving between channels makes little sense. It might makes sense
  //               in the NUMA CMP context where MCs are distributed.
//  PARAMCHECK((num_chip_ == subrank_width_)
//             || (num_channel == 1)
//             || (subrank_interleaving_granularity_ < channel_interleaving_granularity_));

  CalculateMask();
}

// Example) Eight x8 chips per rank, burst length = 8
//  Chip offset is 1B * 8 = 8B, so the last 3 bits are offset.
//  For the subrank_width of N, adjacent N chips are accessed in tandem. That means consecutive
// 8N bytes are stored in adjacent N chips. Therefore the log2(N) LSb are the subrank field.
//  The rest of the subrank field, (3-N) bits in this example, are placed at the offset of
// subrank_interleave_granularity.
//  The channel field is placed at the offset given by the parameter
// 'channel_interleave_granularity'.
// Care should be taken to avoid the overlap of subrank field and channel field.
//
// subrank_width |98765432109876|543|210
// ------------------------------------
// 1             |sss           |   |CCC
// 2             |ss            |  s|CCC
// 4             |s             | ss|CCC
// 8             |              |sss|CCC
//
void AddrDecoder::CalculateMask() {
  vector<char> mask;

  UINT32 chip_access_granularity_log2 = log2(chip_width_ * burst_length_);
  mask.insert(mask.end(), chip_access_granularity_log2, 'C');
  mask.insert(mask.end(), subrank_width_log2_, 's');

  if (bank_interleaving_) {
    mask.insert(mask.end(), num_bank_log2_, 'b');
    mask.insert(mask.end(), num_column_log2_ - chip_access_granularity_log2, 'C');
  } else {
    mask.insert(mask.end(), num_column_log2_ - chip_access_granularity_log2, 'C');
    mask.insert(mask.end(), num_bank_log2_, 'b');
  }

  if (rank_interleaving_) {
    mask.insert(mask.end(), num_rank_log2_, 'r');
    mask.insert(mask.end(), num_row_log2_, 'R');
  } else {
    mask.insert(mask.end(), num_row_log2_, 'R');
    mask.insert(mask.end(), num_rank_log2_, 'r');
  }

  if (subrank_interleaving_granularity_ < channel_interleaving_granularity_) {
    mask.insert(mask.begin() + (subrank_interleaving_granularity_log2_),
                num_chip_log2_ - subrank_width_log2_, 's');
    mask.insert(mask.begin() + (channel_interleaving_granularity_log2_), num_channel_log2_, 'c');
  } else {
    mask.insert(mask.begin() + (channel_interleaving_granularity_log2_), num_channel_log2_, 'c');
    mask.insert(mask.begin() + (subrank_interleaving_granularity_log2_),
                num_chip_log2_ - subrank_width_log2_, 's');
  }

  vector<char>::reverse_iterator rit = mask.rbegin();
  mask_column_ = mask_subrank_ = mask_bank_ = mask_rank_ = mask_row_ = mask_channel_ = 0;
  for (ADDRESS bit = (ADDRESS)1 << (mask.size() - 1); bit != 0; bit >>= 1, rit++) {
    switch (*rit) {
    case 'C': mask_column_ |= bit; break;
    case 's': mask_subrank_ |= bit; break;
    case 'b': mask_bank_ |= bit; break;
    case 'r': mask_rank_ |= bit; break;
    case 'R': mask_row_ |= bit; break;
    case 'c': mask_channel_ |= bit; break;
    default: PARAMCHECK(false, *rit); break;
    }
  }

  INT32 row_start = FirstNonZeroIndex(mask_row_);
  mask_subrank_xor_ = SetBits(num_chip_log2_ - subrank_width_log2_, row_start);
  mask_channel_xor_ = mask_channel_ << num_channel_log2_;
  if (bank_xor_hashing_offset_ != 0) {
    mask_bank_xor_ = (num_bank_ - 1) << bank_xor_hashing_offset_;
  } else {
    mask_bank_xor_ = mask_bank_ << num_bank_log2_;
  }

  PARAMCHECK((mask_subrank_ & mask_channel_) == 0, "Masks should not overlap");

  // Count the number of colum command repeat needed to serve rank-access-granularity sized request.
  // Divide the size with the number of commands then we get the size per column command.
  UINT32 rank_access_granularity = num_chip_ * chip_width_ * burst_length_;
  UINT32 num_repeat = 0x1 << GetNumOne(((rank_access_granularity - 1) & mask_column_)
                                       >> (chip_width_log2_ + burst_length_log2_));
  critical_word_size_ =  rank_access_granularity / num_repeat;

  bank_offset_ = FirstNonZeroIndex(mask_bank_);
}

INT32 AddrDecoder::FirstNonZeroIndex(ADDRESS addr) {
  UINT32 i = 0;
  ADDRESS bit = 1;
  for (bit = 1; bit && !(bit & addr); bit = bit << 1) {
    i++;
  }
  if (bit)
    return i;
  else
    return -1;
}

ADDRESS AddrDecoder::SetBits(UINT32 num_one, UINT32 offset) {
  return (((ADDRESS)0x1 << num_one) - 1) << offset;
}

ostream& AddrDecoder::PrintMask(ostream &os) {
  os << "decoder id: " << id_ << endl;
  for (INT32 i = sizeof(ADDRESS)*8 - 1; i >= 0; i--) {
    os << (i%10);
  }
  os << endl;

  // Print masks from MSb.
  for (ADDRESS bit = (ADDRESS)1 << (sizeof(ADDRESS) * 8 - 1); bit != 0; bit >>= 1) {
    bool printed = false;
    if (bit & mask_channel_) {
      printed = true;
      os << "c";
    }
    if (bit & mask_rank_) {
      ASSERT(!printed);
      os << "r";
      printed = true;
    }
    if (bit & mask_subrank_) {
      ASSERT(!printed);
      os << "s";
      printed = true;
    }
    if (bit & mask_bank_) {
      ASSERT(!printed);
      os << "b";
      printed = true;
    }
    if (bit & mask_row_) {
      ASSERT(!printed);
      os << "R";
      printed = true;
    }
    if (bit & mask_column_) {
      ASSERT(!printed);
      os << "C";
      printed = true;
    }
    if (!printed) os << "_";
  }
  os << endl;

  if (channel_xor_hashing_) PrintMask(os, mask_channel_xor_, 'c');
  if (subrank_xor_hashing_) PrintMask(os, mask_subrank_xor_, 's');
  if (bank_xor_hashing_) PrintMask(os, mask_bank_xor_, 'b');

  return os;
}

ostream& AddrDecoder::PrintMask(ostream &os, ADDRESS mask, char c) {
  // bit starting with MSb = 1
  for (ADDRESS bit = (ADDRESS)1 << (sizeof(ADDRESS) * 8 - 1); bit != 0; bit >>= 1) {
    os << (bit & mask ? c : '_');
  }
  os << endl;
  return os;
}

ADDRESS AddrDecoder::ApplyMask(ADDRESS a, ADDRESS mask) {
  UINT32 bit_pos = 0;
  ADDRESS anded = a & mask;
  ADDRESS compacted = 0;

  for (UINT32 i = 0; i < sizeof(ADDRESS)*8; i++) {
    if ((mask & 0x1) == 1) {
      compacted |= (anded & 0x1) << bit_pos;
      bit_pos++;
    }
    mask >>= 1;
    anded >>= 1;
  }

  return compacted;
}

UINT32 AddrDecoder::GetChannel(ADDRESS paddr) {
  UINT32 channel = ApplyMask(paddr, mask_channel_);
  if (channel_xor_hashing_) {
    UINT32 channel_xor = ApplyMask(paddr, mask_channel_xor_);
    channel = channel ^ channel_xor;
  }

  return channel;
}

UINT32 AddrDecoder::GetRank(ADDRESS paddr) {
  return (UINT32)ApplyMask(paddr, mask_rank_);
}

UINT32 AddrDecoder::GetSubrank(ADDRESS paddr) {
  UINT32 subrank = ApplyMask(paddr, mask_subrank_);
  if (subrank_xor_hashing_) {
    UINT32 subrank_xor = ApplyMask(paddr, mask_subrank_xor_);
    subrank = subrank ^ (subrank_xor << subrank_width_log2_);
  }

  return subrank;
}

UINT32 AddrDecoder::GetBank(ADDRESS paddr) {
  UINT32 bank = ApplyMask(paddr, mask_bank_);
  if (bank_xor_hashing_) {
    UINT32 bank_xor = ApplyMask(paddr, mask_bank_xor_);
    bank = bank ^ bank_xor;
  }
  return bank;
}

UINT32 AddrDecoder::GetRow(ADDRESS paddr) {
  return (UINT32)ApplyMask(paddr, mask_row_);
}

UINT32 AddrDecoder::GetColumn(ADDRESS paddr) {
  UINT32 column = (UINT32)ApplyMask(paddr, mask_column_);
  return column;
}

UINT32 AddrDecoder::GetNumOne(ADDRESS value) {
  UINT32 count = 0;

  while (value) {
    if (value & 0x1) {
      count++;
    }
    value >>= 1;
  }

  return count;
}

//  How a <physical address, size> pair is decoded using the mask.
//  The pair represents the bytes at [physical_address, physical_address + size).
//
// Example)
// 64B access needs the bytes from
//
// 98765432109876|543|210
// ----------------------
// ??????????????|000|000 to
// ??????????????|111|111
//
//  Assuming chips are 1B wide (x8) and the their burst length is 8 (DDR3), one column command to a
//  chip accesses 8B. For 64B, we need to access 8 of those 8B block. It could be one column command
//  to all 8 chips, two commands to 4 chips, four commands to 2 chips, or eight commands to 1 chip.
//  Decoding determines which chips to access (daddr->subrank_mask), and the number of times to send
//  column commands (daddr->num_cmd_repeat).
//
// Example Mask:
// 98765432109876|543|210
// ----------------------
// RRRRbbbCCCCCss|CCs|CCC
//
// num_cmd_repeat can be determined by counting how many 'C's in the mask are within the access size
// offset. For 64B access, it's bit [0:5]. In the given example mask, for 64B, there are 2 'C's
// (excluding the chip-access-granularity-offset).
// So the num_cmd_repeat = 2^2 = 4
//
// subrank_mask will have as many 1 as the number of chips to access, which is the number of 's's
// within the access size offset. In the given example mask, there is one 's'.
// So there will be 2^1 = 2 ones in the mask. The position of those two ones are determined by
// the value of 's's out of the offset. If the bit[7:6] is 00, then chip 0,1 is accessed.
//
// mask          x   y         subrank_mask
// 98765432109876|543|210      01234567
// ----------------------      --------
// RRRRbbbCCCCCss|CCs|CCC
//             00              11000000
//             01              00110000
//             10              00001100
//             11              00000011
// x = log2(size)
// y = log2(chip_width * burst_length)
//
void AddrDecoder::Decode(ADDRESS paddr, UINT32 size, DRAMAddr* daddr) {
  PARAMCHECK(isPowerOfTwo(size), "Access size is not a power of 2: " << size);
  PARAMCHECK(paddr < capacity_, "Addr is out of range: 0x" << hex << paddr << " > 0x" << capacity_);
//  PARAMCHECK((paddr & (size - 1)) == 0); // Check size-alignment.

  ADDRESS line_addr = (paddr / size) * size;

  daddr->channel = GetChannel(line_addr);
  daddr->rank = GetRank(line_addr);
  daddr->bank = GetBank(line_addr);
  daddr->row = GetRow(line_addr);
  daddr->column = GetColumn(line_addr);

  UINT32 chip_offset = GetSubrank(line_addr);
  ASSERT(chip_offset % subrank_width_ == 0);
  UINT32 num_chip_accessed = ApplyMask((size - 1), mask_subrank_) + 1;
  UINT32 num_cmd_repeat = 0x1 << GetNumOne(((size - 1) & mask_column_)
                                           >> (chip_width_log2_ + burst_length_log2_));

  daddr->chip_cmd_count.SetAll(0);
  for (UINT32 i = 0; i < num_chip_accessed; i++)
    daddr->chip_cmd_count[chip_offset + i] = num_cmd_repeat;
  daddr->num_chip_accessed = num_chip_accessed;
  daddr->num_cmd_repeat = num_cmd_repeat;

  // This assumes a critical word size <= (chip_width * burst_length).
  daddr->critical_word_chip = chip_offset + ApplyMask(paddr, (size - 1) & mask_subrank_);

  DBG_LOG(TRACE_DECODER, "PA 0x" << setw(8) << hex << paddr << " DA " << setw(0) << *daddr << endl);
}

SubrankMap fromString(const string &str) {
  if (str.compare("M0") == 0) {
    return(M0);
  } else if (str.compare("M1") == 0) {
    return(M1);
  } else if (str.compare("M2") == 0) {
    return(M2);
  } else if (str.compare("M3") == 0) {
    return(M3);
  } else if (str.compare("M4") == 0) {
    return(M4);
  } else if (str.compare("M5") == 0) {
    return(M5);
  } else if (str.compare("M6") == 0) {
    return(M6);
  } else {
    PARAMCHECK(false, "Invalid subrank_map string: " << str);
    return INVALID_SUBRANK_MAP;
  }
}

}  // namespace drsim
