/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Doe Hyun Yoon
 *          Min Kyu Jeong
 */

#include <cstdlib>
#include "args.h"
#include "anyoption.h" 
#include "log.h"

namespace drsim {

// macros for option processing
#define SET_OPT(longname, shortname, usage)\
  opt->addUsage( (usage) );\
  opt->setOption( (longname), (shortname) )
#define SET_OPTL(longname, usage)\
  opt->addUsage( (usage) );\
  opt->setOption( (longname) )
#define SET_FLAG(longname, shortname, usage)\
  opt->addUsage( (usage) );\
  opt->setFlag( (longname), (shortname))
#define SET_FLAGS(shortname, usage)\
  opt->addUsage( (usage) );\
  opt->setFlag( (shortname) )

#define GET_FLAG( name )\
  opt->getFlag( (name) )
#define GET_STR( name )\
  opt->getValue( (name) )
#define GET_NUM( name )\
  atoi( opt->getValue( (name) ) )

#define NUM_OPTION( target, name )\
  if( GET_STR( (name) ) != NULL ) { \
    (target) = GET_NUM( (name) );\
  }
#define NUM_OPTIONF( function, name )\
  if( GET_STR( (name) ) != NULL ) { \
    (function)( GET_NUM( (name) ) );\
  }

void Args::parse_input_args(int argc, char *argv[]) {
  AnyOption *opt = new AnyOption(100, 100); // 100 arguments -- probably enough
  opt->setVerbose();
  opt->autoUsagePrint( true );

  parse_input_args(opt, argc, argv);

  delete opt;
}

void Args::parse_input_args(AnyOption* opt, int argc, char *argv[]) {
  opt->addUsage( "Dr.Sim Options:" );

  // system wide options/flags
  opt->addUsage("Global control options:");
  SET_FLAG(    "help",      'h',   "     -h  --help                            Print this help" );
  SET_OPTL(    "config",           "         --config       <filename>         Configuration File Name" );
  SET_OPTL(    "config2",          "         --config2      <filename>         Configuration File Name 2" );
  SET_OPTL(    "config3",          "         --config3      <filename>         Configuration File Name 3" );
  SET_OPTL(    "config4",          "         --config4      <filename>         Configuration File Name 4" );
#ifdef LOG_ENABLE
  SET_OPTL(    "dbg_level",        "         --dbg_level    \"<n>,<n>,...\"    Debug message levels to enable");
  SET_OPTL(    "dbg_file",         "         --dbg_file     <filename>         Debug message output filename");
  SET_OPTL(    "dbg_purge",        "         --dbg_purge    <n>                Debug message purge period in dclk (default:0");
  SET_OPTL(    "dbg_begin",        "         --dbg_begin    <n>                Debug message start dclk (default:0");
  SET_OPTL(    "dbg_end",          "         --dbg_end      <n>                Debug message end dclk (default:0");
#endif
  SET_OPTL(    "dram_freq",        "         --dram_freq    <n>                DRAM Clock Frequency (default:667, unit:MHz, min granularity:33.33MHz)" );
//  SET_OPTL(    "sys_freq",         "         --sys_freq     <n>                System Clock Frequency (default:3000, unit:MHz, min granularity:33.33MHz)" );
  opt->addUsage( "" );

  // scheduling policy args
  opt->addUsage("Memory controller scheduling parameters:");
  SET_OPTL(    "scheduler",        "         --scheduler    <scheduler>        Scheduler: FIFO or FR_FCFS (default:FR_FCFS)");
  SET_OPTL(    "closed_page",      "         --closed_page  <n>                Closed page policy (default:0)" );
  SET_OPTL(    "rq_size",          "         --rq_size      <n>                Read Queue Size (default:32)" );
  SET_OPTL(    "wq_size",          "         --wq_size      <n>                Write Queue Size (default:32)" );
  SET_OPTL(    "wm_enter",         "         --wm_enter     <n>                Write mode enter threshold (default:28)" );
  SET_OPTL(    "wm_exit",          "         --wm_exit      <n>                Write mode exit threshold (default:6)" );
  SET_OPTL(    "wm_allow_read_CAS","         --wm_allow_read_CAS <n>           Allow read CAS in write mode (default:0)" );
  SET_OPTL(    "rm_allow_write_ACTPRE","         --rm_allow_write_ACTPRE <n>       Allow write ACT/PRE in read mode (default:0)" );
  SET_OPTL(    "per_req_policy",   "         --per_req_policy <n>              Use per-request scheduling policy (default:0)" );
  SET_OPTL(    "split",            "         --split <n>                       Issue partial requests if possible (default:0)");
  SET_OPTL(    "cw_priority",      "         --cw_priority <n>                 Downgrade priority of a request once a critical-word is issued (default:0)");
  SET_OPTL(    "use_repeat_queue", "         --use_repeat_queue <n>            Schedule multi-CAS requests back-to-back (default:0)");
  SET_OPTL(    "age_promo_threshold","         --age_promo_threshold <n>         Age promotion threshold in dclk (default: 1000)" );
  SET_OPTL(    "powerdown",        "         --powerdown    <n>                Turn on DRAM powerdown mode (default:0)" );
  SET_OPTL(    "cg_priority",      "         --cg_priority  <n>                Give wide accesses priority over narrow (default:0)" );
  SET_OPTL(    "rot_q_search_off", "         --rot_q_search_off <n>               Advance buffer search start offset always (default:0)" );
  SET_OPTL(    "multi_cas",        "         --multi-cas <n>                   Multi-CAS 0: none, 1: non-CW, 2: all (default:0)");
  SET_OPTL(    "highp8",           "         --highp8    <n>                   High priority for sr_width==8 (default:0)" );
  SET_OPTL(    "split_cwf",        "         --split_cwf <n>                   Split cwf for AGMS+ (default:0)" );
  opt->addUsage( "" );

  // module parameter
  opt->addUsage("DRAM module paramters:");
  SET_OPT(    "channel",  'c',    "     -c  --channel      <n>                Number of DRAM channels (default:1)" );
  SET_OPT(    "rank",      'r',   "     -r  --rank         <n>                Number of Ranks (default:1)" );
  SET_OPTL(   "chip",             "         --chip         <n>                Number of Chips (default:8)" );
  SET_OPT(    "bank",      'b',   "     -b  --bank         <n>                Number of Banks (default:8)" );
  SET_OPTL(   "row",              "         --row          <n>                Number of Rows (default:8192)" );
  SET_OPTL(   "col",              "         --col          <n>                Number of Columns (default:1024)" );
  SET_OPTL(   "num_ABUS",         "         --num_ABUS     <n>                Number of ABUS per channel (default:1)" );
  SET_OPTL(   "chip_width",       "         --chip_width   <n>                DRAM chip width used in bytes. (default:1)" );
  opt->addUsage( "" );

  // address decoder parameter
  opt->addUsage("Physical->DRAM address mapping paramters:");
  SET_OPTL(    "subrank_map",      "         --subrank_map  <sr_map>           Subrank Map: M0, M1, or M2 (default:M1)" );
  SET_OPTL(    "subrank",          "         --subrank      <n>                Deprecated: Number of Subranks (default:2)" );
//  SET_OPTL(    "sr_offset",        "         --sr_offset    <n>                SR offset (default:0)" );
//  SET_OPTL(    "ch_offset",        "         --ch_offset    <n>                CH offset (default:0)" );
//  SET_OPTL(    "addr_mask",        "         --addr_mask    <mask_string>      Address mask string. overrides all other options. use carefully.");
//  SET_OPTL(    "xor_mask",         "         --xor_mask     <mask_string>      XOR hash mask string. use with --addr_mask.");

  SET_OPTL(    "bank_interleave",  "         --bank_interleave <n>             Enable bank interleaving (default:0)" );
  SET_OPTL(    "rank_interleave",  "         --rank_interleave <n>             Enable rank interleaving (default:1)" );
  SET_OPTL(    "ch_xor_hash",      "         --ch_xor_hash  <n>                Enable XOR hasing on channel index (default:1)" );
  SET_OPTL(    "sr_xor_hash",      "         --sr_xor_hash  <n>                Enable XOR hasing on subrank index (default:1)" );
  SET_OPTL(    "bk_xor_hash",      "         --bk_xor_hash  <n>                Enable XOR hasing on bank index (default:1)" );

  SET_OPTL(    "subrank_width",    "         --subrank_width <n>               Subrank width in number of chips (default:8)" );
  SET_OPTL(    "subrank_ig",       "         --subrank_ig    <n>               Subrank interleaving granularity in bytes (default:64)" );
  SET_OPTL(    "channel_ig",       "         --channel_ig    <n>               Channel interleaving granularity in bytes (default:64)" );
  opt->addUsage( "" );

  // DRAM settings
  opt->addUsage("DRAM timing paramters in DRAM clock:");
  SET_OPTL(    "ABUS_delay",       "         --ABUS_delay   <n>                ABUS additional delay to model ABUS register (default:0)" );
  SET_OPTL(    "burst",            "         --burst        <n>                DRAM Burst (default:8)" );
  SET_OPTL(    "AL",               "         --AL           <n>                DRAM Posted Cas Additional Latency (default:0)" );
  SET_OPTL(    "CAS",              "         --CAS          <n>                DRAM CAS latency. =CL (default:10)" );
  SET_OPTL(    "CCD",              "         --CCD          <n>                DRAM Column to Column delay (default:4)" );
  SET_OPTL(    "CMD",              "         --CMD          <n>                DRAM Command transport duration (default:1)" );
  SET_OPTL(    "CWD",              "         --CWD          <n>                DRAM Column Write delay =CWL (default:8)" );
  SET_OPTL(    "FAW",              "         --FAW          <n>                DRAM Four Activation Window (default: 24)" );
  SET_OPTL(    "OST",              "         --OST          <n>                DRAM ODT Switching Time (default:0)" );
  SET_OPTL(    "RAS",              "         --RAS          <n>                DRAM Row activation and data restoration (default:28)" );
  SET_OPTL(    "RC",               "         --RC           <n>                DRAM Row cycle time (default:38)" );
  SET_OPTL(    "RCD",              "         --RCD          <n>                DRAM Ras to Cas time (default:10)" );
  SET_OPTL(    "RFC",              "         --RFC          <n>                DRAM Refresh cycle (default:88)" );
  SET_OPTL(    "RP",               "         --RP           <n>                DRAM Row Precharching (default:10)" );
  SET_OPTL(    "RRD",              "         --RRD          <n>                DRAM ACT to ACT time (default:5)" );
  SET_OPTL(    "RTP",              "         --RTP          <n>                DRAM Read to Precharge (default:6)" );
  SET_OPTL(    "RTRS",             "         --RTRS         <n>                DRAM Row switching time (default:2)" );
  SET_OPTL(    "WR",               "         --WR           <n>                DRAM Write recovery time (default:12)" );
  SET_OPTL(    "WTR",              "         --WTR          <n>                DRAM write to read time (default:6)" );
  SET_OPTL(    "CKEMIN",           "         --CKEMIN       <n>                DRAM CKE MIN pulse width (default:3)" );
  SET_OPTL(    "XP",               "         --XP           <n>                DRAM XP (default:6)" );
  SET_OPTL(    "XPDLL",            "         --XPDLL        <n>                DRAM XPDLL (default:20)" );
  SET_OPTL(    "REFI",             "         --REFI         <n>                DRAM Refresh Interval refresh period (default:6240)" );
  opt->addUsage( "" );

  opt->addUsage("DRAM power paramters:");
  SET_OPTL(    "IDD0",             "         --IDD0         <n>                ACT-PRE                          " );
  SET_OPTL(    "IDD1",             "         --IDD1         <n>                ACT-RD-PRE                       " );
  SET_OPTL(    "IDD2P0",           "         --IDD2P0       <n>                Precharged PDN (slow, locked DLL)" );
  SET_OPTL(    "IDD2P1",           "         --IDD2P1       <n>                Precharged PDN (fast)            " );
  SET_OPTL(    "IDD2N",            "         --IDD2N        <n>                Precharged standby               " );
  SET_OPTL(    "IDD3P",            "         --IDD3P        <n>                Active PDN                       " );
  SET_OPTL(    "IDD3N",            "         --IDD3N        <n>                Active standby                   " );
  SET_OPTL(    "IDD4R",            "         --IDD4R        <n>                Read                             " );
  SET_OPTL(    "IDD4W",            "         --IDD4W        <n>                Write                            " );
  SET_OPTL(    "IDD6",             "         --IDD6         <n>                Self-refresh                     " );
  SET_OPTL(    "IDD6ET",           "         --IDD6ET       <n>                Self-refresh extende temperature " );

  // Interface parameters
  SET_OPTL(    "sched_lat",        "         --sched_lat    <n>                Time until considered for scheduling (defulat:0)" );
  SET_OPTL(    "return_lat",       "         --return_lat   <n>                Time overhead to return (defulat:0)" );

  // Process command-line args
  opt->processCommandArgs(argc, argv);

  // Print usage if necessary
  if (GET_FLAG("help") || GET_FLAG('h')) {
    opt->printUsage();
    delete opt;
    exit(0);
  }

  // Read config file.
  if (GET_STR("config") != NULL) {
    opt->processFile(GET_STR("config"));
    opt->processCommandArgs(argc, argv); // override
  }
  if (GET_STR("config2") != NULL) {
    opt->processFile(GET_STR("config2"));
    opt->processCommandArgs(argc, argv); // override
  }
  if (GET_STR("config3") != NULL) {
    opt->processFile(GET_STR("config3"));
    opt->processCommandArgs(argc, argv); // override
  }
  if (GET_STR("config4") != NULL) {
    opt->processFile(GET_STR("config4"));
    opt->processCommandArgs(argc, argv); // override
  }

  // Debugging log levels
#ifdef LOG_ENABLE
  if ( GET_STR( "dbg_level" ) != NULL ) {
    char* dbg_level = GET_STR( "dbg_level" );
    char* ptr = strtok(dbg_level, ",");
    while (ptr) {
      int level = atoi(ptr);
      ptr = strtok(NULL, ",");
      DBG_LEVEL[level] = true;
    }
  }
  if (GET_STR("dbg_file") != NULL) {
    char* dbg_file = GET_STR("dbg_file");
    strncpy(_dbg_log_filename, dbg_file, 255);
    _dbg_log_filename[255]='\0';
    _dbg_log_out = new std::ofstream(dbg_file);
  } else {
    _dbg_log_out = &(std::cout);
  }
  NUM_OPTION(_dbg_log_purge, "dbg_purge");
  NUM_OPTION(_dbg_log_begin, "dbg_begin");
  NUM_OPTION(_dbg_log_end,   "dbg_end");
#endif

  // Copy the processed args to the Arg structure
  NUM_OPTIONF(set_dram_freq,          "dram_freq");
  NUM_OPTIONF(set_num_channel,        "channel" );
  NUM_OPTIONF(set_num_channel,        'c' );

  NUM_OPTIONF(set_sched_lat,          "sched_lat");
  NUM_OPTIONF(set_read_queue_size,    "rq_size" );
  NUM_OPTIONF(set_write_queue_size,   "wq_size" );
//  NUM_OPTIONF(set_sr_offset,          "sr_offset" );
//  NUM_OPTIONF(set_ch_offset,          "ch_offset" );
  NUM_OPTIONF(set_bank_interleave,    "bank_interleave" );
  NUM_OPTIONF(set_rank_interleave,    "rank_interleave" );
  NUM_OPTIONF(set_age_promo_threshold,   "age_promo_threshold" );
  NUM_OPTIONF(set_channel_xor_hash,        "ch_xor_hash" );
  NUM_OPTIONF(set_subrank_xor_hash,        "sr_xor_hash" );
  NUM_OPTIONF(set_bank_xor_hash,           "bk_xor_hash" );

  NUM_OPTIONF(set_num_subrank,        "subrank" );

  NUM_OPTIONF(set_subrank_width,      "subrank_width");
  NUM_OPTIONF(set_subrank_interleaving_granularity, "subrank_ig");
  NUM_OPTIONF(set_channel_interleaving_granularity, "channel_ig");

//  if ( GET_STR( "addr_mask" ) != NULL )
//    set_addr_mask( GET_STR( "addr_mask" ) );
//
//  if ( GET_STR( "xor_mask" ) != NULL )
//    set_xor_mask( GET_STR( "xor_mask" ) );

  if ( GET_STR( "subrank_map" ) != NULL ) {
    string str( GET_STR( "subrank_map" ) );
    set_subrank_map(fromString(str));
  }

  if ( GET_STR( "scheduler" ) != NULL ) {
    set_scheduler(GET_STR("scheduler"));
  }
  NUM_OPTIONF(  set_closed_page,                "closed_page" );
  NUM_OPTIONF(  set_wmode_enter_threshold,      "wm_enter" );
  NUM_OPTIONF(  set_wmode_exit_threshold,       "wm_exit" );
  NUM_OPTIONF(  set_wmode_allow_read_CAS,       "wm_allow_read_CAS" );
  NUM_OPTIONF(  set_rmode_allow_write_ACTPRE,   "rm_allow_write_ACTPRE" );
  NUM_OPTIONF(  set_per_req_policy,             "per_req_policy" );
  NUM_OPTIONF(  set_split,                      "split" );
  NUM_OPTIONF(  set_cw_priority,                "cw_priority" );
  NUM_OPTIONF(  set_use_repeat_queue,           "use_repeat_queue" );

  NUM_OPTIONF(  set_powerdown,                  "powerdown" );
  NUM_OPTIONF(  set_cg_priority,                "cg_priority" );
  NUM_OPTIONF(  set_rot_q_search_off,           "rot_q_search_off" );
  NUM_OPTIONF(  set_multi_cas,                  "multi_cas" );
  NUM_OPTIONF(  set_highp8,                     "highp8" );
  NUM_OPTIONF(  set_split_cwf,                  "split_cwf" );

  NUM_OPTIONF(  set_num_rank,            "rank" );
  NUM_OPTIONF(  set_num_rank,            'r' );
  NUM_OPTIONF(  set_num_chip,            "chip" );
  NUM_OPTIONF(  set_num_bank,            "bank" );
  NUM_OPTIONF(  set_num_bank,            'b' );
  NUM_OPTIONF(  set_num_row,             "row" );
  NUM_OPTIONF(  set_num_column,          "col" );
  NUM_OPTIONF(  set_burst_length,        "burst" );
  NUM_OPTIONF(  set_num_ABUS,            "num_ABUS" );
  NUM_OPTIONF(  set_chip_width,          "chip_width" );

  NUM_OPTIONF(  set_IDD0,               "IDD0" );
  NUM_OPTIONF(  set_IDD1,               "IDD1" );
  NUM_OPTIONF(  set_IDD2P0,              "IDD2P0" );
  NUM_OPTIONF(  set_IDD2P1,              "IDD2P1" );
  NUM_OPTIONF(  set_IDD2N,              "IDD2N" );
  NUM_OPTIONF(  set_IDD3P,              "IDD3P" );
  NUM_OPTIONF(  set_IDD3N,              "IDD3N" );
  NUM_OPTIONF(  set_IDD4R,              "IDD4R" );
  NUM_OPTIONF(  set_IDD4W,              "IDD4W" );
  NUM_OPTIONF(  set_IDD6,               "IDD6" );
  NUM_OPTIONF(  set_IDD6ET,              "IDD6ET" );

  PARAMCHECK( (module_config().burst_length % DRAMTiming::DataClockRate) == 0, "" );
  DRAMTiming::BURST = module_config().burst_length / DRAMTiming::DataClockRate;
  NUM_OPTION(    DRAMTiming::ABUS_delay,             "ABUS_delay" );
  NUM_OPTION(    DRAMTiming::AL,                     "AL" );
  NUM_OPTION(    DRAMTiming::CAS,                    "CAS" );
  NUM_OPTION(    DRAMTiming::CCD,                    "CCD" );
  NUM_OPTION(    DRAMTiming::CMD,                    "CMD" );
  NUM_OPTION(    DRAMTiming::CWD,                    "CWD" );
  NUM_OPTION(    DRAMTiming::FAW,                    "FAW" );
  NUM_OPTION(    DRAMTiming::OST,                    "OST" );
  NUM_OPTION(    DRAMTiming::RAS,                    "RAS" );
  NUM_OPTION(    DRAMTiming::RC,                     "RC" );
  NUM_OPTION(    DRAMTiming::RCD,                    "RCD" );
  NUM_OPTION(    DRAMTiming::RFC,                    "RFC" );
  NUM_OPTION(    DRAMTiming::RP,                     "RP" );
  NUM_OPTION(    DRAMTiming::RRD,                    "RRD" );
  NUM_OPTION(    DRAMTiming::RTP,                    "RTP" );
  NUM_OPTION(    DRAMTiming::RTRS,                   "RTRS" );
  NUM_OPTION(    DRAMTiming::WR,                     "WR" );
  NUM_OPTION(    DRAMTiming::WTR,                    "WTR" );
  NUM_OPTION(    DRAMTiming::CKEMIN,                 "CKEMIN" );
  NUM_OPTION(    DRAMTiming::XP,                     "XP" );
  NUM_OPTION(    DRAMTiming::XPDLL,                  "XPDLL" );
  NUM_OPTION(    DRAMTiming::REFI,                   "REFI" );
}

}  // namespace drsim
