/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Doe Hyun Yoon
 */

#include "common.h"
#include "clock.h"
#include "log.h"

using std::endl;

namespace drsim {

UINT64 clock::s_freqDRAM;     // dram_clk frequency in MHz
UINT64 clock::s_freqSys;      // sys_clk frequency in MHz
UINT64 clock::s_freqMaster;   // master_clk frequency in MHz

TIME clock::s_clkMaster_to_Sys;   // how many master cycles tick a sys_clk
TIME clock::s_clkMaster_to_DRAM;  // how many dram cycles tick a dram_clk

TIME clock::s_timeDRAM;
TIME clock::s_timeSys;
TIME clock::s_timeMaster;

bool clock::s_bTickDRAM;
bool clock::s_bTickSys;

// greatest common divisor routine from web: forgot the site name
int clock::gcd(const UINT64 aa, const UINT64 bb) {
  UINT64 a = aa;
  UINT64 b = bb;

  while (1) {
    a = a % b;
    if (a == 0) {
      return b;
    }
    b = b % a;
    if (b == 0) {
      return a;
    }
  }
}

void clock::xInitialize(const UINT64 freqDRAM, const UINT64 freqSys) {
  DBG_LOG(TRACE_CLOCK, "DRAM freq: " << freqDRAM << "MHz, Sys Freq: " <<
          freqSys << "MHz" << endl);

  // assign to state variables
  s_freqDRAM = freqDRAM;
  s_freqSys = freqSys;

  // we allow clock frequencies in unit of 100MHz or 33/67 in last two digits
  // (e.g., 133, 167)
  double dFreqDRAM = static_cast<double>(freqDRAM);
  double dFreqSys = static_cast<double>(freqSys);
  double dFreqDRAM_in_100 = 100*round(dFreqDRAM/100.0);
  double dFreqSys_in_100 = 100*round(dFreqSys/100.0);
  bool bDRAMFreq_3x = (dFreqDRAM != dFreqDRAM_in_100);
  bool bSysFreq_3x = (dFreqSys != dFreqSys_in_100);
  // ASSUMPTION: if a freq doesn't end with xx00 MHz, it shall be xx33 MHz or
  // xx67 MHz. multiply by 3 to make it yy00 MHz.

  DBG_LOG(TRACE_CLOCK, "DRAM clock 3x: " << (bDRAMFreq_3x?"yes":"no") <<
          ", Sys clock 3x: " << (bSysFreq_3x?"yes":"no") << endl);

  UINT64 dram_clk_t = bDRAMFreq_3x ? 3*freqDRAM : freqDRAM;
  UINT64 sys_clk_t = bSysFreq_3x ? 3*freqSys : freqSys;

  // need only 100MHz granularity -- also important for xx33 or xx67MHz
  dram_clk_t = 100*round((static_cast<double>(dram_clk_t))/100.0);
  sys_clk_t = 100*round((static_cast<double>(sys_clk_t))/100.0);
  // it's important: 33 is, in fact, 33.3333 and 67 is 66.6666

  DBG_LOG(TRACE_CLOCK, "DRAM clock_t: " << dram_clk_t << "MHz, Sys clock_t: "
          << sys_clk_t << "MHz" << endl);

  UINT64 gcd_d_s = gcd(dram_clk_t, sys_clk_t);  // use greatest common divisor
  DBG_LOG(TRACE_CLOCK, "GCD: " << gcd_d_s << "MHz" << endl);
  ASSERT((dram_clk_t % gcd_d_s) == 0);
  ASSERT((sys_clk_t % gcd_d_s) == 0);
  s_freqMaster = dram_clk_t * sys_clk_t / gcd_d_s;  // it's the least common multiple
  DBG_LOG(TRACE_CLOCK, "Master Freq: " << s_freqMaster << "MHz" << endl);

  s_clkMaster_to_DRAM = s_freqMaster / dram_clk_t;
  s_clkMaster_to_Sys = s_freqMaster / sys_clk_t;

  s_clkMaster_to_DRAM = bDRAMFreq_3x ? 3*s_clkMaster_to_DRAM : s_clkMaster_to_DRAM;
  s_clkMaster_to_Sys = bSysFreq_3x ? 3*s_clkMaster_to_Sys : s_clkMaster_to_Sys;
  DBG_LOG(TRACE_CLOCK, "Master/DRAM: " << s_clkMaster_to_DRAM <<
          ", Master/Sys: " << s_clkMaster_to_Sys << endl);

  s_timeDRAM = 0;
  s_timeSys = 0;
  s_timeMaster = 0;

  s_bTickDRAM = false;
  s_bTickSys = false;
}

void clock::xSysClkTick(void) {
  ASSERT(s_freqSys > s_freqDRAM);
  s_bTickDRAM = false;
  s_bTickSys = true;

  s_timeSys++;
  s_timeMaster += s_clkMaster_to_Sys;
  if ((s_timeMaster - (s_clkMaster_to_DRAM * s_timeDRAM))
      > s_clkMaster_to_DRAM) {
    s_timeDRAM++;
    s_bTickDRAM = true;
    ASSERT((s_timeMaster - (s_clkMaster_to_DRAM * s_timeDRAM))
           < s_clkMaster_to_DRAM);
  }
}
void clock::xTick(void) {
  s_bTickDRAM = false;
  s_bTickSys = false;

  s_timeMaster++;

  if ( (s_timeMaster % s_clkMaster_to_DRAM) == 0 ) {
    s_timeDRAM++;
    s_bTickDRAM = true;
  }
  if ( (s_timeMaster % s_clkMaster_to_Sys) == 0 ) {
    s_timeSys++;
    s_bTickSys = true;
  }
  using std::setw;
  using std::setfill;
#define CDFORMAT setw(4) << setfill('0')
  DBG_LOG(TRACE_CLOCK, "clock::xTick() -- Master: " << CDFORMAT << s_timeMaster
          << "\t Sys: " << CDFORMAT << s_timeSys << " tick: " << (s_bTickSys ? "yes" : "no")
          << "\t DRAM: " << CDFORMAT << s_timeDRAM << " tick: " << (s_bTickDRAM ? "yes" : "no")
          << endl);
#undef CDFORMAT
}

}  // namespace drsim
