/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 *          Doe Hyun Yoon
 */

#include <string>
#include <boost/format.hpp>
#include "common.h"
#include "dram.h"

namespace drsim {

using std::endl;

ChipMask::ChipMask() {
  for (INT32 i = 0; i < MAX_NUM_SUBRANK; i++)
    chip_i[i] = 0;
  first_nz_i = -1;
}

ChipMask ChipMask::operator* (UINT32 b) {
  ChipMask c;
  for (INT32 i = num_chip - 1; i >= 0; i--)
    if ((c[i] = (chip_i[i] * b)))
      c.first_nz_i = i;
  return c;
}

ChipMask ChipMask::operator&& (ChipMask b) {
  ChipMask c;
  for (INT32 i = num_chip - 1; i >= 0; i--) {
    if ((c[i] = (chip_i[i] && b[i])))
      c.first_nz_i = i;
  }
  return c;
}

ChipMask ChipMask::operator|| (ChipMask b) {
  ChipMask c;
  for (INT32 i = num_chip - 1; i >= 0; i--)
    if ((c[i] = (chip_i[i] || b[i])))
      c.first_nz_i = i;
  return c;
}

ChipMask ChipMask::operator- (ChipMask b) {
  ChipMask c;
  for (INT32 i = num_chip - 1; i >= 0; i--)
    if ((c[i] = (chip_i[i] > b[i]) ? chip_i[i] - b[i] : 0))
      c.first_nz_i = i;
  return c;
}

ChipMask ChipMask::operator-= (ChipMask b) {
  *this = *this - b;
  return *this;
}

void ChipMask::SetAll(UINT32 v) {
  for (INT32 i = num_chip - 1; i >= 0; i--)
    chip_i[i] = v;
  first_nz_i = (v != 0 ? 0 : -1);
}

bool ChipMask::Empty() const {
  bool empty = true;
  for (INT32 i = 0; i < num_chip; i++)
    empty &= (chip_i[i] == 0);
  return empty;
}

bool ChipMask::AndNotEmpty(ChipMask const& b) {
  bool exists_one = false;
  for (INT32 i = 0; i < num_chip; i++)
    exists_one = exists_one || (chip_i[i] && b[i]);
  return exists_one;
}

bool ChipMask::IsSet(INT32 offset, INT32 n) {
  bool set = true;
  if (n == 0)
    return false;
  for (INT32 i = 0; i < n; i++)
    set = set && chip_i[offset+i];
  return set;
}

void ChipMask::Mask(ChipMask const& b) {
  for (INT32 i = 0; i < num_chip; i++)
    chip_i[i] = b[i] ? chip_i[i] : 0;
}

void ChipMask::Normalize() {
  for (INT32 i = 0; i < num_chip; i++)
    chip_i[i] = chip_i[i] == 0 ? 0 : 1;
}

INT32 ChipMask::num_chip = 8;

DRAMAddr::DRAMAddr() {
  reset();
}

void DRAMAddr::reset() {
  channel = rank = bank = row = column = num_chip_accessed = num_cmd_repeat = 0;
  chip_cmd_count.SetAll(0);
  critical_word_chip = -1;
}

bool DRAMAddr::isSrConflict(const DRAMAddr &b) const {
  for (INT32 sr = 0; sr < ChipMask::num_chip; sr++) {
    if (chip_cmd_count[sr] > 0 && b.chip_cmd_count[sr] > 0) {
      return true;
    }
  }

  return false;
}

bool DRAMAddr::isSamePerChipAddr(const DRAMAddr &b) const {
  return (channel == b.channel && rank == b.rank && bank == b.bank && row == b.row);
}

bool DRAMAddr::isSameRow(const DRAMAddr &b) const {
  bool sr_conflict = isSrConflict(b);

  return (channel == b.channel && rank == b.rank && bank == b.bank
          && row == b.row && sr_conflict);
}

bool DRAMAddr::isSameBank(const DRAMAddr &b) const {
  bool sr_conflict = isSrConflict(b);

  return (channel == b.channel && rank == b.rank && bank == b.bank
          && sr_conflict);
}

bool DRAMAddr::isBankConflict(const DRAMAddr &b) const {
  bool sr_conflict = isSrConflict(b);

  return (channel == b.channel && rank == b.rank && bank == b.bank
          && row != b.row && sr_conflict);
}

// [channel:rank:bank:{subranks}:row:column]critical_word_chip
ostream & operator<<(ostream & os, const DRAMAddr & dram_addr) {
  os << "[" << dram_addr.channel << ":" << dram_addr.rank << ":" << dram_addr.bank << ":";
  os << dram_addr.chip_cmd_count << ":";
  os << "0x" << std::setfill('0') << std::hex << std::setw(4) << dram_addr.row << ":";
  os << "0x" << std::setfill('0') << std::hex << std::setw(4) << dram_addr.column << "]";
  os << "cw" << dram_addr.critical_word_chip;
  os << std::dec;

  return os;
}

ostream & operator<<(ostream & os, const ChipMask chip_mask) {
  os << "{";
  for (INT32 sr = 0; sr < ChipMask::num_chip; sr++) {
    os << chip_mask[sr];
    if (sr < (ChipMask::num_chip - 1)) {
      os << ":";
    }
  }
  os << "}";
  return os;
}

void boolArrayOR(bool* dest_src1, bool* src2, INT32 length) {
  for (INT32 i = 0; i < length; i++)
    dest_src1[i] = dest_src1[i] || src2[i];
}

INT32 boolArrayAND(bool* dest_src1, bool* src2, INT32 length) {
  INT32 count = 0;
  for (INT32 i = 0; i < length; i++) {
    dest_src1[i] = dest_src1[i] && src2[i];
    if (dest_src1[i]) count++;
  }
  return count;
}

void boolArrayNeg(bool* dest_src1, INT32 length) {
  for (INT32 i = 0; i < length; i++)
    dest_src1[i] = !dest_src1[i];
}

std::string g_stringMemCmd[NUM_Cmd] = {
  "ACT",
  "RD",       // Column Read
  "RDAP",     // Column Read with precharge
  "WR",       // Column Write
  "WRAP",     // Column Write with precharge
  "PRE",      // Precharge (a bank)
  "PREA",     // Precharge All banks
  "REF",      // Refresh All banks
  "PDN",      // Power Down
  "PUP",      // Power Up
  "NOP",      // NO operation
};

ostream & operator<<(ostream & os, eMemCmd dram_cmd) {
  return (os << g_stringMemCmd[dram_cmd]);
};

std::string g_stringPowerState[NUM_PowerState] = {
  "ACTIVE",
  "ACTIVE_PDN",
  "PRE",
  "PRE_PDN",
  "PRE_PDN_DLLOFF"
};
ostream & operator<<(ostream & os, PowerState state) {
  return (os << g_stringPowerState[state]);
};

bool isPowerDown(PowerState state) {
  return (state == ACTIVE_PDN || state == PRE_PDN || state == PRE_PDN_DLLOFF);
}

ostream & operator<<(ostream & os, const DRAMModule & module) {
  using boost::format;
  os << format("%-30s : %s") % "Number of Ranks per channel" % module.num_rank << endl;
  os << format("%-30s : %s") % "Number of Chips per rank" % module.num_chip << endl;
  os << format("%-30s : %s") % "Number of ABUS per channel" % module.num_ABUS << endl;
  os << endl;

  os << "[Chip Configuration]" << endl;
  os << format("%-30s : %s") % "Number of Banks per chip" % module.num_bank << endl;
  os << format("%-30s : %s") % "Number of Rows per bank" % module.num_row << endl;
  os << format("%-30s : %s") % "Number of Columns per row" % module.num_column << endl;
  os << format("%-30s : %s") % "Chip width in Bytes" % module.chip_width << endl;
  os << format("%-30s : %s") % "Burst length" % module.burst_length << endl;

  return os;
}

ostream & operator<<(ostream & os, const DRAMIddSpec & spec) {
  using boost::format;
  os << format("%-30s : %s") % "IDD0" << spec.IDD0 << endl;
  os << format("%-30s : %s") % "IDD1" << spec.IDD1 << endl;
  os << format("%-30s : %s") % "IDD2P0" << spec.IDD2P0 << endl;
  os << format("%-30s : %s") % "IDD2P1" << spec.IDD2P1 << endl;
  os << format("%-30s : %s") % "IDD2N" << spec.IDD2N << endl;
  os << format("%-30s : %s") % "IDD3P" << spec.IDD3P << endl;
  os << format("%-30s : %s") % "IDD3N" << spec.IDD3N << endl;
  os << format("%-30s : %s") % "IDD4R" << spec.IDD4R << endl;
  os << format("%-30s : %s") % "IDD4W" << spec.IDD4W << endl;
  os << format("%-30s : %s") % "IDD6" << spec.IDD6 << endl;
  os << format("%-30s : %s") % "IDD6ET" << spec.IDD6ET << endl;

  return os;
}

// DRAM timing related variables
// borrowed from Buce Jacob's book Table 15.7
namespace DRAMTiming {
const UINT32 DataClockRate = 2;  // 2 for double data rate (DDRx)
// 4 for quad data rate (maybe GDDRx ?)
// 8 for octal data rate (XDR ?)
// 16 for 16x data rate (XDR2 ?), hmm but XDR2's A Bus is also 16x...
UINT32 ABUS_delay = 0;  // additional latency to model ABUS register

// timing in unit of DRAM clock cycles (ABUS)
// numbers from Micron 1Gb DDR3 1600
UINT32 AL = 0;    // Added Latency to column access for posted cas
// I'm not gonna use this stupid posted CAS option...
UINT32 BURST = 4;  // data BURST duration in ABUS cycle: e.g., 2 for DDR2's 4, 4 for DDR3's 8
// DDR3 -- burst 8, but it's double data rate, so 4 cycles in ABUS cycle
// FIXME: Args has DRAMModule that includes BL: BURST should be a half of DRAMModule's BL
UINT32 CAS = 10;  // Column Access Strobe latency: column read to data burst
UINT32 CCD = 4;   // Column-to-Column Delay: intra-device column to column command
UINT32 CMD = 1;   // CoMmanD transport duration: IT SHALL BE 1
UINT32 CWD = 8;   // Column Write Delay: time btw column write and data to DRAM
UINT32 FAW = 24;  // Four Activation Window: a rolling time-frame of 4 ACTs in a rank
UINT32 OST = 0;   // ODT Switching Time: ODT on-time - ODT off-time
// is it correct? Micron 1Gb DRAM has the same ODTL on and ODTL off time
UINT32 RAS = 28;  // Row Access Strobe: time to access a row and data restoration
UINT32 RC = 38;   // Row Cycle: ACT to ACT in a bank
UINT32 RCD = 10;  // Row to Column command Delay: ACT to column command
UINT32 RFC = 88;  // ReFresh Cycle: time btw Refresh and ACT
// 88 for 1Gb, 128 for 2Gb, 240 for 4Gb
UINT32 RP = 10;   // Row Precharge: time to precharge a bank
UINT32 RRD = 5;   // Row activation to Row activation Delay: ACT to ACT on a rank
UINT32 RTP = 6;   // Read To Precharge: time btw Column Rd and Precharge
UINT32 RTRS = 2;  // Rank-To-Rank Switching
// not sure: from Micron Spec. page 142: min RD to WR timing = RL+CCD-WL+ 2clocks
UINT32 WR = 12;   // Write Recovery: time btw end of write data burst and Precharge, 15ns
UINT32 WTR = 6;   // Write to Read: time btw end of write data burst and column read

UINT32 CKEMIN = 3;  // =PD(MIN) CKE MIN pulse width: minimum PDN to PUP
UINT32 XP = 6;      // greater of 3 CLK (1.25*3=3.75ns for 800Mhz) or 7.5ns
UINT32 XPDLL = 20;  // greater of 10 CLK (1.25*10=12.5ns for 800Mhz) or 24ns

UINT32 REFI = 6240;  // REFresh Interval; refresh period (64ms) / 8k = 7.8us
                     // in ABUS clock: 7.8us / 1.25ns = 6240 cycles
UINT32 refresh_defer_allowance = 8;  // up to 8 refreshes can be deferred

void print_dram_timing(ostream &os) {
  os << "[DRAM timing configuration]" << endl;
  os << "Data Clock Rate: " << DataClockRate << endl;
  os << "ABus additional delay: " << ABUS_delay << endl;
  os << "Burst: \t" <<  BURST << " in A Bus clock, \t" << BURST*DataClockRate
     << " in D Bus clock" << endl;
  os << "AL: \t" <<     AL << "\t";
  os << "CAS: \t" <<    CAS << "\t";
  os << "CCD: \t" <<    CCD << "\t";
  os << "CMD: \t" <<    CMD << "\t";
  os << "CWD: \t" <<    CWD << endl;
  os << "FAW: \t" <<    FAW << "\t";
  os << "OST: \t" <<    OST << "\t";
  os << "RAS: \t" <<    RAS << "\t";
  os << "RC: \t" <<     RC << "\t";
  os << "RCD: \t" <<    RCD << endl;
  os << "RFC: \t" <<    RFC << "\t";
  os << "RP: \t" <<     RP << "\t";
  os << "RRD: \t" <<    RRD << "\t";
  os << "RTP: \t" <<    RTP << "\t";
  os << "RTRS: \t"<<   RTRS << endl;
  os << "WR: \t" <<     WR << "\t";
  os << "WTR: \t" <<    WTR << "\t";
  os << "REFI: \t" <<   REFI << endl;
}
}  // namespace DRAMTiming

}  // namespace drsim
