/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 *          Doe Hyun Yoon
 */

#include <sstream>
#include <algorithm>
#include <list>
#include <vector>
#include <stdio.h>
#include <cstring>
#include <boost/format.hpp>
#include "memorychannel.h"
#include "log.h"

using std::endl;
using std::cout;
using std::cerr;
using std::stringstream;

namespace drsim {

list<MemoryChannel*> MemoryChannel::memChannels;

MemoryChannel::MemoryChannel(const Args& args,
                             DRAMModule module_cfg,
                             UINT32 readq_size,
                             UINT32 writeq_size,
                             UINT32 ch_id) :
  id_(ch_id),
  num_rank_(module_cfg.num_rank),
  num_chip_(module_cfg.num_chip),
  num_bank_(module_cfg.num_bank),
  num_row_(module_cfg.num_row),
  num_column_(module_cfg.num_column),
  BL_(module_cfg.burst_length),
  chip_width_(module_cfg.chip_width),
  num_ABUS_(module_cfg.num_ABUS),
  age_promo_threshold_(args.age_promo_threshold()),
  use_powerdown_mode_(args.powerdown()),
  cg_priority_(args.cg_priority()),
  split_(args.split()),
  cw_priority_(args.cw_priority()),
  multi_cas_(args.multi_cas()),
  highp8_(args.highp8()),
//  use_repeat_queue_(args.use_repeat_queue()),
  split_cwf_(args.split_cwf()),
  sched_lat_(args.sched_lat()),
  dram_idd_spec_(args.module_idd_spec()),
//  vdd_(args.module_vdd()),
  dram_freq_(args.dram_freq()),
  stat_(this) {
  memChannels.push_back(this);

  in_queue_ = new ReqBuffer(sched_lat_ + 1);
  read_queue_ = new ReqBuffer(readq_size);
  write_queue_ = new ReqBuffer(writeq_size);
  inflight_queue_ = new InFlightReqBuffer(num_chip_);
//  repeat_queue_ = new RepeatReqBuffer(readq_size + writeq_size);
  timing_logic_ = new cTimingModel(id_, num_rank_, num_bank_,
      num_chip_, num_ABUS_, args.dram_freq());

  if (args.scheduler().compare("FIFO") == 0) {
    scheduler_ = new Scheduler_FIFO(id_,
        read_queue_, write_queue_, inflight_queue_,
        timing_logic_, args.closed_page());
  } else if (args.scheduler().compare("FR_FCFS") == 0) {
    scheduler_ = new Scheduler_FRFCFS(id_,
        read_queue_, write_queue_,
        inflight_queue_, timing_logic_, args.closed_page(),
        args.wmode_enter_threshold(), args.wmode_exit_threshold(),
        args.wmode_allow_read_CAS(), args.rmode_allow_write_ACTPRE(),
        args.per_req_policy(), args.split(), args.rot_q_search_off(), args.multi_cas());
  } else {
    PARAMCHECK(false, "Invalid Schdeuler");
  }

  pdn_control_ = new PDNController(id_, num_rank_, num_chip_,
                                    read_queue_, write_queue_,
                                    inflight_queue_, timing_logic_);
}

MemoryChannel::~MemoryChannel() {
  delete read_queue_;
  delete write_queue_;
  delete inflight_queue_;
  delete timing_logic_;
  delete scheduler_;
  delete pdn_control_;
}

void MemoryChannel::sendRequest(Req* req) {
  ReqRowStatus bank_status = timing_logic_->getBankStatus(req->dram_addr());
  req->set_row_status(bank_status);
  req->set_subrank_powerdown(pdn_control_->is_any_sr_powerdown(req->dram_addr()));
  req->set_time_received(clock::get_dram_clk());
  if (bank_status == ROW_HIT) {
    req->set_time_page_hit(clock::get_dram_clk());
  }
  if (cw_priority_ && multi_cas_ == 0) {
    req->RaisePriority();
  }

  if (req->is_read()) {
    if (CheckWriteBufferHit(req)) {
      // TODO(mjeong): return the read right away
      read_queue_->Insert(req);
    } else {
      read_queue_->Insert(req);
    }

    DBG_LOG(TRACE_MC, *req << " --- Received by Ch " << id_
                      << " --- rq (" << read_queue_->size() << "/"
                      << read_queue_->max_size() << ")" << endl);
    stat_.num_rd_in_rk_bk.increment(req->dram_addr().rank * num_bank_ + req->dram_addr().bank);
  } else {
    if (CheckWriteBufferHit(req)) {
      // TODO(mjeong): merge the writes
      write_queue_->Insert(req);
    } else {
      write_queue_->Insert(req);
    }

    DBG_LOG(TRACE_MC, *req << " --- Received by Ch " << id_
                      << " --- wq (" << write_queue_->size() << "/"
                      << write_queue_->max_size() << ")" << endl);
    stat_.num_wr_in_rk_bk.increment(req->dram_addr().rank * num_bank_ + req->dram_addr().bank);

    // for writes, we call callback when we receive them
    req->completed_callback();
  }

  stat_.num_req_received.increment();
  stat_.num_req_received_core[req->source_id()]->increment();
  stat_.IncrementSrStat(stat_.num_req_received_in_sr, req->dram_addr());
  // time_to_PH
  DRAMAddr daddr = req->dram_addr();
  if (req->is_read() && timing_logic_->getRowStatusChipMask(&daddr, ROW_HIT, split_))
    stat_.SampleTimeToPH(req, daddr);
}

void MemoryChannel::sendRequestInQ(Req* req) {
  DBG_LOG(TRACE_MC, *req << " --- Received by IngressQ" << endl);
  req->set_time_received(clock::get_dram_clk() + sched_lat_);

  in_queue_->Insert(req);
}

TIME cycle_hook = 731535;
void cycle_hook_f() {
  cout << "cycle hook hit" << endl;
}

void MemoryChannel::cycle() {
  timing_logic_->xCycle();

#if 0
  if (clock::get_dram_clk() == cycle_hook) {
    cycle_hook_f();
  }
#endif

  ReceiveReqs();
  CheckReqsFinished();
  UpdatePriorities();
  ScheduleDRAMCommand();

#ifdef LOG_ENABLE
  if (_dbg_log_purge != 0 && (clock::get_dram_clk() % _dbg_log_purge == 0)) {
    if (_dbg_log_out != &(std::cout)) {
      char buf[255];
      strncpy(buf, _dbg_log_filename, 255);
      size_t eol = strlen(buf);
      buf[eol] = '1';
      buf[eol+1] = '\0';

      remove(buf);  // delete the old period

      ((std::ofstream*)_dbg_log_out)->close();
      rename(_dbg_log_filename, buf);

      _dbg_log_out = new std::ofstream(_dbg_log_filename);
    }
  }
#endif
}

void MemoryChannel::wrapUpStats() {
  for (UINT32 rk = 0; rk < num_rank_; rk++) {
    // if a subrank is precharged, record how long it has been
    for (UINT32 sr = 0; sr < num_chip_; sr++) {
      if (!timing_logic_->isRankOpen(rk, sr)) {
        UINT32 i = rk*num_chip_ + sr;
        TIME precharged_cycles = clock::get_dram_clk() - stat_.last_precharged_in_sr_[i];
        stat_.num_cycles_precharged_in_sr.increment(i, precharged_cycles);
      }
    }
  }

  pdn_control_->wrapUpStats();
}

ostream& MemoryChannel::printStatus(ostream &os) {
  os << "Read Queue" << endl;
  os << *read_queue_ << endl;
  os << "Write Queue" << endl;
  os << *write_queue_ << endl;
  os << "In-flight Read Queue" << endl;
  os << *inflight_queue_ << endl;
  scheduler_->DumpLog(os);

  return os;
}

ostream& MemoryChannel::printChannelStats(ostream &os) {
  TIME dclk = clock::get_dram_clk();
  UINT32 num_channels = memChannels.size();
  UINT32 num_chips = (*memChannels.begin())->num_chip();
  UINT32 num_ABUS = (*memChannels.begin())->num_ABUS();
//  UINT64 total_REQ = 0;  // over channel and subranks
  UINT64 total_cmds = 0;
  UINT64 total_ACT = 0;
  UINT64 total_PRE = 0;
  UINT64 total_RD = 0;
  UINT64 total_RDAP = 0;
  UINT64 total_WR = 0;
  UINT64 total_WRAP = 0;
  UINT64 total_REF = 0;
  UINT64 total_REQ = 0;
  UINT64 total_PUP = 0;
  UINT64 total_PDN = 0;
  ArrayCounterStat total_REQ_ch("Req dist in channels", num_channels, 8);

//  double total_Psch_PRE_PDN = 0;
//  double total_Psch_ACT_PDN = 0;
//  double total_Psch_PRE_STBY = 0;
//  double total_Psch_ACT_STBY = 0;
//  double total_Psch_ACT = 0;
//  double total_Psch_WR = 0;
//  double total_Psch_RD = 0;
//  double total_Psch_DQ = 0;
//  double total_Psch_termW = 0;
//  double total_Psch_termRoth = 0;
//  double total_Psch_termWoth = 0;
//  double total_Psch_total = 0;

  list<MemoryChannel*>::const_iterator it = memChannels.begin();
  for (UINT32 ch = 0 ; it != memChannels.end(); it++, ch++) {
    UINT64 num_cmds = 0;
    MemoryChannel* currChannel = *it;

    UINT64 num_ACT  = currChannel->stat_.num_ACT_issued_in_sr.get_sum();
    UINT64 num_PRE  = currChannel->stat_.num_PRE_issued_in_sr.get_sum();
    UINT64 num_RD   = currChannel->stat_.num_RD_issued_in_sr.get_sum();
    UINT64 num_RDAP = currChannel->stat_.num_RDAP_issued_in_sr.get_sum();
    UINT64 num_WR   = currChannel->stat_.num_WR_issued_in_sr.get_sum();
    UINT64 num_WRAP = currChannel->stat_.num_WRAP_issued_in_sr.get_sum();
    UINT64 num_REF  = currChannel->stat_.num_REF_issued_in_sr.get_sum();
    UINT64 num_PUP  = currChannel->stat_.num_PUP_issued_in_sr.get_sum();
    UINT64 num_PDN  = currChannel->stat_.num_PDN_issued_in_sr.get_sum();

    UINT64 num_rw_cmd = currChannel->stat_.num_rw_cmd_in_sr.get_sum();
    // num_cmds = num_ACT + num_PRE + num_RD + num_RDAP + num_WR + num_WRAP +
    //            num_REF + num_PUP + num_PDN;
    num_cmds = currChannel->stat_.num_cmds_in_ABUS.get_sum();

    os << "--------------------------------------------------" << endl;
    os << "Channel: " << currChannel->id() << endl;
    os << "--------------------------------------------------" << endl;

    os << "ABUS utilization: "
       << num_cmds << " commands / (" << dclk << " cycles * " << num_ABUS << ") -- "
       << 100 * ((double)num_cmds) / ((double)(num_ABUS * dclk)) << "%" << endl;

    os << "DBUS utilization: "
       << num_rw_cmd << " subrank access * " << DRAMTiming::BURST << " burst / "
       << dclk << " cycles / " << num_chips << " subranks -- "
       << 100 * ((double)num_rw_cmd) * DRAMTiming::BURST / ((double)dclk)
          / ((double)num_chips) << "%" << endl;

    os << "Commands per Transaction: "
       << num_cmds << " commands / " << num_rw_cmd << " transactions -- "
       << ((double)num_cmds) / ((double)num_rw_cmd) << " cmds per transaction" << endl;

    // NOTE(mjeong): Page hit rate can mean two different things when more than one
    //               column command is isssued for one memory request.
    //               1. num ACT / Number of memory requests received
    //               2. num_ACT / Number of column commands issued
    os << "Page hit rate: 1 - (" << num_ACT << " ACTs / " << num_rw_cmd << " transactions) -- "
       << 100 * (1- (((double)num_ACT) / ((double)num_rw_cmd))) << "%" << endl;

    total_cmds += num_cmds;
    total_ACT += num_ACT, total_PRE += num_PRE, total_RD += num_RD;
    total_WR += num_WR, total_REF += num_REF;
    total_RDAP += num_RDAP, total_WRAP += num_WRAP;
    total_REQ += num_rw_cmd;
    total_REQ_ch.set(ch, num_rw_cmd);
    total_PUP += num_PUP;
    total_PDN += num_PDN;
  }

//  if (num_channels > 1) {
    using boost::format;
    os << "--------------------------------------------------" << endl;
    os << "total stats over channels" << endl;
    os << "--------------------------------------------------" << endl;
    os << format("ABUS utilization: %d commands / (%d cycles * %d) / %d channels -- %.2f%%")
          % total_cmds % dclk % num_ABUS % num_channels
          % (100.0 * total_cmds / (num_ABUS * dclk) / num_channels)
       << endl;

    os << format("DBUS utilization: %d subrank access * %d burst / %d cycles / %d subranks "
                 "/ %d channels -- %.2f%%")
          % total_REQ % DRAMTiming::BURST % dclk % num_chips % num_channels
          % (100.0 * total_REQ * DRAMTiming::BURST / dclk / num_chips / num_channels)
       << endl;

    os << format("Commands per Transaction: %d commands / %d transactions -- %d ")
          % total_cmds % total_REQ % (((double)total_cmds) / total_REQ)
       << endl;
    os << format("Page hit rate: 1 - (%d ACTs / %d transactions) -- %d%%")
          % total_ACT % total_REQ
          % (100.0 * (1 - ((double)total_ACT) / total_REQ))
       << endl;

    os << total_REQ_ch << endl;

#if 0
os << "Power --------------------------------------------" << endl;
os << setw(15) << left << "mc.Psch_PRE_PDN " << right << setw(15) << total_Psch_PRE_PDN << endl;
os << setw(15) << left << "mc.Psch_ACT_PDN " << right << setw(15) << total_Psch_ACT_PDN << endl;
os << setw(15) << left << "mc.Psch_PRE_STBY" << right << setw(15) << total_Psch_PRE_STBY << endl;
os << setw(15) << left << "mc.Psch_ACT_STBY" << right << setw(15) << total_Psch_ACT_STBY << endl;
os << setw(15) << left << "mc.Psch_ACT     " << right << setw(15) << total_Psch_ACT << endl;
os << setw(15) << left << "mc.Psch_WR      " << right << setw(15) << total_Psch_WR << endl;
os << setw(15) << left << "mc.Psch_RD      " << right << setw(15) << total_Psch_RD << endl;
os << setw(15) << left << "mc.Psch_DQ      " << right << setw(15) << total_Psch_DQ << endl;
os << setw(15) << left << "mc.Psch_termW   " << right << setw(15) << total_Psch_termW << endl;
os << setw(15) << left << "mc.Psch_termRoth" << right << setw(15) << total_Psch_termRoth << endl;
os << setw(15) << left << "mc.Psch_termWoth" << right << setw(15) << total_Psch_termWoth << endl;
os << "--------------------------------------------------" << endl;
os << setw(15) << left << "mc.total_power" << right << setw(15) << total_Psch_total << endl;
  }
#endif
  // TODO(dhyoon): more things to collect
  // 1. Rank to Rank switching overhead -- Profile inside timing model?
  // 2. Read to Write or Write to Read turn around overhead -- Profile inside timing model?
  return os;
}

INT32 cwSize(UINT32 cw_chip, ChipMask mask) {
  int n = 2;
  int i = n * ((int)(cw_chip / n));

  while (n <= (int)mask.num_chip) {
    if (mask.IsSet(i, n)) {
      n *= 2;
      i = n * ((int)(cw_chip / n));
    } else {
      return n/2;
    }
  }
  return n;
}

void MemoryChannel::CheckReqsFinished() {
  InFlightReqBuffer::iterator it = inflight_queue_->begin();

  while (it != inflight_queue_->end()) {
    Req* req = *it;

    if (req->time_completed() == clock::get_dram_clk()) {
      DBG_LOG(TRACE_MC, *req << " --- Finished" << endl);

      if (req->is_read()) {
        req->completed_callback();
        stat_.SampleTimeToFinish(req);
      }

      it = inflight_queue_->Erase(it);
      delete req;

      stat_.num_req_finished.increment();
    } else {
      if (split_cwf_) {
        // assuming full-rank access only
        if (req->is_read() && (req->time_cw_completed() - (BL_/2) + 1) == clock::get_dram_clk()) {
          ASSERT(req->time_completed() == req->time_cw_completed());
          req->critical_word_callback(8);
          DBG_LOG(TRACE_MC, *req << " --- split CW block returned (8B)" << req->cw_block() << endl);
        }
      } else {
        if (req->is_read() && req->time_cw_completed() == clock::get_dram_clk()) {
          INT32 cw_block_size = cwSize(req->dram_addr().critical_word_chip, req->cw_block());
          req->critical_word_callback(cw_block_size * chip_width_ * BL_);
          DBG_LOG(TRACE_MC, *req << " --- CW block returned (" << cw_block_size * chip_width_ * BL_
                            << "B) " << req->cw_block() << endl);
        }
      }

      it++;
    }
  }
}

/**
 * Checks the age of each request and promote if starved.
 */
void MemoryChannel::UpdatePriorities() {
  for (ReqBuffer::iterator it = read_queue_->begin(); it != read_queue_->end(); it++) {
    Req* req = *it;
    ASSERT(clock::get_dram_clk() >= req->time_received());
    UINT32 num_clocks_waited = clock::get_dram_clk() - req->time_received();
    if (req->priority() != kStarved && num_clocks_waited >= age_promo_threshold_) {
      req->RaisePriority();
      req->RaisePriority();
      req->RaisePriority();
      req->RaisePriority();
      DBG_LOG(TRACE_MC, *req << " --- Starvation promoted " << endl);
      stat_.num_req_starved.increment();
    }

    // FIXME(dhyoon): higher priority for larger req size
#if 0
    if (cg_priority_) {
      UINT32 curr_dbus_granularity = 32768;

      // find the smallest subrank_width in the queue
      for (InFlightReqBuffer::iterator it = inflight_queue_->begin();
           it != inflight_queue_->end();
           it++) {
        UINT32 num_chip_accessed = (*it)->dram_addr().num_chip_accessed;
        if (num_chip_accessed < curr_dbus_granularity) {
          curr_dbus_granularity = num_chip_accessed;
        }
      }

      // Increase priority for anyone who's larger than that
      // Hmm.. even at the cost of precharge???
      if (req->priority() == kNormal  // STARVED
          && req->dram_addr().num_chip_accessed > curr_dbus_granularity
          && req->req_status() == REQ_WAIT) {
//        ReqRowStatus row_status = req->row_status();
//        eMemCmd cmd = Cmd_NOP;
//        if (row_status == ROW_HIT) {
//          cmd = Cmd_RD;
//        } else if (row_status == ROW_EMPTY) {
//          cmd = Cmd_ACT;
//        } else if (row_status == ROW_MISS) {
//          cmd = Cmd_PRE;
//        } else {
//          assert(0);
//        }
//        ChipMask chip_issuable;
//        bool can_issue = timing_logic_->checkCommand(cmd, req->dram_addr(),
//                                                     false, &chip_issuable);
//        if (!can_issue && !chip_issuable.Empty()) {
          DBG_LOG(TRACE_MC, *req << " --- large req promoted " << endl);
          req->RaisePriority();
          stat_.num_req_promote_large_size.increment();
//        }
      }
    }
#endif
  }
}

/**
 * Update states for the command being issued.
 * @param[in] daddr The dram address the command is being issued to.
 * @param[in] cmd The dram command being issued.
 * @param[in] cand The request the dram command is being issued for.
 */
void MemoryChannel::IssueDRAMCommand(DRAMAddr daddr, eMemCmd cmd, Req* cand) {
//  // Multi-CAS together
//  UINT32 n = 1;
//  if ((cmd == Cmd_RD || cmd == Cmd_WR)
//      && use_repeat_queue_
//      && cand->time_cw_completed() != TIME_MAX) {
//    ChipMask chip_cmd_left = cand->dram_addr().chip_cmd_count - daddr.chip_cmd_count;
//    while (chip_cmd_left.AndNotEmpty(daddr.chip_cmd_count)) {
//      chip_cmd_left -= daddr.chip_cmd_count;
//      n++;
//    }
//  }
//
  TIME cmd_latency = timing_logic_->xIssueCommand(cmd, daddr);

  if (cmd == Cmd_PDN) {
    DBG_LOG(TRACE_MC, daddr << " --- Issuing --- " << cmd << endl);
    pdn_control_->powerdown(daddr);
  } else if (cmd == Cmd_PUP) {
    DBG_LOG(TRACE_MC, *cand  << " --- Issuing --- " << cmd << endl);
    pdn_control_->powerup(daddr);
  } else {
    DBG_LOG(TRACE_MC, *cand  << " --- Issuing --- " << cmd << " " << daddr.chip_cmd_count << endl);
    ASSERT(!pdn_control_->is_any_sr_powerdown(daddr));

    // update the row hit status of the requests in the queue (ACT/PRE)
    // calculate when will the issued req finish (RD/WR)
    switch (cmd) {
    case Cmd_ACT:
      ASSERT(split_ || cand->row_status() == ROW_EMPTY);
      pdn_control_->activated(daddr);
      cand->ACTissued_ = true;
      break;

    case Cmd_RD:
    case Cmd_RDAP:
    case Cmd_WR:
    case Cmd_WRAP:
      ASSERT(split_ || cand->row_status() == ROW_HIT);
      ASSERT(!cand->dram_addr().chip_cmd_count.Empty());

      if (cw_priority_ && multi_cas_ == 0 && cand->priority() == kLow
          && cand->time_cw_completed() < TIME_MAX) {
        cand->RaisePriority();
      }

      // Critical-word containing RD is being issued.
      if (cand->time_cw_completed() == TIME_MAX
          && daddr.chip_cmd_count[daddr.critical_word_chip]) {
        TIME one_rd_latency = timing_logic_->getLatency(Cmd_RD);
        cand->set_time_cw_completed(clock::get_dram_clk() + one_rd_latency);
        ChipMask norm_mask = daddr.chip_cmd_count;
        norm_mask.Normalize();
        cand->set_cw_block(cand->cw_block() - (cand->dram_addr().chip_cmd_count - norm_mask));

        if (cw_priority_) {
          if (multi_cas_ == 1) {
            cand->LowerPriority();
          } else if (multi_cas_ == 0 && !(highp8_ && cand->dram_addr().num_chip_accessed == 8)) {
            cand->LowerPriority();
            cand->LowerPriority();
          }
        }
      }

      cand->ColumnCmdIssued(daddr.chip_cmd_count);

      // First CAS is being issued for this request.
      if (cand->req_status() == REQ_WAIT) {
        cand->set_time_issued(clock::get_dram_clk());
        cand->set_req_status(REQ_CAS_STARTED);
        cand->set_time_completed(TIME_MAX);
        inflight_queue_->Insert(cand);
      }

//      // For non-cw repeating requests, put it back to the repeat queue.
//      if (use_repeat_queue_
//          && cand->time_cw_completed() != TIME_MAX
//          && cand->dram_addr().chip_cmd_count.AndNotEmpty(daddr.chip_cmd_count)) {
//        UINT32 n = 0;
//        do {
//          n++;
//          cand->ColumnCmdIssued(daddr.chip_cmd_count);
//        } while (cand->dram_addr().chip_cmd_count.AndNotEmpty(daddr.chip_cmd_count));
//
//        DBG_LOG(TRACE_MC, *cand  << " --- Repeat queue inserted --- " << cmd
//                          << " " << daddr.chip_cmd_count << endl);
//        repeat_queue_->Insert(cand, daddr);
//        timing_logic_->xReserveCommand(cmd, daddr);
//      }

      // Last CAS is being issued for this request.
      if (cand->IsColumnCmdAllIssued()) {
        cand->set_time_completed(clock::get_dram_clk() + cmd_latency);
        cand->set_req_status(REQ_LAST_CAS_ISSUED);
        if (cmd == Cmd_RD || cmd == Cmd_RDAP) {
          read_queue_->Erase(cand);
        } else {
          write_queue_->Erase(cand);
          DBG_LOG(TRACE_MC, *cand << " --- Finished --- wq ("
                            << write_queue_->size() << "/"
                            << write_queue_->max_size() << ")" << endl);
        }
      }

      // AP commands need to update the row hit/miss status
      if (cmd == Cmd_RDAP || cmd == Cmd_WRAP) {
        pdn_control_->precharged(daddr);
      }
      break;

    case Cmd_PRE:
    case Cmd_PREA:
      ASSERT(split_ || cand->row_status() == ROW_MISS);
      pdn_control_->precharged(daddr);
      break;

    case Cmd_REF:
      ASSERT(split_ || cand->row_status() == ROW_EMPTY);
      PANIC("Refresh not implemented yet.");
      break;

    case Cmd_NOP:
    default:
      break;
    }
  }
}

void MemoryChannel::ScheduleDRAMCommand() {
  Req* req;
  DRAMAddr daddr;
  eMemCmd cmd;

  bool found = false;
  do {
    found = false;

//    if (use_repeat_queue_)
//      found = FindRepeater(&daddr, &cmd, &req);
    if (!found && use_powerdown_mode_)
      found = pdn_control_->findPUPCandidate(&daddr, &cmd, &req, scheduler_->IsReadMode());
    if (!found)
      found = scheduler_->FindCandidate(&daddr, &cmd, &req);
    if (!found && use_powerdown_mode_)
      found = pdn_control_->findPDNCandidate(&daddr, &cmd);

    if (found) {
      PreIssueStatUpdate(daddr, cmd, req);
      IssueDRAMCommand(daddr, cmd, req);
      PostIssueStatUpdate(daddr, cmd, req);
    }
  } while (found && (timing_logic_->canIssueMoreCommand() > 0));
}

void MemoryChannel::PreIssueStatUpdate(const DRAMAddr& daddr, eMemCmd cmd, Req* req) {
  // if the rank was precharged, record how long it was
  if (cmd == Cmd_ACT) {
    for (UINT32 sr = 0; sr < num_chip_; sr++) {
      if (daddr.chip_cmd_count[sr] && !timing_logic_->isRankOpen(daddr.rank, sr)) {
        UINT32 i = daddr.rank*num_chip_ + sr;
        TIME cycles = clock::get_dram_clk() - stat_.last_precharged_in_sr_[i];
        stat_.num_cycles_precharged_in_sr.increment(i, cycles);
      }
    }
  }

  if (cmd == Cmd_RD || cmd == Cmd_RDAP) {
    stat_.SampleTimeToRD(req, daddr);
  }

  switch (cmd) {
    case Cmd_RD:
    case Cmd_RDAP:
    case Cmd_WR:
    case Cmd_WRAP: {
      break;
    }
    default:
      break;
  }
}

void MemoryChannel::PostIssueStatUpdate(const DRAMAddr& daddr, eMemCmd cmd, Req* req) {
  UINT32 rk = daddr.rank;
  UINT32 bk = daddr.bank;

  switch (cmd) {
  case Cmd_ACT: {
    UpdateRowStatusInQueues(daddr, cmd, req);

    stat_.num_ACT_issued.increment();
    stat_.IncrementSrStat(stat_.num_ACT_issued_in_sr, daddr);
    for (INT32 i = 0; i < daddr.chip_cmd_count.num_chip; i++) {
      if (daddr.chip_cmd_count[i]) {
        stat_.num_ACT_issued_core[req->source_id()]->increment();
        if (req->is_read())
          stat_.num_ACT_RD_issued_core[req->source_id()]->increment();
        else
          stat_.num_ACT_WR_issued_core[req->source_id()]->increment();
      }
    }
    break;
  }

  case Cmd_RD:
  case Cmd_RDAP:
  case Cmd_WR:
  case Cmd_WRAP:
    if (cmd == Cmd_RDAP || cmd == Cmd_WRAP) {
      UpdateRowStatusInQueues(daddr, cmd, req);

      for (UINT32 sr = 0; sr < num_chip_; sr++) {
        if (daddr.chip_cmd_count[sr] && !timing_logic_->isRankOpen(rk, sr)) {
          UINT32 i = rk*num_chip_ + sr;
          stat_.last_precharged_in_sr_[i] = clock::get_dram_clk();
        }
      }
    }

    switch (cmd) {
    case Cmd_RD:
      stat_.num_RD_issued.increment();
      stat_.IncrementSrStat(stat_.num_RD_issued_in_sr, daddr);
      for (INT32 i = 0; i < daddr.chip_cmd_count.num_chip; i++) {
        if (daddr.chip_cmd_count[i])
          stat_.num_RD_issued_core[req->source_id()]->increment();
      }
      break;
    case Cmd_RDAP:
      stat_.num_RDAP_issued.increment();
      stat_.IncrementSrStat(stat_.num_RDAP_issued_in_sr, daddr);
      for (INT32 i = 0; i < daddr.chip_cmd_count.num_chip; i++) {
        if (daddr.chip_cmd_count[i])
          stat_.num_RD_issued_core[req->source_id()]->increment();
      }
      break;
    case Cmd_WR:
      stat_.num_WR_issued.increment();
      stat_.IncrementSrStat(stat_.num_WR_issued_in_sr, daddr);
      for (INT32 i = 0; i < daddr.chip_cmd_count.num_chip; i++) {
        if (daddr.chip_cmd_count[i])
          stat_.num_WR_issued_core[req->source_id()]->increment();
      }
      break;
    case Cmd_WRAP:
      stat_.num_WRAP_issued.increment();
      stat_.IncrementSrStat(stat_.num_WRAP_issued_in_sr, daddr);
      for (INT32 i = 0; i < daddr.chip_cmd_count.num_chip; i++) {
        if (daddr.chip_cmd_count[i])
          stat_.num_WR_issued_core[req->source_id()]->increment();
      }
      break;
    default:
      PANIC("Undefined command: " << cmd);
    }

    stat_.num_req_in_rk_bk.increment(rk*num_bank_ + bk);
    for (UINT32 sr = 0; sr < num_chip_; sr++) {
      if (daddr.chip_cmd_count[sr])
        stat_.num_rw_cmd_in_sr.increment(sr);
    }
    break;

  case Cmd_PRE:
  case Cmd_PREA:
    UpdateRowStatusInQueues(daddr, cmd, req);

    stat_.num_PRE_issued.increment();
    stat_.IncrementSrStat(stat_.num_PRE_issued_in_sr, daddr);
    for (INT32 i = 0; i < daddr.chip_cmd_count.num_chip; i++) {
      if (daddr.chip_cmd_count[i])
        stat_.num_PRE_issued_core[req->source_id()]->increment();
        if (req->is_read())
          stat_.num_PRE_RD_issued_core[req->source_id()]->increment();
        else
          stat_.num_PRE_WR_issued_core[req->source_id()]->increment();
    }

    for (UINT32 sr = 0; sr < num_chip_; sr++) {
      if (daddr.chip_cmd_count[sr] && !timing_logic_->isRankOpen(rk, sr)) {
        UINT32 i = rk*num_chip_ + sr;
        stat_.last_precharged_in_sr_[i] = clock::get_dram_clk();
      }
    }
    break;

  case Cmd_REF:
    stat_.num_REF_issued.increment();
    stat_.IncrementSrStat(stat_.num_REF_issued_in_sr, daddr);
    break;

  case Cmd_PUP:
    stat_.IncrementSrStat(stat_.num_PUP_issued_in_sr, daddr);
    break;

  case Cmd_PDN:
    stat_.IncrementSrStat(stat_.num_PDN_issued_in_sr, daddr);
    break;

  case Cmd_NOP:
  default:
    break;
  }  // switch (cmd)

  if (cmd != Cmd_NOP) {
    for (UINT32 a = 0; a < num_ABUS_; a++) {
      for (UINT32 sr = a*num_chip_/num_ABUS_; sr < (a+1)*num_chip_/num_ABUS_; sr++) {
        if (daddr.chip_cmd_count[sr]) {
          stat_.num_cmds_in_ABUS.increment(a);
          break;
        }
      }
    }
  }
}

void MemoryChannel::UpdateRowStatusInQueues(const DRAMAddr& daddr,
                                            eMemCmd cmd,
                                            Req* req_issued,
                                            ReqBuffer* buffer) {
  for (ReqBuffer::iterator it = buffer->begin(); it != buffer->end(); it++) {
    Req* req = *it;

    if (req->dram_addr().isSameBank(daddr)) {
      if (cmd == Cmd_ACT && req->is_read()) {
        TIME PH_time = clock::get_dram_clk() + timing_logic_->getLatency(Cmd_ACT);
        stat_.SampleTimeToPH(req, req->dram_addr().chip_cmd_count && daddr.chip_cmd_count, PH_time);
      }

      // This per req row-status is purely for a debugging purpose.
      ReqRowStatus prev = req->row_status();
      ReqRowStatus current = timing_logic_->getBankStatus(req->dram_addr());
      req->set_row_status(current);

      if (prev != current)
        CheckRedundantActPre(prev, req, req_issued);
    }
  }
}

void MemoryChannel::UpdateRowStatusInQueues(const DRAMAddr& daddr, eMemCmd cmd, Req* req) {
  UpdateRowStatusInQueues(daddr, cmd, req, read_queue_);
  UpdateRowStatusInQueues(daddr, cmd, req, write_queue_);
}

bool MemoryChannel::CheckWriteBufferHit(Req* new_req) {
  ReqBuffer::iterator it = write_queue_->begin();
  for (; it != write_queue_->end(); it++) {
    Req* write_req = *it;

    if ((write_req->phy_addr() <= new_req->phy_addr())
        && (write_req->phy_addr() + write_req->size()
            >= new_req->phy_addr() + new_req->size())) {
      // Hit
      DBG_LOG(TRACE_MC, new_req->id() << (new_req->is_read() ?
              " RAW hit on " : " WAW hit on ") << write_req->id() << endl);
    }
    return true;
  }

  return false;
}

bool MemoryChannel::isReqPending() {
  return (read_queue_->size() != 0 ||
          write_queue_->size() != 0 ||
          inflight_queue_->size() != 0);
}

void MemoryChannel::CheckRedundantActPre(ReqRowStatus prev_status, Req* req, Req* act_pre_issuer) {
  ReqRowStatus curr_status = req->row_status();

  // TODO(mjeong): check this in the scheduler, not here?
  if (prev_status == ROW_HIT
      && curr_status == ROW_EMPTY
      && req->req_status() != REQ_LAST_CAS_ISSUED
      && req->ACTissued_
      && req->priority() >= act_pre_issuer->priority()
      && !(!scheduler_->IsReadMode() && req->is_read() && !act_pre_issuer->is_read())) {
    // Req row status became from HIT to MISS
    // ACT was issued for this request, but no CAS was issued
    cout << "[" << std::setw(8) << std::setfill('0') << clock::get_dram_clk();
    cout << "] WARNING: redundant ACT/PRE " << endl;
    cout << "The following req became ROW_EMPTY status, but previously ";
    cout << "we issued ACT but no CAS!" << endl;
    cout << *req << endl << endl;
    stat_.num_Redundant_ACT_PRE.increment();
    printStatus(cout);
  }
}

#if 0
bool MemoryChannel::FindRepeater(DRAMAddr* daddr, eMemCmd* cmd, Req** req) {
  for (UINT32 i = 0; i < repeat_queue_->size(); i++) {
    *req = (*repeat_queue_)[i];
    *daddr = repeat_queue_->Daddr(i);
    *cmd = (*req)->is_read() ? Cmd_RD : Cmd_WR;

    if (timing_logic_->checkCommand(*cmd, *daddr)) {
      ASSERT(timing_logic_->getRowStatusChipMask(daddr, ROW_HIT, false));
      DBG_LOG(TRACE_SCHEDULER, **req << " - repeat issue cand found - " << *cmd << endl);
      repeat_queue_->Erase(i);
      return true;
    } else {
      DBG_LOG(TRACE_SCHEDULER_V, **req << " testing CAS candadicy : 0" << endl);
    }
  }

  return false;
}
#endif

MemoryChannel::MCStat::MCStat(MemoryChannel* ch)
    :  ch_(ch),
       num_req_received(fmt_str_ch("mc_num_req_received", ch->id())),
       num_ACT_issued(fmt_str_ch("mc_num_ACT_issued", ch->id())),
       num_PRE_issued(fmt_str_ch("mc_num_PRE_issued", ch->id())),
       num_RD_issued(fmt_str_ch("mc_num_RD_issued", ch->id())),
       num_RDAP_issued(fmt_str_ch("mc_num_RDAP_issued", ch->id())),
       num_WR_issued(fmt_str_ch("mc_num_WR_issued", ch->id())),
       num_WRAP_issued(fmt_str_ch("mc_num_WRAP_issued", ch->id())),
       num_REF_issued(fmt_str_ch("mc_num_REF_issued", ch->id())),
       num_req_finished(fmt_str_ch("mc_num_req_finished", ch->id())),
       num_req_starved(fmt_str_ch("mc_num_req_starved", ch->id())),
       num_req_promote_large_size(fmt_str_ch("mc_num_req_promted_for_size", ch->id())),
       num_cmds_in_ABUS(fmt_str_ch("mc_cmds_in_ABUS", ch->id()),
         ch->num_ABUS(), ch->num_ABUS(), false),
       num_req_received_in_sr(fmt_str_ch("mc_num_req_in_sr", ch->id(), "rk\\sr"),
         ch->num_rank() * ch->num_chip(), ch->num_chip(), false),
       num_ACT_issued_in_sr(fmt_str_ch("mc_num_ACT_issued_in_sr", ch->id()),
         ch->num_rank() * ch->num_chip(), ch->num_chip(), false),
       num_PRE_issued_in_sr(fmt_str_ch("mc_num_PRE_issued_in_sr", ch->id()),
         ch->num_rank() * ch->num_chip(), ch->num_chip(), false),
       num_RD_issued_in_sr(fmt_str_ch("mc_num_RD_issued_in_sr", ch->id()),
         ch->num_rank() * ch->num_chip(), ch->num_chip(), false),
       num_RDAP_issued_in_sr(fmt_str_ch("mc_num_RDAP_issued_in_sr", ch->id()),
         ch->num_rank() * ch->num_chip(), ch->num_chip(), false),
       num_WR_issued_in_sr(fmt_str_ch("mc_num_WR_issued_in_sr", ch->id()),
         ch->num_rank() * ch->num_chip(), ch->num_chip(), false),
       num_WRAP_issued_in_sr(fmt_str_ch("mc_num_WRAP_issued_in_sr", ch->id()),
         ch->num_rank() * ch->num_chip(), ch->num_chip(), false),
       num_REF_issued_in_sr(fmt_str_ch("mc_num_REF_issued_in_sr", ch->id()),
         ch->num_rank() * ch->num_chip(), ch->num_chip(), false),
       num_PUP_issued_in_sr(fmt_str_ch("mc_num_PUP_issued_in_sr", ch->id()),
         ch->num_rank() * ch->num_chip(), ch->num_chip(), false),
       num_PDN_issued_in_sr(fmt_str_ch("mc_num_PDN_issued_in_sr", ch->id()),
         ch->num_rank() * ch->num_chip(), ch->num_chip(), false),
       num_cycles_precharged_in_sr(fmt_str_ch("mc_num_cycles_precharged_in_sr", ch->id()),
         ch->num_rank() * ch->num_chip(), ch->num_chip(), false),
       num_rw_cmd_in_sr(fmt_str_ch("mc_num_rw_cmd_in_sr", ch->id()),
         ch->num_chip(), ch->num_chip()),
       num_req_in_rk_bk(fmt_str_ch("mc_req_distribution", ch->id(), "rk\\bk"),
         ch->num_rank() * ch->num_bank(), ch->num_bank()),
       num_rd_in_rk_bk(fmt_str_ch("mc_rd_req_distribution", ch->id(), "rk\\bk"),
         ch->num_rank() * ch->num_bank(), ch->num_bank()),
       num_wr_in_rk_bk(fmt_str_ch("mc_wr_req_distribution", ch->id(), "rk\\bk"),
         ch->num_rank() * ch->num_bank(), ch->num_bank()),
       num_Redundant_ACT_PRE(fmt_str_ch("Redundant_ACT_PRE", ch->id())) {
  // stats that need dynamic allocation
  for (INT32 i = 0; i < MAX_SOURCE_ID; i++) {
    num_req_received_core[i] = new CounterStat(fmt_str_ch("num_req_received_core", ch->id(), i));
    num_req_received_core[i]->set_print_periodic(true);

    num_ACT_issued_core[i] = new CounterStat(fmt_str_ch("num_ACT_core", ch->id(), i));
    num_RD_issued_core[i] = new CounterStat(fmt_str_ch("num_RD_core", ch->id(), i));
    num_RD_issued_core[i]->set_print_periodic(true);
    num_WR_issued_core[i] = new CounterStat(fmt_str_ch("num_WR_core", ch->id(), i));
    num_WR_issued_core[i]->set_print_periodic(true);
    num_PRE_issued_core[i] = new CounterStat(fmt_str_ch("num_PRE_core", ch->id(), i));

    num_ACT_RD_issued_core[i] = new CounterStat(fmt_str_ch("num_ACT_RD_core", ch->id(), i));
    num_PRE_RD_issued_core[i] = new CounterStat(fmt_str_ch("num_PRE_RD_core", ch->id(), i));
    num_ACT_WR_issued_core[i] = new CounterStat(fmt_str_ch("num_ACT_WR_core", ch->id(), i));
    num_PRE_WR_issued_core[i] = new CounterStat(fmt_str_ch("num_PRE_WR_core", ch->id(), i));
  }

  last_precharged_in_sr_ = new TIME[ch->num_rank()*ch->num_chip()];
  for (UINT32 i = 0; i < ch->num_rank()*ch->num_chip(); i++) {
    last_precharged_in_sr_[i] = 0;
  }

  // +1 for the total. x2 for cw/rest
  // core0: [0] cw, [5] rest
  // core1: [1] cw, [6] rest
  // total: [4] cw, [9] rest
  char buf[256];
  for (UINT32 i = 0; i < MAX_SOURCE_ID; i++) {
    snprintf(buf, sizeof(buf), "ch%d.c%d.mc_dist_read_latency", ch->id(), i);
    dist_read_latency[i] = new SSHistogram(buf, "read latency distribution per core");
    dist_read_latency[i]->set_print_periodic(true);
  }
  snprintf(buf, sizeof(buf), "ch%d.mc_dist_read_latency", ch->id());
  dist_read_latency[MAX_SOURCE_ID] = new SSHistogram(buf, "read latency distribution");
  dist_read_latency[MAX_SOURCE_ID]->set_print_periodic(true);

  for (UINT32 i = 0; i < MAX_SOURCE_ID; i++) {
    snprintf(buf, sizeof(buf), "ch%d.c%d.mc_dist_time_to_PH_cw", ch->id(), i);
    dist_time_to_PH[i] = new SSHistogram(buf, "time to page_hit critical word per core");
    dist_time_to_PH[i]->set_print_periodic(true);
    snprintf(buf, sizeof(buf), "ch%d.c%d.mc_dist_time_to_PH_noncw", ch->id(), i);
    dist_time_to_PH[i + MAX_SOURCE_ID + 1] =
      new SSHistogram(buf, "time to page_hit non critical word per core");
    dist_time_to_PH[i + MAX_SOURCE_ID + 1]->set_print_periodic(true);
  }
  snprintf(buf, sizeof(buf), "ch%d.mc_dist_time_to_PH_cw", ch->id());
  dist_time_to_PH[MAX_SOURCE_ID] = new SSHistogram(buf, "time to page_hit critical word");
  dist_time_to_PH[MAX_SOURCE_ID]->set_print_periodic(true);
  snprintf(buf, sizeof(buf), "ch%d.mc_dist_time_to_PH_noncw", ch->id());
  dist_time_to_PH[2*MAX_SOURCE_ID + 1] = new SSHistogram(buf, "time to page_hit non critical word");

  for (UINT32 i = 0; i < MAX_SOURCE_ID; i++) {
    snprintf(buf, sizeof(buf), "ch%d.c%d.mc_dist_time_to_RD_cw", ch->id(), i);
    dist_time_to_RD[i] = new SSHistogram(buf, "time to page_hit critical word per core");
    snprintf(buf, sizeof(buf), "ch%d.c%d.mc_dist_time_to_RD_noncw", ch->id(), i);
    dist_time_to_RD[i + MAX_SOURCE_ID + 1] =
      new SSHistogram(buf, "time to page_hit non critical word per core");
  }
  snprintf(buf, sizeof(buf), "ch%d.mc_dist_time_to_RD_cw", ch->id());
  dist_time_to_RD[MAX_SOURCE_ID] = new SSHistogram(buf, "time to page_hit critical word");
  snprintf(buf, sizeof(buf), "ch%d.mc_dist_time_to_RD_noncw", ch->id());
  dist_time_to_RD[2*MAX_SOURCE_ID + 1] = new SSHistogram(buf, "time to page_hit non critical word");
};

void MemoryChannel::MCStat::SampleTimeToPH(Req* req, const DRAMAddr& daddr) {
  for (INT32 i = 0; i < daddr.chip_cmd_count.num_chip; i++) {
    if (daddr.chip_cmd_count[i]) {
      TIME last_act_time = ch_->timing_logic_->getLastActTime(daddr.rank, daddr.bank, i);
      TIME PH_time = std::max(last_act_time + ch_->timing_logic_->getLatency(Cmd_ACT),
                              clock::get_dram_clk());
//      TIME lat = PH_time - req->time_received();

//      UINT32 cw = (req->time_cw_completed() == TIME_MAX
//                   && i == req->dram_addr().critical_word_chip ? 0 : MAX_SOURCE_ID + 1);
//      stat_add_sample(dist_time_to_PH[cw + MAX_SOURCE_ID], lat);
//      stat_add_sample(dist_time_to_PH[cw + req->source_id()], lat);
//
//      DBG_LOG(TRACE_STAT, *req << " --- time_to_PH: " << (cw == 0 ? "cw: " : "noncw: ")
//                          << lat << endl);
//
      ASSERT(req->time_page_hit_chip_[i] == TIME_MAX);
      req->time_page_hit_chip_[i] = PH_time;
    }
  }
}

void MemoryChannel::MCStat::SampleTimeToPH(Req* req, const ChipMask& mask, TIME PH_time) {
//  TIME lat = PH_time - req->time_received();
//  ASSERT(lat >= 0);
  for (INT32 i = 0; i < mask.num_chip; i++) {
    if (mask[i]) {
//      UINT32 cw = (req->time_cw_completed() == TIME_MAX
//                   && i == req->dram_addr().critical_word_chip ? 0 : MAX_SOURCE_ID + 1);
//      stat_add_sample(dist_time_to_PH[cw + MAX_SOURCE_ID], lat);
//      stat_add_sample(dist_time_to_PH[cw + req->source_id()], lat);
//
//      DBG_LOG(TRACE_STAT, *req << " --- time_to_PH: "
//                               << (cw == 0 ? "cw: " : "noncw: ") << lat << endl);
//
      req->time_page_hit_chip_[i] = PH_time;
    }
  }
}

void MemoryChannel::MCStat::SampleTimeToRD(Req* req, const DRAMAddr& daddr) {
  ChipMask chip_cmd_count = daddr.chip_cmd_count;
  TIME repeat_interval = ch_->timing_logic_->GetRepeatInterval(Cmd_RD);
  ASSERT(repeat_interval > 0);

  for (INT32 i = 0; i < daddr.chip_cmd_count.num_chip; i++) {
    if (chip_cmd_count[i]) {
      TIME to_ph = req->time_page_hit_chip_[i] - req->time_received();
      TIME ph_to_rd = clock::get_dram_clk() - req->time_page_hit_chip_[i];
      ASSERT(to_ph >= 0);
      ASSERT(ph_to_rd >= 0);

      UINT32 cw = (req->time_cw_completed() == TIME_MAX
                   && i == req->dram_addr().critical_word_chip ? 0 : MAX_SOURCE_ID + 1);

      for (UINT32 j = 0; j < chip_cmd_count[i]; j++) {
//        stat_add_sample(dist_time_to_PH[cw + MAX_SOURCE_ID], to_ph);
//        stat_add_sample(dist_time_to_PH[cw + req->source_id()], to_ph);
//        stat_add_sample(dist_time_to_RD[cw + MAX_SOURCE_ID], ph_to_rd);
//        stat_add_sample(dist_time_to_RD[cw + req->source_id()], ph_to_rd);
        dist_time_to_PH[cw + MAX_SOURCE_ID]->add_sample(to_ph);
        dist_time_to_PH[cw + req->source_id()]->add_sample(to_ph);
        dist_time_to_RD[cw + MAX_SOURCE_ID]->add_sample(ph_to_rd);
        dist_time_to_RD[cw + req->source_id()]->add_sample(ph_to_rd);
        cw = MAX_SOURCE_ID + 1;
        ph_to_rd += repeat_interval;
      }
    }
  }
}

void MemoryChannel::MCStat::SampleTimeToFinish(Req* req) {
  if (req->time_received() != TIME_MAX) {
    TIME arrived_to_fin = clock::get_dram_clk() - req->time_received();
    if (arrived_to_fin > 0) {
//      stat_add_sample(dist_read_latency[MAX_SOURCE_ID], arrived_to_fin);
//      stat_add_sample(dist_read_latency[req->source_id()], arrived_to_fin);
      dist_read_latency[MAX_SOURCE_ID]->add_sample(arrived_to_fin);
      dist_read_latency[req->source_id()]->add_sample(arrived_to_fin);
      DBG_LOG(TRACE_MC, *req << " --- Latency: " << arrived_to_fin << endl);
    }
  }
}

void MemoryChannel::MCStat::IncrementSrStat(ArrayCounterStat& stat, const DRAMAddr& daddr) {
  for (INT32 sr = 0; sr < ChipMask::num_chip; sr++) {
    if (daddr.chip_cmd_count[sr]) {
      UINT32 i = daddr.rank*ChipMask::num_chip + sr;
      stat.increment(i);
    }
  }
}

void MemoryChannel::ReceiveReqs() {
  ReqBuffer::iterator it = in_queue_->begin();

  while (it != in_queue_->end()) {
    Req* req = *it;
    bool full = (req->is_read() && isReadQueueFull())
                || (!req->is_read() && isWriteQueueFull());

    if (!full && (req->time_received() <= clock::get_dram_clk())) {
      it = in_queue_->Erase(it);
      sendRequest(req);
      continue;
    }

    it++;
  }
}

}  // namespace drsim
