/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#include "clock.h"
#include "pdncontroller.h"
#include "log.h"

using std::cout;
using std::endl;
using std::cerr;

namespace drsim {

PDNController::PDNController(UINT32 channel_id,
                             UINT32 num_rank,
                             UINT32 num_chip,
                             ReqBuffer* read_queue,
                             ReqBuffer* write_queue,
                             InFlightReqBuffer* inflight_read_queue,
                             cTimingModel* timing_logic)
    :  channel_id_(channel_id),
       num_rank_(num_rank),
       num_chip_(num_chip),
       read_queue_(read_queue),
       write_queue_(write_queue),
       inflight_read_queue_(inflight_read_queue),
       timing_logic_(timing_logic),
       num_cycles_pre_pdn_in_sr_(fmt_str_ch("mc_num_cycles_pre_pdn_in_sr", channel_id),
                                  num_rank * num_chip, num_chip, false),
       num_cycles_act_pdn_in_sr_(fmt_str_ch("mc_num_cycles_act_pdn_in_sr", channel_id),
                                  num_rank * num_chip, num_chip, false) {
  chip_pdn_state_ = new PowerState*[num_rank_];
  last_pre_pdn_in_sr_ = new TIME*[num_rank_];
  last_act_pdn_in_sr_ = new TIME*[num_rank_];
  for (UINT32 rk = 0; rk < num_rank_; rk++) {
    chip_pdn_state_[rk] = new PowerState[num_chip_];
    last_pre_pdn_in_sr_[rk] = new TIME[num_chip_];
    last_act_pdn_in_sr_[rk] = new TIME[num_chip_];
    for (UINT32 sr = 0; sr < num_chip_; sr++) {
      chip_pdn_state_[rk][sr] = PRE;
      last_pre_pdn_in_sr_[rk][sr] = -1;
      last_act_pdn_in_sr_[rk][sr] = -1;
    }
  }
}

PDNController::~PDNController() {
  for (UINT32 rk = 0; rk < num_rank_; rk++) {
    delete chip_pdn_state_[rk];
    delete last_pre_pdn_in_sr_[rk];
    delete last_act_pdn_in_sr_[rk];
  }
  delete chip_pdn_state_;
  delete last_pre_pdn_in_sr_;
  delete last_act_pdn_in_sr_;
}

bool PDNController::is_any_sr_powerdown(const DRAMAddr& daddr) {
  bool powerdown = false;

  for (UINT32 i = 0; i < num_chip_; i++) {
    if (daddr.chip_cmd_count[i]) {
      powerdown = powerdown || isPowerDown(chip_pdn_state_[daddr.rank][i]);
    }
  }

  return powerdown;
}

bool PDNController::is_any_sr_powerdown(const DRAMAddr& daddr, ChipMask* pdn_chips) {
  bool powerdown = false;

  pdn_chips->SetAll(0);
  for (UINT32 i = 0; i < num_chip_; i++) {
    if (daddr.chip_cmd_count[i] && isPowerDown(chip_pdn_state_[daddr.rank][i])) {
        powerdown = true;
        (*pdn_chips)[i] = 1;
    }
  }

  return powerdown;
}

bool PDNController::findPDNCandidate(DRAMAddr* daddr, eMemCmd* dram_cmd) {
  bool found = false;
  ChipMask chip_issuable;
  ChipMask readq_idle_sr;
  ChipMask writeq_idle_sr;
  ChipMask ifreadq_idle_sr;

  for (UINT32 rk = 0; rk < num_rank_; rk++) {
    daddr->reset();
    daddr->rank = rk;

//    DBG_LOG(TRACE_PDN, *daddr << " PDN checking rank " << rk << endl);

    // create a mask of subranks we can send PDN this cycle
    for (UINT32 sr = 0; sr < num_chip_; sr++) {
      PowerState pdn_state = chip_pdn_state_[rk][sr];
      daddr->chip_cmd_count[sr] = (pdn_state == PRE || pdn_state == ACTIVE);
    }
    found = timing_logic_->checkCommand(Cmd_PDN, *daddr, true, &chip_issuable);

//    DBG_LOG(TRACE_PDN_V, *daddr << " PDN issuable subranks" << endl);

    // create a mask of subranks to which there are no pending request
    if (found
        && (read_queue_->GetIdleChipMask(rk, &readq_idle_sr) > 0)
        && (write_queue_->GetIdleChipMask(rk, &writeq_idle_sr) > 0)
        && (inflight_read_queue_->GetIdleChipMask(rk, &ifreadq_idle_sr) > 0)) {
      daddr->chip_cmd_count = (chip_issuable && readq_idle_sr && ifreadq_idle_sr && writeq_idle_sr);
      found = !(daddr->chip_cmd_count.Empty());
    } else {
      found = false;
    }

    if (found) {
      *dram_cmd = Cmd_PDN;
      DBG_LOG(TRACE_PDN, *daddr << " PDN candidate found" << endl);
      break;
    }
  }

  return found;
}

bool PDNController::findPUPCandidate(DRAMAddr* daddr,
                                     eMemCmd* dram_cmd,
                                     Req** cand,
                                     bool readonly) {
  ReqBuffer::iterator it;

  for (it = read_queue_->begin(); it != read_queue_->end(); it++) {
    Req* req = *it;
    *daddr = req->dram_addr();

    if (req->subrank_powerdown()
        && is_any_sr_powerdown(req->dram_addr(), &daddr->chip_cmd_count)
        && timing_logic_->checkCommand(Cmd_PUP, *daddr, true, &daddr->chip_cmd_count)) {
      *dram_cmd = Cmd_PUP;
      *cand = req;
      return true;
    }
  }

  if (!readonly) {
    for (it = write_queue_->begin(); it != write_queue_->end(); it++) {
      Req* req = *it;
      *daddr = req->dram_addr();

      if (req->subrank_powerdown()
          && is_any_sr_powerdown(req->dram_addr(), &daddr->chip_cmd_count)
          && timing_logic_->checkCommand(Cmd_PUP, *daddr, true, &daddr->chip_cmd_count)) {
        *dram_cmd = Cmd_PUP;
        *cand = req;
        return true;
      }
    }
  }

  return false;
}

void PDNController::activated(const DRAMAddr& daddr) {
  for (UINT32 i = 0; i < num_chip_; i++) {
    if (daddr.chip_cmd_count[i]) {
      ASSERT(chip_pdn_state_[daddr.rank][i] == PRE
             || chip_pdn_state_[daddr.rank][i] == ACTIVE);

      chip_pdn_state_[daddr.rank][i] = ACTIVE;
    }
  }
}

void PDNController::precharged(const DRAMAddr& daddr) {
  for (UINT32 sr = 0; sr < num_chip_; sr++) {
    if (daddr.chip_cmd_count[sr]) {
      if (chip_pdn_state_[daddr.rank][sr] == ACTIVE
          && !timing_logic_->isRankOpen(daddr.rank, sr)) {
        chip_pdn_state_[daddr.rank][sr] = PRE;
      }
    }
  }
}

void PDNController::powerdown(const DRAMAddr& daddr) {
  for (UINT32 i = 0; i < num_chip_; i++) {
    if (daddr.chip_cmd_count[i]) {
      if (chip_pdn_state_[daddr.rank][i] == ACTIVE) {
        DBG_LOG(TRACE_PDN, daddr << " sr " << i << " enters ACTIVE_PDN" << endl);

        chip_pdn_state_[daddr.rank][i] = ACTIVE_PDN;
        // stat
        last_act_pdn_in_sr_[daddr.rank][i] = clock::get_dram_clk();
      } else if (chip_pdn_state_[daddr.rank][i] == PRE) {
        DBG_LOG(TRACE_PDN, daddr << " sr " << i << " enters PRE_PDN" << endl);

        chip_pdn_state_[daddr.rank][i] = PRE_PDN;
        // stat
        last_pre_pdn_in_sr_[daddr.rank][i] = clock::get_dram_clk();
      } else {
        PANIC("Trying to power down an already power-down subrank: "
             << daddr << " subrank " << i);
      }
    }
  }

  updatePowerStatusInQueues();
  DBG_LOG_ASIS(TRACE_PDN, printPDNState(*_dbg_log_out));
}

void PDNController::powerup(const DRAMAddr& daddr) {
  for (UINT32 i = 0; i < num_chip_; i++) {
    if (daddr.chip_cmd_count[i]) {
      if (chip_pdn_state_[daddr.rank][i] == ACTIVE_PDN) {
        DBG_LOG(TRACE_PDN, daddr << " sr " << i << " exits ACTIVE_PDN" << endl);

        chip_pdn_state_[daddr.rank][i] = ACTIVE;
        // stat
        ASSERT(last_act_pdn_in_sr_[daddr.rank][i] > 0);
        TIME pdn_cycles = clock::get_dram_clk() - last_act_pdn_in_sr_[daddr.rank][i];
        ASSERT(pdn_cycles > 0);
        num_cycles_act_pdn_in_sr_.increment(daddr.rank*num_chip_ + i, pdn_cycles);
      } else if (chip_pdn_state_[daddr.rank][i] == PRE_PDN) {
        DBG_LOG(TRACE_PDN, daddr << " sr " << i << " exits PRE_PDN" << endl);

        chip_pdn_state_[daddr.rank][i] = PRE;
        // stat
        ASSERT(last_pre_pdn_in_sr_[daddr.rank][i] > 0);
        TIME pdn_cycles = clock::get_dram_clk() - last_pre_pdn_in_sr_[daddr.rank][i];
        ASSERT(pdn_cycles > 0);
        num_cycles_pre_pdn_in_sr_.increment(daddr.rank*num_chip_ + i, pdn_cycles);
      } else {
        // TODO(mjeong):
      }
    }
  }

  updatePowerStatusInQueues();
  DBG_LOG_ASIS(TRACE_PDN, printPDNState(*_dbg_log_out));
}

void PDNController::updatePowerStatusInQueues() {
  ReqBuffer::iterator it;
  for (it = read_queue_->begin(); it != read_queue_->end(); it++) {
    Req* it_req = *it;
    DRAMAddr it_dram_addr = it_req->dram_addr();

    it_req->set_subrank_powerdown(is_any_sr_powerdown(it_dram_addr));
  }
  for (it = write_queue_->begin(); it != write_queue_->end(); it++) {
    Req* it_req = *it;
    DRAMAddr it_dram_addr = it_req->dram_addr();

    it_req->set_subrank_powerdown(is_any_sr_powerdown(it_dram_addr));
  }
}

void PDNController::wrapUpStats() {
  for (UINT32 rk = 0; rk < num_rank_; rk++) {
    // if a subrank is power-down, record how long it has been
    for (UINT32 sr = 0; sr < num_chip_; sr++) {
      if (chip_pdn_state_[rk][sr] == ACTIVE_PDN) {
        ASSERT(last_act_pdn_in_sr_[rk][sr] > 0);
        TIME pdn_cycles = clock::get_dram_clk() - last_act_pdn_in_sr_[rk][sr];
        ASSERT(pdn_cycles >= 0);
        num_cycles_act_pdn_in_sr_.increment(rk*num_chip_ + sr, pdn_cycles);
      }
      else if (chip_pdn_state_[rk][sr] == PRE_PDN) {
        ASSERT(last_pre_pdn_in_sr_[rk][sr] > 0);
        TIME pdn_cycles = clock::get_dram_clk() - last_pre_pdn_in_sr_[rk][sr];
        ASSERT(pdn_cycles >= 0);
        num_cycles_pre_pdn_in_sr_.increment(rk*num_chip_ + sr, pdn_cycles);
      }
    }
  }
}

ostream& PDNController::printPDNState(ostream &os) {
  for (UINT32 rk = 0; rk < num_rank_; rk++) {
    for (UINT32 sr = 0; sr < num_chip_; sr++) {
      os << std::setw(13) << std::setfill(' ') << chip_pdn_state_[rk][sr];
    }
    os << endl;
  }
  return os;
}

UINT64 PDNController::get_num_cycles_act_pdn_in_sr(UINT32 rk, UINT32 sr) {
  return num_cycles_act_pdn_in_sr_.get_counter(rk*num_chip_ + sr);
}

UINT64 PDNController::get_num_cycles_pre_pdn_in_sr(UINT32 rk, UINT32 sr) {
  return num_cycles_pre_pdn_in_sr_.get_counter(rk*num_chip_ + sr);
}

}  // namespace drsim
