/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#include <cstdlib>
#include <list>
#include "reqbuffer.h"

namespace drsim {

ostream& operator<<(ostream& os, const ReqRowStatus status) {
  switch (status) {
    case ROW_HIT:    os << "HIT  ";    break;
    case ROW_MISS:   os << "MISS ";    break;
    case ROW_EMPTY:  os << "EMPTY";    break;
    case ROW_MIX:    os << "MIX  ";    break;
    default:         os << "INVALID";  break;
  }
  return os;
}

ostream& operator<<(ostream& os, const ReqStatus status) {
  switch (status) {
    case REQ_WAIT:            os << "WAIT   ";  break;
    case REQ_CAS_STARTED:     os << "STARTED";  break;
    case REQ_LAST_CAS_ISSUED: os << "LAST   ";  break;
    default:                  os << "INVALID";  break;
  }
  return os;
}

Req::Req(UINT64 id,
         bool is_read,
         ADDRESS phy_addr,
         UINT32 size,
         DRAMAddr dram_addr,
         bool is_prefetch,
         INT32 source_id)
    : id_(id), is_read_(is_read), phy_addr_(phy_addr), size_(size), dram_addr_(dram_addr),
      priority_(kNormal), subrank_powerdown_(false),
      cw_block_(dram_addr.chip_cmd_count), source_id_(source_id),
      is_prefetch_(is_prefetch), row_status_(ROW_EMPTY), req_status_(REQ_WAIT),
      time_received_(TIME_MAX), time_page_hit_(TIME_MAX),
      time_issued_(TIME_MAX), time_cw_completed_(TIME_MAX), time_completed_(TIME_MAX),
      is_closed_page_(false), allow_rmode_write_ACTPRE_(false) {
  ACTissued_ = false;
  for (int i = 0; i < 8; i++)
    time_page_hit_chip_[i] = TIME_MAX;
}

ostream & operator<<(ostream & os, const Req & req) {
  os << "Req " << std::setw(3) << std::setfill('0') << req.id_;
  if (req.is_prefetch_) {
    os << " PF";
  } else if (req.is_read_) {
    os << " RD";
  } else {
    os << " WR";
  }
  os << req.dram_addr_;
  os << " 0x" << std::hex << std::setw(8) << req.phy_addr_ << std::dec;
//  os << " " << std::hex << (req.phy_addr_ >> 13) << std::dec;

  os << " " << req.row_status_;
  os << " " << req.req_status_;
  os << " " << req.subrank_powerdown_;
  os << " P" << req.priority_;
  req.PrintExtra(os);

  return os;
}

ostream & operator<<(ostream &os, ReqBuffer &rb) {
  UINT32 i = 0;
  for (ReqBuffer::iterator it = rb.begin(); it != rb.end(); it++, i++)
    os << std::setw(2) << i << " " << **it << std::endl;
  return os;
}

/**
 * Inserts a request into the buffer. We're using STL push_back(), thus the
 * requests in the buffer are sorted by their arrival time. If the scheduler
 * searches the buffer from the front, effectively oldest-first policy is used.
 * In real implmentation, this is not the case and additional logic is
 * required to schedule oldest-first.
 *
 * @param[in] req The request to insert.
 */
void ReqBuffer::Insert(Req* req) {
  ASSERT(entries_.size() < max_size_);
  entries_.push_back(req);
}

void ReqBuffer::Erase(Req* req) {
  for (iterator it = entries_.begin(); it != entries_.end(); it++) {
    if (*it == req) {
      entries_.erase(it);
      return;
    }
  }

  PANIC("Erase item not found." << req);
}

ReqBuffer::iterator ReqBuffer::Erase(iterator it) {
  return entries_.erase(it);
}

UINT32 ReqBuffer::NumReqsToRow(const DRAMAddr& addr) {
  UINT32 count = 0;
  iterator it;

  for (it = begin(); it != end(); it++) {
    if ((*it)->dram_addr().isSameRow(addr))
      count++;
  }

  return count;
}

UINT32 ReqBuffer::GetIdleChipMask(UINT32 rank, ChipMask* chip_idle) {
  chip_idle->SetAll(0);

  // Get 'busy' chip mask.
  for (iterator it = begin(); it != end(); it++) {
    if ((*it)->dram_addr().rank == rank)
      *chip_idle = *chip_idle || (*it)->dram_addr().chip_cmd_count;
  }

  // And inverse it.
  UINT32 count = 0;
  for (INT32 i = 0; i < ChipMask::num_chip; i++) {
    (*chip_idle)[i] = !(*chip_idle)[i];
    if ((*chip_idle)[i])
      count++;
  }

  return count;
}

InFlightReqBuffer::InFlightReqBuffer(UINT32 num_chip)
    : ReqBuffer(0), num_chip_(num_chip) {
//  dq_busy_until_ = new TIME[num_chip];
//  for (UINT32 i = 0; i < num_chip; i++) {
//    dq_busy_until_[i] = 0;
//  }
}

void InFlightReqBuffer::Insert(Req* req) {
  entries_.push_back(req);
//    dq_busy_until_[req->get_dram_addr().chip] = current_cycle+tCAS+BL;
}

#if 0
bool InFlightReqBuffer::isDqFree(UINT32 chip, TIME from, TIME until) const {
//  return (dq_busy_until_[chip] < until);
}
#endif

void RepeatReqBuffer::Insert(Req* req, DRAMAddr daddr) {
  if (entries_.size() < max_size_) {
    entries_.push_back(req);
    daddrs_.push_back(daddr);
  } else {
    std::cerr << "buffer is full" << std::endl;
    std::exit(1);
  }
}

void RepeatReqBuffer::Erase(UINT32 i) {
  ASSERT(i < entries_.size());
  entries_.erase(entries_.begin() + i);
  daddrs_.erase(daddrs_.begin() + i);
}

void RepeatReqBuffer::Erase(Req* req) {
//  for (UINT32 i = 0; i < entries_.size(); i++) {
//    if (entries_[i] == req) {
//      entries_.erase(entries_.begin() + i);
//      daddrs_.erase(daddrs_.begin() + i);
//      return;
//    }
//  }
//
  std::cerr << "This function should not be used." << std::endl;
  std::exit(1);
}

}  // namespace drsim
