/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#include "scheduler.h"
#include "clock.h"
#include "log.h"

using std::endl;

namespace drsim {

// local log wrapper
// NOTE(mjeong): use this macro to print out the last scheduling cycle before simu dies.
#ifdef DEBUG
#define S_DBG_LOG_(xx, ...)  log_buffer_ << __VA_ARGS__;  DBG_LOG_NOCLK(xx, __VA_ARGS__)
#define S_DBG_LOG(xx, ...)   log_buffer_ << __VA_ARGS__;  DBG_LOG(xx, __VA_ARGS__)
#else
#define S_DBG_LOG_(xx, ...)  DBG_LOG_NOCLK(xx, __VA_ARGS__)
#define S_DBG_LOG(xx, ...)   DBG_LOG(xx, __VA_ARGS__)
#endif

/**
 * Constructor.
 */
Scheduler::Scheduler(UINT32 id,
                     ReqBuffer* read_queue,
                     ReqBuffer* write_queue,
                     InFlightReqBuffer* inflight_read_queue,
                     cTimingModel* timing_logic,
                     bool closed_page)
    : id_(id),
      read_queue_(read_queue),
      write_queue_(write_queue),
      inflight_read_queue_(inflight_read_queue),
      timing_logic_(timing_logic),
      keep_search_offset_(false),
      rq_search_offset_(0),
      wq_search_offset_(0),
      closed_page_(closed_page) {
}

Scheduler::~Scheduler() {
}

/**
 * Constructor.
 */
Scheduler_FIFO::Scheduler_FIFO(UINT32 id,
                               ReqBuffer* read_queue,
                               ReqBuffer* write_queue,
                               InFlightReqBuffer* inflight_read_queue,
                               cTimingModel* timing_logic,
                               bool closed_page)
    : Scheduler(id, read_queue, write_queue, inflight_read_queue, timing_logic, closed_page) {
}

bool Scheduler_FIFO::FindCandidate(DRAMAddr* daddr, eMemCmd* dram_cmd, Req** cand) {
  *cand = NULL;

//  // Find the two oldest requests in read/write queues
//  // in case of close-page policy, the second oldest would decide
//  // if we send RD or RDAP
//  // FIXME(mjeong): check (!req->subrank_powerdown())
//  Req* oldest[2];
//  ReqBuffer::iterator wq_it = write_queue_->begin();
//  ReqBuffer::iterator rq_it = read_queue_->begin();
//  for (UINT32 i = 0; i < 2; i++) {
//    if (rq_it != read_queue_->end() && wq_it != write_queue_->end()) {
//      if ((*rq_it)->get_time_received() <= (*wq_it)->get_time_received()) {
//        oldest[i] = *rq_it;
//        rq_it++;
//      }
//      else {
//        oldest[i] = *wq_it;
//        wq_it++;
//      }
//    }
//    else if (rq_it != read_queue_->end()) {
//      oldest[i] = *rq_it;
//      rq_it++;
//    }
//    else if (wq_it != write_queue_->end()) {
//      oldest[i] = *wq_it;
//      wq_it++;
//    }
//    else {
//      oldest[i] = NULL;
//    }
//  }
//
//  if (oldest[0] != NULL
//      && (oldest[0]->get_req_status() == REQ_WAIT
//          || oldest[0]->get_req_status() == REQ_CAS_STARTED)) {
//    // determine what DRAM command to send
//    switch (oldest[0]->get_row_status()) {
//    case ROW_HIT:
//      if (oldest[0]->is_read()) {
//        *dram_cmd = Cmd_RD;
//        if (closed_page_
//            && (oldest[1] == NULL || oldest[1]->get_row_status() != ROW_HIT)
//            && oldest[0]->get_num_burst() == 1)
//          *dram_cmd = Cmd_RDAP;
//      }
//      else {
//        *dram_cmd = Cmd_WR;
//        if (closed_page_
//            && (oldest[1] == NULL || oldest[1]->get_row_status() != ROW_HIT)
//            && oldest[0]->get_num_burst() == 1)
//          *dram_cmd = Cmd_WRAP;
//      }
//      break;
//    case ROW_MISS:      *dram_cmd = Cmd_PRE;      break;
//    case ROW_EMPTY:     *dram_cmd = Cmd_ACT;      break;
//    default:            assert(false);            break;
//    }
//
//    // check DRAM timing constraints
//    if (timing_logic_->checkCommand(*dram_cmd, oldest[0]->dram_addr()))
//      *cand = oldest[0];
//  }

  return (*cand == NULL);
}

/**
 * Constructor.
 */
Scheduler_FRFCFS::Scheduler_FRFCFS(UINT32 id,
                                   ReqBuffer* read_queue,
                                   ReqBuffer* write_queue,
                                   InFlightReqBuffer* inflight_read_queue,
                                   cTimingModel* timing_logic,
                                   bool closed_page,
                                   UINT32 wmode_enter_threshold,
                                   UINT32 wmode_exit_threadhold,
                                   bool wmode_allow_read_CAS,
                                   bool rmode_allow_write_ACTPRE,
                                   bool per_req_policy,
                                   bool split,
                                   UINT32 rot_q_search_off,
                                   UINT32 multi_cas)
    : Scheduler(id, read_queue, write_queue, inflight_read_queue, timing_logic, closed_page),
      wmode_enter_threshold_(wmode_enter_threshold),
      wmode_exit_threshold_(wmode_exit_threadhold),
      wmode_allow_read_CAS_(wmode_allow_read_CAS),
      rmode_allow_write_ACTPRE_(rmode_allow_write_ACTPRE),
      per_req_policy_(per_req_policy),
      split_(split),
      rot_q_search_off_(rot_q_search_off),
      multi_cas_(multi_cas),
      num_RM_to_WM_switch_(fmt_str_ch("num_RM_to_WM_switch", id)),
      num_WM_to_RM_switch_(fmt_str_ch("num_WM_to_RM_switch", id)) {
  rwmode_ = READ_MODE;
}

/**
 * Finds a issuable DRAM command.
 * @param[out] daddr The DRAM address.
 * @param[out] dram_cmd The DRAM command.
 * @param[out] cand The Request the command is for.
 */
bool Scheduler_FRFCFS::FindCandidate(DRAMAddr* daddr, eMemCmd* dram_cmd, Req** cand) {
  bool found = false;

  UpdateRWMode();
  log_buffer_.str("");

  if (rot_q_search_off_ > 0 && !keep_search_offset_) {
      if (read_queue_->size() > 0)
        rq_search_offset_ = (rq_search_offset_ + 1) % read_queue_->size();
      if (write_queue_->size() > 0)
        wq_search_offset_ = (wq_search_offset_ + 1) % write_queue_->size();
  }

  // search from the highest priority requests
  for (Priority p = kStarved; !found && p > kInvalidLow; p--) {
    if (!found)
      found = FindCASCandidate(p, daddr, dram_cmd, cand);
    if (!found)
      found = FindACTCandidate(p, daddr, dram_cmd, cand);
    if (!found)
      found = FindPRECandidate(p, daddr, dram_cmd, cand);
  }

  keep_search_offset_ = false;
  if (found) {
    DBG_LOG(TRACE_SCHEDULER, **cand << " --- issue candidate found --- " << *dram_cmd << endl);
    ASSERT((*cand)->dram_addr().isSamePerChipAddr(*daddr));
    if (rot_q_search_off_ == 1
        && (*dram_cmd == Cmd_RD || *dram_cmd == Cmd_RDAP)
        && !((*cand)->dram_addr().chip_cmd_count - daddr->chip_cmd_count).Empty()) {
      // then we do not advance
      keep_search_offset_ = true;
    }
  }

  return found;
}

/**
 * Updates read-write mode based on the current queue occupancy and mode threshold.
 */
void Scheduler_FRFCFS::UpdateRWMode() {
  if (rwmode_ == READ_MODE
      && write_queue_->size() >= wmode_enter_threshold_) {
    rwmode_ = WRITE_MODE;
    num_RM_to_WM_switch_.increment();
    DBG_LOG(TRACE_MC, "Switching from READ mode to WRITE mode" << endl);
    DBG_LOG(TRACE_MC, "Write Queue: \n" << *write_queue_ << endl);
  } else if (rwmode_ == WRITE_MODE
             && write_queue_->size() <= wmode_exit_threshold_) {
    rwmode_ = READ_MODE;
    num_WM_to_RM_switch_.increment();
    DBG_LOG(TRACE_MC, "Switching from WRITE mode to READ mode" << endl);
  }
}

/**
 * Checks if the queue has a request to the same bank and row hit.
 * @param[in] q The queue being searched.
 * @param[in] daddr The dram address whose bank address is searched.
 * @return Highest priority of the conflicting hit requests.
 */
Priority Scheduler_FRFCFS::HasBankConflictHit(ReqBuffer* q, const DRAMAddr& daddr) {
  Priority highest = kInvalidLow;

  for (ReqBuffer::iterator it = q->begin(); it != q->end(); it++) {
    DRAMAddr hit_daddr = (*it)->dram_addr();
    if (timing_logic_->getRowStatusChipMask(&hit_daddr, ROW_HIT, true)
        && hit_daddr.isBankConflict(daddr)
        && (*it)->priority() > highest)
      highest = (*it)->priority();
  }

  return highest;
}

/**
 * Checks if there is a higher priority request that would be delayed if a CAS
 * is issued to the given addr.
 * @param[in] daddr The dram address.
 * @param[in] p Priority higher than given p is searched.
 */
bool Scheduler_FRFCFS::HasHighpReqDelay(DRAMAddr daddr, Priority p) {
  for (ReqBuffer::iterator it = read_queue_->begin(); it != read_queue_->end(); it++) {
    Req* req = *it;

    if (p < req->priority()) {
      DRAMAddr cas_daddr = req->dram_addr();

      // If it's a row-hit then suppress commands to its chips.
      bool hit_cond = timing_logic_->getRowStatusChipMask(&cas_daddr, ROW_HIT, split_);
      hit_cond = hit_cond && cas_daddr.chip_cmd_count.AndNotEmpty(daddr.chip_cmd_count);
      if (hit_cond)
        return true;

      // If it's a CanPreempt() row-miss then suppress commands to its banks so it can issue PRE.
      DRAMAddr pre_daddr = req->dram_addr();
      bool miss_cond = req->CanPreempt();
      miss_cond = miss_cond && timing_logic_->getRowStatusChipMask(&pre_daddr, ROW_MISS, true);
      miss_cond = miss_cond && pre_daddr.isBankConflict(daddr);
      if (miss_cond)
        return true;

//      if (req->row_status() == ROW_HIT
//          && (req->req_status() == REQ_WAIT || req->req_status() == REQ_CAS_STARTED)
//          && req->dram_addr().isSrConflict(daddr)) {
//        return true;
//      }
//      // if it's a row-miss then suppress commands to its banks
//      // NOTE(mjeong): only STARVED miss can suppres CAS
//      else if (req->row_status() == ROW_MISS
//               && (req->req_status() == REQ_WAIT || req->req_status() == REQ_CAS_STARTED)
//               && req->dram_addr().isBankConflict(daddr)
//               && req->priority() == STARVED) {
//        return true;
//      }
    }
  }

  return false;
}

/**
 * Checks if a column read/write command is issuable for the request.
 * @param[in] req The Request the column read/write command is for.
 * @param[in] p Issuable request priority.
 * @param[out] daddr The dram address column RD/WR command is issuable to.
 */
bool Scheduler_FRFCFS::IsCASIssuable(Req const* req, Priority p, DRAMAddr* daddr) {
  *daddr = req->dram_addr();
  eMemCmd cmd = req->is_read() ? Cmd_RD : Cmd_WR;
  bool ignore_rds = (rwmode_ == WRITE_MODE) && !req->is_read();

  S_DBG_LOG(TRACE_SCHEDULER_V, *req << " testing CAS candadicy :");
  bool cond = req->priority() >= p;
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && timing_logic_->getRowStatusChipMask(daddr, ROW_HIT, split_);
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && timing_logic_->checkCommand(cmd, *daddr, split_, &(daddr->chip_cmd_count));
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && (req->req_status() == REQ_WAIT || req->req_status() == REQ_CAS_STARTED);
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && !req->subrank_powerdown();
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  // This last condition checks if the issuing of a command for this request can delay
  // any higher priority read request. In write mode, it is bypassed.
  cond = cond && (ignore_rds || !HasHighpReqDelay(*daddr, p));
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond << endl);

  // Back-to-back CAS issue
  if (multi_cas_ == 2 || (multi_cas_ == 1 && req->time_cw_completed() != TIME_MAX)) {
    UINT32 n = 1;
    ChipMask chip_cmd_left = req->dram_addr().chip_cmd_count - daddr->chip_cmd_count;
    while (chip_cmd_left.AndNotEmpty(daddr->chip_cmd_count)) {
      chip_cmd_left -= daddr->chip_cmd_count;
      n++;
    }
    daddr->chip_cmd_count = daddr->chip_cmd_count * n;
  }

  return cond;
}

/**
 * Checks if the column command being issued should be auto-precharged.
 * @param[in] req
 * @param[in] daddr The dram address column RD/WR command is issuable to.
 */
bool Scheduler_FRFCFS::AndPrecharge(Req const* req, DRAMAddr const* daddr) {
  if (((per_req_policy_ && req->is_closed_page()) || (!per_req_policy_ && closed_page_))
      && read_queue_->NumReqsToRow(req->dram_addr()) == (req->is_read() ? 1 : 0)
      && write_queue_->NumReqsToRow(req->dram_addr()) == (req->is_read() ? 0 : 1)) {
    // Checks if the chips we are issuing (daddr->chip_cmd_count) have chip_cmd_count == 1.
    ChipMask sub = req->dram_addr().chip_cmd_count - daddr->chip_cmd_count;
    return (sub && daddr->chip_cmd_count).Empty();
  }

  return false;
}

/**
 * Searches an issuable column command in a given queue.
 * @param[in] queue The queue to search.
 * @param[in] p The priority of the request to search.
 * @param[out] daddr The dram address column RD/WR command is issuable to.
 * @param[out] dram_cmd The dram command to issue.
 * @param[out] req The Request the command is for.
 */
bool Scheduler_FRFCFS::SearchCASIssuable(ReqBuffer* queue,
                                         UINT32 offset,
                                         Priority p,
                                         DRAMAddr* daddr,
                                         eMemCmd* dram_cmd,
                                         Req** cand) {
  for (UINT32 i = 0; i < queue->size(); i++) {
    Req* req = (*queue)[(offset + i) % queue->size()];

    if (IsCASIssuable(req, p, daddr)) {
      *cand = req;
      *dram_cmd = req->is_read() ? Cmd_RD : Cmd_WR;

      if (AndPrecharge(req, daddr))
        *dram_cmd = req->is_read() ? Cmd_RDAP : Cmd_WRAP;

      return true;
    }
  }

  return false;
}

/**
 * Searches an issuable column command in read/write queues.
 * @param[in] p The priority of the request to search.
 * @param[out] daddr The dram address column RD/WR command is issuable to.
 * @param[out] dram_cmd The dram command to issue.
 * @param[out] req The Request the command is for.
 */
bool Scheduler_FRFCFS::FindCASCandidate(Priority p,
                                        DRAMAddr* daddr,
                                        eMemCmd* dram_cmd,
                                        Req** cand) {
  bool found = false;

  if (rwmode_ == READ_MODE) {
    found = SearchCASIssuable(read_queue_, rq_search_offset_, p, daddr, dram_cmd, cand);
    if (!found)
      found = SearchCASIssuable(write_queue_, wq_search_offset_, p, daddr, dram_cmd, cand);
  } else if (rwmode_ == WRITE_MODE) {
    found = SearchCASIssuable(write_queue_, wq_search_offset_, p, daddr, dram_cmd, cand);
    if (!found && wmode_allow_read_CAS_)
      found = SearchCASIssuable(read_queue_, rq_search_offset_, p, daddr, dram_cmd, cand);
  } else {
    PANIC("Scheduler is in neither READ or WRITE mode: " << rwmode_);
  }

  return found;
}

/**
 * Checks if a row activate command is issuable for the request.
 * @param[in] req The Request the activate command is for.
 * @param[in] p Issuable request priority.
 * @param[out] daddr The dram address a row activate command is issuable to.
 */
bool Scheduler_FRFCFS::IsACTIssuable(Req const* req, Priority p, DRAMAddr* daddr) {
  *daddr = req->dram_addr();

  S_DBG_LOG(TRACE_SCHEDULER_V, *req << " testing ACT candadicy :");
  bool cond = req->priority() == p;
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && (req->req_status() == REQ_WAIT || req->req_status() == REQ_CAS_STARTED);
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && !req->subrank_powerdown();
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && timing_logic_->getRowStatusChipMask(daddr, ROW_EMPTY, split_);
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && timing_logic_->checkCommand(Cmd_ACT, *daddr, split_, &(daddr->chip_cmd_count));
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond << endl);

  return cond;
}

/**
 * Searches an issuable row activate command in read/write queues.
 * @param[in] p The priority of the request to search.
 * @param[out] daddr The dram address row activate command is issuable to.
 * @param[out] dram_cmd The dram command to issue.
 * @param[out] req The Request the command is for.
 */
bool Scheduler_FRFCFS::FindACTCandidate(Priority p,
                                        DRAMAddr* daddr,
                                        eMemCmd* dram_cmd,
                                        Req** cand) {
  ReqBuffer::iterator it;
  if (rwmode_ == READ_MODE) {
    for (UINT32 i = 0; i < read_queue_->size(); i++) {
      Req* req = (*read_queue_)[(rq_search_offset_ + i) % read_queue_->size()];
      if (IsACTIssuable(req, p, daddr)) {
        *cand = req;
        *dram_cmd = Cmd_ACT;
        return true;
      }
    }

    if (!per_req_policy_ && !rmode_allow_write_ACTPRE_)
      return false;

    for (UINT32 i = 0; i < write_queue_->size(); i++) {
      Req* req = (*write_queue_)[(wq_search_offset_ + i) % write_queue_->size()];
      if ((!per_req_policy_ || (per_req_policy_ && req->allow_rmode_write_ACTPRE()))
          && IsACTIssuable(req, p, daddr)) {
        *cand = req;
        *dram_cmd = Cmd_ACT;
        return true;
      }
    }
  } else if (rwmode_ == WRITE_MODE) {
    for (UINT32 i = 0; i < write_queue_->size(); i++) {
      Req* req = (*write_queue_)[(wq_search_offset_ + i) % write_queue_->size()];
      if (IsACTIssuable(req, p, daddr)) {
        *cand = req;
        *dram_cmd = Cmd_ACT;
        return true;
      }
    }
    // NOTE(mjeong): We don't issue ACT for reads in a write mode.
  }
  return false;
}

/**
 * Checks if a row precharge command is issuable for the request.
 * @param[in] req The Request the precharge command is for.
 * @param[in] p Issuable request priority.
 * @param[out] daddr The dram address a row precharge command is issuable to.
 */
bool Scheduler_FRFCFS::IsPREIssuable(Req const* req, Priority p, DRAMAddr* daddr) {
  *daddr = req->dram_addr();
  bool ignore_rds = (rwmode_ == WRITE_MODE) && !req->is_read() && !wmode_allow_read_CAS_;

  S_DBG_LOG(TRACE_SCHEDULER_V, *req << " testing PRE candadicy :");
  bool cond = (req->priority() == p);
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && (req->req_status() == REQ_WAIT || req->req_status() == REQ_CAS_STARTED);
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && !req->subrank_powerdown();  // NOTE(mjeong): this is for ANY subrank
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && timing_logic_->getRowStatusChipMask(daddr, ROW_MISS, split_);
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  cond = cond && timing_logic_->checkCommand(Cmd_PRE, *daddr, split_, &(daddr->chip_cmd_count));
  S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  // PREcharge can be issued if
  // a. No request to the current open row exists.
  //  OR
  // b. It can preempt (CanPreempt()), and requests to the current open row has
  //    lower priority.
  // If the scheduler is in write mode and read CAS is not allowed in a write
  // mode(!wmode_allow_read_CAS_), then read requests are not considered.
  // TODO(mjeong): once it's stable, should turn into if (.. && .. && ..)
  if (cond) {
    Priority rd_hit_p = kInvalidLow;
    Priority wr_hit_p = kInvalidLow;
    if (!ignore_rds)
      rd_hit_p = HasBankConflictHit(read_queue_, *daddr);
    wr_hit_p = HasBankConflictHit(write_queue_, *daddr);

    bool cond1 = (rd_hit_p == kInvalidLow) && (wr_hit_p == kInvalidLow);
    S_DBG_LOG_(TRACE_SCHEDULER_V, cond1);
    bool cond2 = req->CanPreempt() && (rd_hit_p < p) && (wr_hit_p < p);
    S_DBG_LOG_(TRACE_SCHEDULER_V, cond2);
    cond = cond && (cond1 || cond2);
    S_DBG_LOG_(TRACE_SCHEDULER_V, cond);
  }
  S_DBG_LOG_(TRACE_SCHEDULER_V, endl);

  return cond;
}

/**
 * Searches an issuable row precharge command in read/write queues.
 * @param[in] p The priority of the request to search.
 * @param[out] daddr The dram address row precharge command is issuable to.
 * @param[out] dram_cmd The dram command to issue.
 * @param[out] req The Request the command is for.
 */
bool Scheduler_FRFCFS::FindPRECandidate(Priority p,
                                        DRAMAddr* daddr,
                                        eMemCmd* dram_cmd,
                                        Req** cand) {
  ReqBuffer::iterator it;
  if (rwmode_ == READ_MODE) {
    for (UINT32 i = 0; i < read_queue_->size(); i++) {
      Req* req = (*read_queue_)[(rq_search_offset_ + i) % read_queue_->size()];
      if (IsPREIssuable(req, p, daddr)) {
        *cand = req;
        *dram_cmd = Cmd_PRE;
        return true;
      }
    }

    if (!per_req_policy_ && !rmode_allow_write_ACTPRE_)
      return false;

    for (UINT32 i = 0; i < write_queue_->size(); i++) {
      Req* req = (*write_queue_)[(wq_search_offset_ + i) % write_queue_->size()];
      if ((!per_req_policy_ || (per_req_policy_ && req->allow_rmode_write_ACTPRE()))
          && IsPREIssuable(req, p, daddr)) {
        *cand = req;
        *dram_cmd = Cmd_PRE;
        return true;
      }
    }
  } else if (rwmode_ == WRITE_MODE) {
    for (UINT32 i = 0; i < write_queue_->size(); i++) {
      Req* req = (*write_queue_)[(wq_search_offset_ + i) % write_queue_->size()];
      *daddr = req->dram_addr();
      if (IsPREIssuable(req, p, daddr)) {
        *cand = req;
        *dram_cmd = Cmd_PRE;
        return true;
      }
    }
  }
  return false;
}

void Scheduler_FRFCFS::DumpLog(ostream& os) {
  os << "Log from the last scheduling cycle" << endl;
  os << rwmode_ << endl;
  os << log_buffer_.str();
}

}  // namespace drsim
