/* stats.c - statistical package routines */
/*
 * SimpleScalar Ô Tool Suite
 * © 1994-2003 Todd M. Austin, Ph.D. and SimpleScalar, LLC
 * All Rights Reserved.
 * 
 * THIS IS A LEGAL DOCUMENT BY DOWNLOADING SIMPLESCALAR, YOU ARE AGREEING TO
 * THESE TERMS AND CONDITIONS.
 * 
 * No portion of this work may be used by any commercial entity, or for any
 * commercial purpose, without the prior, written permission of SimpleScalar,
 * LLC (info@simplescalar.com). Nonprofit and noncommercial use is permitted as
 * described below.
 * 
 * 1. SimpleScalar is provided AS IS, with no warranty of any kind, express or
 * implied. The user of the program accepts full responsibility for the
 * application of the program and the use of any results.
 * 
 * 2. Nonprofit and noncommercial use is encouraged.  SimpleScalar may be
 * downloaded, compiled, executed, copied, and modified solely for nonprofit,
 * educational, noncommercial research, and noncommercial scholarship purposes
 * provided that this notice in its entirety accompanies all copies. Copies of
 * the modified software can be delivered to persons who use it solely for
 * nonprofit, educational, noncommercial research, and noncommercial
 * scholarship purposes provided that this notice in its entirety accompanies
 * all copies.
 * 
 * 3. ALL COMMERCIAL USE, AND ALL USE BY FOR PROFIT ENTITIES, IS EXPRESSLY
 * PROHIBITED WITHOUT A LICENSE FROM SIMPLESCALAR, LLC (info@simplescalar.com).
 * 
 * 4. No nonprofit user may place any restrictions on the use of this software,
 * including as modified by the user, by any other authorized user.
 * 
 * 5. Noncommercial and nonprofit users may distribute copies of SimpleScalar
 * in compiled or executable form as set forth in Section 2, provided that
 * either: (A) it is accompanied by the corresponding machine-readable source
 * code, or (B) it is accompanied by a written offer, with no time limit, to
 * give anyone a machine-readable copy of the corresponding source code in
 * return for reimbursement of the cost of distribution. This written offer
 * must permit verbatim duplication by anyone, or (C) it is distributed by
 * someone who received only the executable form, and is accompanied by a copy
 * of the written offer of source code.
 * 
 * 6. SimpleScalar was developed by Todd M. Austin, Ph.D. The tool suite is
 * currently maintained by SimpleScalar LLC (info@simplescalar.com). US Mail:
 * 2395 Timbercrest Court, Ann Arbor, MI 48105.
 * 
 * Copyright © 1994-2003 by Todd M. Austin, Ph.D. and SimpleScalar, LLC.
 */

#include <cstdio>
#include <cstring>
#include <cmath>
#include "ss_stats.h"

namespace SimpleScalar {
using namespace drsim;

void panic(const char* str) {
  printf("%s", str);
  abort();
}
void fatal(const char* str) {
  printf("%s", str);
  exit(1);
}

/* create a sparse array distribution in stat database SDB, while the sparse
   array consumes more memory per bucket than an array distribution, it can
   efficiently map any number of indicies from 0 to 2^32-1, PF specifies the
   distribution components to print for optional format FORMAT; the indicies
   may be optionally replaced with the strings from IMAP, or the entire
   distribution can be printed with the optional user-specified print function
   PRINT_FN */
  struct stat_stat_t *
stat_reg_sdist(//struct stat_sdb_t *sdb,	/* stat database */
    const char *name,		/* stat variable name */
    const char *desc,		/* stat variable description */
    unsigned int init_val,	/* dist initial value */
    int pf,			/* print format, use PF_* defs */
    const char *format,		/* optional variable output format */
    print_fn_t print_fn
    )	/* optional user print function */
{
  struct stat_stat_t *stat;
  struct bucket_t **sarr;

  stat = (struct stat_stat_t *)calloc(1, sizeof(struct stat_stat_t));
  if (!stat) {
    printf("out of virtual memory");
    exit(1);
  }

  stat->name = strdup(name);
  stat->desc = strdup(desc);
  stat->print_me = true;
  stat->format = format ? format : NULL;
  stat->sc = sc_sdist;
  stat->variant.for_sdist.init_val = init_val;
  stat->variant.for_sdist.pf = pf;
  stat->variant.for_sdist.print_fn = print_fn;
  stat->variant.for_sdist.sum_samples = 0;
  stat->variant.for_sdist.sum_sq_samples = 0;

  /* allocate hash table */
  sarr = (struct bucket_t **)calloc(HTAB_SZ, sizeof(struct bucket_t *));
  if (!sarr) {
    printf("out of virtual memory");
    exit(1);
  }
  stat->variant.for_sdist.sarr = sarr;

  /* link onto SDB chain */
//  add_stat(sdb, stat);

  return stat;
}

/* reset sparse array distribution STAT */
  void
stat_reset_sdist(struct stat_stat_t * stat)
{
  struct bucket_t *bucket;

  for (int i=0; i < HTAB_SZ; i++) {
    bucket = stat->variant.for_sdist.sarr[i];
    while (bucket != NULL) {
      struct bucket_t *next_bucket = bucket->next;
      free(bucket);
      bucket = next_bucket;
    }
    stat->variant.for_sdist.sarr[i] = NULL;
  }
  stat->variant.for_sdist.sum_samples = 0;
  stat->variant.for_sdist.sum_sq_samples = 0;
}

/* add NSAMPLES to array or sparse array distribution STAT */
  void
stat_add_samples(struct stat_stat_t *stat,/* stat database */
   UINT32 index,	/* distribution index of samples */
   int nsamples)		/* number of samples to add to dist */
{
  switch (stat->sc)
  {
    case sc_dist:
      {
        unsigned int i = 0;

        /* compute array index */
        i = index / stat->variant.for_dist.bucket_sz;

        /* check for overflow */
        if (i >= stat->variant.for_dist.arr_sz)
          stat->variant.for_dist.overflows += nsamples;
        else
          stat->variant.for_dist.arr[i] += nsamples;

        stat->variant.for_dist.sum_samples += index;
        stat->variant.for_dist.sum_sq_samples += (index*index);
      }
      break;
    case sc_sdist:
      {
        struct bucket_t *bucket;
        int hash = HTAB_HASH(index);

        if (hash < 0 || hash >= HTAB_SZ)
          panic("hash table index overflow");

        /* find bucket */
        for (bucket = stat->variant.for_sdist.sarr[hash];
            bucket != NULL;
            bucket = bucket->next)
        {
          if (bucket->index == index)
            break;
        }
        if (!bucket)
        {
          /* add a new sample bucket */
          bucket = (struct bucket_t *)calloc(1, sizeof(struct bucket_t));
          if (!bucket)
            fatal("out of virtual memory");
          bucket->next = stat->variant.for_sdist.sarr[hash];
          stat->variant.for_sdist.sarr[hash] = bucket;
          bucket->index = index;
          bucket->count = stat->variant.for_sdist.init_val;
        }
        bucket->count += nsamples;

        stat->variant.for_sdist.sum_samples += index;
        stat->variant.for_sdist.sum_sq_samples += (index*index);
      }
      break;
    default:
      panic("stat variable is not an array distribution");
  }
}

/* add a single sample to array or sparse array distribution STAT */
  void
stat_add_sample(struct stat_stat_t *stat,/* stat variable */
    UINT32 index)	/* index of sample */
{
  stat_add_samples(stat, index, 1);
}

/* compare two indicies in a sparse array hash table, used by qsort() */
  static int
compare_fn(void *p1, void *p2)
{
  struct bucket_t **pb1 = (struct bucket_t**) p1, **pb2 = (struct bucket_t**) p2;

  /* compare indices */
  if ((*pb1)->index < (*pb2)->index)
    return -1;
  else if ((*pb1)->index > (*pb2)->index)
    return 1;
  else /* ((*pb1)->index == (*pb2)->index) */
    return 0;
}

/* print a sparse array distribution breifly */
void
print_sdist_short(struct stat_stat_t *stat,	/* stat variable */
                  char *fd)			/* output stream */
{
  unsigned int i, bcount;
  UINT32 imax, imin;
  double btotal, bvar, bavg, bsqsum;
  struct bucket_t *bucket;

  /* count and sum entries */
  bcount = 0; btotal = 0.0; bvar = 0.0; bsqsum = 0.0;
  imax = 0; imin = UINT32_MAX;
  for (i=0; i<HTAB_SZ; i++)
  {
    for (bucket = stat->variant.for_sdist.sarr[i];
        bucket != NULL;
        bucket = bucket->next)
    {
      bcount++;
      btotal += bucket->count;
      /* on-line variance computation, tres cool, no!?! */
      bsqsum += ((double)bucket->count * (double)bucket->count);
      bavg = btotal / (double)bcount;
      bvar = (bsqsum - ((double)bcount * bavg * bavg)) / 
        (double)(((bcount - 1) > 0) ? (bcount - 1) : 1);
      if (bucket->index < imin)
        imin = bucket->index;
      if (bucket->index > imax)
        imax = bucket->index;
    }
  }

  /* print header */
  fd += sprintf(fd, "\n");
  fd += sprintf(fd, "%-28s # %s\n", stat->name, stat->desc);
  fd += sprintf(fd, "%s.count = %u\n", stat->name, bcount);
  fd += sprintf(fd, "%s.total = %.0f\n", stat->name, btotal);
  if (bcount > 0)
  {
    fd += sprintf(fd, "%s.imin = 0x%x\n", stat->name, imin);
    fd += sprintf(fd, "%s.imax = 0x%x\n", stat->name, imax);
  }
  else
  {
    fd += sprintf(fd, "%s.imin = %d\n", stat->name, -1);
    fd += sprintf(fd, "%s.imax = %d\n", stat->name, -1);
  }
  fd += sprintf(fd, "%s.average = %8.4f\n", stat->name, btotal/bcount);
  fd += sprintf(fd, "%s.std_dev = %8.4f\n", stat->name, sqrt(bvar));
  fd += sprintf(fd, "%s.sample_average = %8.4f\n", stat->name, stat->variant.for_sdist.sum_samples/btotal);
  double variance = (btotal > 1) ? (double)(stat->variant.for_sdist.sum_sq_samples -
      stat->variant.for_sdist.sum_samples*stat->variant.for_sdist.sum_samples/btotal)/(btotal - 1) : 0;
  fd += sprintf(fd, "%s.sample_std_dev = %8.4f\n", stat->name, sqrt(variance));
  fd += sprintf(fd, "%s.overflows = 0\n", stat->name);
}

/* print a sparse array distribution */
void
print_sdist(struct stat_stat_t *stat,	/* stat variable */
            char *fd)			/* output stream */
{
  unsigned int i, bcount;
  UINT32 imax, imin;
  double btotal, bsum, bvar, bavg, bsqsum;
  struct bucket_t *bucket;
  int pf = stat->variant.for_sdist.pf;

  /* count and sum entries */
  bcount = 0; btotal = 0.0; bvar = 0.0; bsqsum = 0.0;
  imax = 0; imin = UINT32_MAX;
  for (i=0; i<HTAB_SZ; i++)
  {
    for (bucket = stat->variant.for_sdist.sarr[i];
        bucket != NULL;
        bucket = bucket->next)
    {
      bcount++;
      btotal += bucket->count;
      /* on-line variance computation, tres cool, no!?! */
      bsqsum += ((double)bucket->count * (double)bucket->count);
      bavg = btotal / (double)bcount;
      bvar = (bsqsum - ((double)bcount * bavg * bavg)) / 
        (double)(((bcount - 1) > 0) ? (bcount - 1) : 1);
      if (bucket->index < imin)
        imin = bucket->index;
      if (bucket->index > imax)
        imax = bucket->index;
    }
  }

  /* print header */
  fd += sprintf(fd, "\n");
  fd += sprintf(fd, "%-28s # %s\n", stat->name, stat->desc);
  fd += sprintf(fd, "%s.count = %u\n", stat->name, bcount);
  fd += sprintf(fd, "%s.total = %.0f\n", stat->name, btotal);
  if (bcount > 0)
  {
    fd += sprintf(fd, "%s.imin = 0x%x\n", stat->name, imin);
    fd += sprintf(fd, "%s.imax = 0x%x\n", stat->name, imax);
  }
  else
  {
    fd += sprintf(fd, "%s.imin = %d\n", stat->name, -1);
    fd += sprintf(fd, "%s.imax = %d\n", stat->name, -1);
  }
  fd += sprintf(fd, "%s.average = %8.4f\n", stat->name, btotal/bcount);
  fd += sprintf(fd, "%s.std_dev = %8.4f\n", stat->name, sqrt(bvar));
  fd += sprintf(fd, "%s.sample_average = %8.4f\n", stat->name, stat->variant.for_sdist.sum_samples/btotal);
  double variance = (btotal > 1) ? (double)(stat->variant.for_sdist.sum_sq_samples -
      stat->variant.for_sdist.sum_samples*stat->variant.for_sdist.sum_samples/btotal)/(btotal - 1) : 0;
  fd += sprintf(fd, "%s.sample_std_dev = %8.4f\n", stat->name, sqrt(variance));
  fd += sprintf(fd, "%s.overflows = 0\n", stat->name);

  fd += sprintf(fd, "# pdf == prob dist fn, cdf == cumulative dist fn\n");
  fd += sprintf(fd, "# %14s ", "index");
  if (pf & PF_COUNT)
    fd += sprintf(fd, "%10s ", "count");
  if (pf & PF_PDF)
    fd += sprintf(fd, "%6s ", "pdf");
  if (pf & PF_CDF)
    fd += sprintf(fd, "%6s ", "cdf");
  fd += sprintf(fd, "\n");

  fd += sprintf(fd, "%s.start_dist\n", stat->name);

  if (bcount > 0)
  {
    unsigned int bindex;
    struct bucket_t **barr;

    /* collect all buckets */
    barr = (struct bucket_t **)calloc(bcount, sizeof(struct bucket_t *));
    if (!barr)
      fatal("out of virtual memory");
    for (bindex=0,i=0; i<HTAB_SZ; i++)
    {
      for (bucket = stat->variant.for_sdist.sarr[i];
          bucket != NULL;
          bucket = bucket->next)
      {
        barr[bindex++] = bucket;
      }
    }

    /* sort the array by index */
    qsort(barr, bcount, sizeof(struct bucket_t *), (int (*)(const void*, const void*))compare_fn);

    /* print the array */
    bsum = 0.0;
    for (i=0; i<bcount; i++)
    {
      bsum += (double)barr[i]->count;
      if (stat->variant.for_sdist.print_fn)
      {
        stat->variant.for_sdist.print_fn(stat,
            barr[i]->index,
            barr[i]->count,
            bsum,
            btotal);
      }
      else
      {
        if (stat->format == NULL)
        {
          fd += sprintf(fd, "0x%x ", barr[i]->index);
          if (pf & PF_COUNT)
            fd += sprintf(fd, "%10u ", barr[i]->count);
          if (pf & PF_PDF)
            fd += sprintf(fd, "%6.2f ",
                (double)barr[i]->count/MAX(btotal, 1.0) * 100.0);
          if (pf & PF_CDF)
            fd += sprintf(fd, "%6.2f ", bsum/MAX(btotal, 1.0) * 100.0);
        }
        else
        {
          if (pf == (PF_COUNT|PF_PDF|PF_CDF))
          {
            fd += sprintf(fd, stat->format,
                barr[i]->index, barr[i]->count,
                (double)barr[i]->count/MAX(btotal, 1.0)*100.0,
                bsum/MAX(btotal, 1.0) * 100.0);
          }
          else if (pf == (PF_COUNT|PF_PDF))
          {
            fd += sprintf(fd, stat->format,
                barr[i]->index, barr[i]->count,
                (double)barr[i]->count/MAX(btotal, 1.0)*100.0);
          }
          else if (pf == PF_COUNT)
          {
            fd += sprintf(fd, stat->format,
                barr[i]->index, barr[i]->count);
          }
          else
            fatal("distribution format not yet implemented");
        }
        fd += sprintf(fd, "\n");
      }
    }

    /* all done, release bucket pointer array */
    free(barr);
  }

  fd += sprintf(fd, "%s.end_dist\n", stat->name);
}

} // namespace SimpleScalar
