/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#include <list>
#include <string>
#include <cstdarg>
#include "stat.h"
#include "ss_stats.h"

using std::cout;
using std::cerr;
using std::endl;

namespace drsim {

list<Stat*> Stat::allstat;

Stat::Stat(const string &name)
    : name_(name), print_default_(true), print_periodic_(false) {
  allstat.push_back(this);
}

Stat::~Stat() {
  list<Stat*>::iterator it;
  for (it = allstat.begin() ; it != allstat.end(); it++) {
    if ((*it) == this) {
      allstat.erase(it);
      return;
    }
  }
  PANIC("Can't find " << *this << "in allstat list");
}

void Stat::printall(ostream &os) {
  list<Stat*>::iterator it;

  for (it = allstat.begin(); it != allstat.end(); it++) {
    if ((**it).name_.size() > 0 && (**it).print_default_)
      os << **it << endl;
  }

//  list<struct stat_stat_t *>::iterator it2;
//  for (it2 = allstat_z.begin(); it2 != allstat_z.end(); it2++) {
//    if ((*it2)->print_me) {
//      char buf[1000000];
//      print_sdist((*it2), buf);
//      os << buf;
//    }
//  }
}

void Stat::printall_periodic(ostream &os) {
  list<Stat*>::iterator it;

  for (it = allstat.begin(); it != allstat.end(); it++) {
    if ((**it).name_.size() > 0 && (**it).print_default_ && (**it).print_periodic_) {
      (*it)->print_periodic(os);
      os << endl;
    }
  }
}

ostream& Stat::print_periodic(ostream &os) {
  PANIC("Not implemented yet.");
}

ostream & operator<<(ostream &os, const Stat &stat) {
  return stat.print(os);
}

CounterStat::CounterStat(const string &name) : Stat(name) {
  reset();
}

CounterStat::CounterStat() : Stat("") {
  reset();
}

CounterStat::~CounterStat() {
}

ostream & CounterStat::print(ostream &os) const {
  os << name_ << ":\t" << counter_;
  return os;
}

ostream & CounterStat::print_periodic(ostream &os) {
  os << name_ << ":\t" << counter_periodic_;
  counter_periodic_ = 0;
  return os;
}

ArrayCounterStat::ArrayCounterStat(const string &name, const UINT64 length,
                                   const UINT64 num_per_line, bool print_percent)
  : Stat(name) {
  counters_ = new UINT64[length];
  len_ = length;
  uiNumPerLine_ = num_per_line;
  printPercent_ = print_percent;
  reset();
}

ArrayCounterStat::~ArrayCounterStat() {
  delete [] counters_;
}

void ArrayCounterStat::reset() {
  for (UINT64 i = 0; i < len_; i++) {
    counters_[i] = 0;
  }
}

UINT64 ArrayCounterStat::get_sum() const {
  UINT64 sum = 0;
  for (UINT64 i = 0; i < len_; i++) {
    sum += counters_[i];
  }
  return sum;
}

ostream& ArrayCounterStat::print(ostream &os) const {
  os << name_ << ": " << endl;
  UINT32 cnt = 0;
  UINT64 sum = get_sum();
  for (UINT64 i = 0; i < len_; i++) {
    if (printPercent_)
      os << std::setprecision(4) << 100*((double)counters_[i])/((double)sum) << "% \t";
    else
      os << std::setprecision(4) << counters_[i] << " \t";

    cnt++;
    if (cnt >= uiNumPerLine_ && (i+1) < len_) {
      os << endl;
      cnt = 0;
    }
  }
  return os;
}

SSHistogram::SSHistogram(const char* name, const char *desc)
    : Stat(string(name)) {
  hist_ = SimpleScalar::stat_reg_sdist(name, desc, 0, PF_ALL, NULL, NULL);
  hist_periodic_ = SimpleScalar::stat_reg_sdist(name, desc, 0, PF_ALL, NULL, NULL);
}

SSHistogram::~SSHistogram() {
}

ostream & SSHistogram::print(ostream &os) const {
  if (hist_->variant.for_sdist.sum_samples > 0) {
    char buf[1000000];
    print_sdist(hist_, buf);
    os << buf;
  }
  return os;
}

ostream & SSHistogram::print_periodic(ostream &os) {
  if (hist_->variant.for_sdist.sum_samples > 0) {
    char buf[1000000];
    print_sdist_short(hist_periodic_, buf);
    stat_reset_sdist(hist_periodic_);
    os << buf;
  }
  return os;
}

void SSHistogram::add_sample(UINT32 index) {
  stat_add_sample(hist_, index);
  if (print_periodic_)
    stat_add_sample(hist_periodic_, index);
}

}  // namespace drsim
