/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Doe Hyun Yoon
 */

#include <algorithm>
#include <string>
#include "common.h"
#include "dram.h"
#include "timing_model.h"
#include "clock.h"
#include "log.h"

namespace drsim {

/********************************************************************
 * cProtocol class
 ********************************************************************/

cProtocol::cProtocol(UINT64 dram_freq) {
  initializeTable();
  xInitializeProtocol(dram_freq);
}

cProtocol::~cProtocol(void) {
}

void cProtocol::initializeTable(void) {
  // initialize the protocol timing table
  for (UINT32 p = 0; p  < NUM_Cmd; p++) {
    for (UINT32 c = 0; c < NUM_Cmd; c++) {
      for (UINT32 r = 0; r < 2; r++) {
        for (UINT32 b = 0; b < 2; b++) {
          m_iMinTime[ p ][ c ][ r ][ b ]  =  0;
        }
      }
    }
  }
  // any entry that are not set in xInitializeProtocol will return
  // 0 min_time

  for (UINT32 c = 0; c < NUM_Cmd; c++) {
    m_iCompleteTime[ c ] = 0;
  }
}

void cProtocol::set(eMemCmd prevCmd, eMemCmd currCmd,
                     eSameDiffAll same_rank, eSameDiffAll same_bank,
                     UINT32 min_time ) {
  switch (same_rank) {
  case SAME:
    switch (same_bank) {
    case SAME:
      m_iMinTime[ prevCmd ][ currCmd ][ true ][ true ] = min_time;
      break;
    case DIFF:
      m_iMinTime[ prevCmd ][ currCmd ][ true ][ false ] = min_time;
      break;
    case ALL:
      m_iMinTime[ prevCmd ][ currCmd ][ true ][ true ] = min_time;
      m_iMinTime[ prevCmd ][ currCmd ][ true ][ false ] = min_time;
      break;
    default:
      ASSERT(0);
      break;
    }
    break;
  case DIFF:
    switch (same_bank) {
    case SAME:
      m_iMinTime[ prevCmd ][ currCmd ][ false ][ true ] = min_time;
      break;
    case DIFF:
      m_iMinTime[ prevCmd ][ currCmd ][ false ][ false ] = min_time;
      break;
    case ALL:
      m_iMinTime[ prevCmd ][ currCmd ][ false ][ true ] = min_time;
      m_iMinTime[ prevCmd ][ currCmd ][ false ][ false ] = min_time;
      break;
    default:
      ASSERT(0);
      break;
    }
    break;
  case ALL:
    switch (same_bank) {
    case SAME:
      m_iMinTime[ prevCmd ][ currCmd ][ true ][ true ] = min_time;
      m_iMinTime[ prevCmd ][ currCmd ][ false ][ true ] = min_time;
      break;
    case DIFF:
      m_iMinTime[ prevCmd ][ currCmd ][ true ][ false ] = min_time;
      m_iMinTime[ prevCmd ][ currCmd ][ false ][ false ] = min_time;
      break;
    case ALL:
      m_iMinTime[ prevCmd ][ currCmd ][ true ][ true ] = min_time;
      m_iMinTime[ prevCmd ][ currCmd ][ true ][ false ] = min_time;
      m_iMinTime[ prevCmd ][ currCmd ][ false ][ true ] = min_time;
      m_iMinTime[ prevCmd ][ currCmd ][ false ][ false ] = min_time;
      break;
    default:
      ASSERT(0);
      break;
    }
    break;
  default:
    ASSERT(0);
    break;
  }
}

void cProtocol::xInitializeProtocol(UINT64 dram_freq) {
  using namespace DRAMTiming;
//  double cycle_time = (double)1000 / dram_freq;  // assuming dram_freq is in Mhz
  // this must be called after initializing timing parameters in namespace_dram_timing

  // min time -- table borrowed from Bruce Jacob's book Table 15.8
  //   PrevCmd    CurrCmd    same_rank  same_bank  min_time
  set(Cmd_ACT,    Cmd_ACT,   SAME,      SAME,      RC          );
  set(Cmd_ACT,    Cmd_ACT,   SAME,      DIFF,      RRD         );
  set(Cmd_PRE,    Cmd_ACT,   SAME,      SAME,      RP          );  // the book, {PRE, ACT, s, d} is wrong obviously!
  set(Cmd_PREA,   Cmd_ACT,   SAME,      SAME,      RP          );  // the book, {PRE, ACT, s, d} is wrong obviously!
  set(Cmd_REF,    Cmd_ACT,   SAME,      SAME,      RFC         );
  set(Cmd_RDAP,   Cmd_ACT,   SAME,      SAME,      (AL+BURST+RTP-CCD  + RP));  // Micron spec is AL+ceil(tRTP+tRP)
  set(Cmd_WRAP,   Cmd_ACT,   SAME,      SAME,      (AL+CWD+BURST+WR  -1  + RP));  // Timing for WR--PRE--ACT
  // constraints on ACT: ACT, PRE, and REF on the same rank

  set(Cmd_ACT,   Cmd_RD,    SAME,      SAME,      (RCD-AL)    );
  set(Cmd_RD,    Cmd_RD,    SAME,      ALL,      std::max(BURST, CCD)  );
  set(Cmd_RD,    Cmd_RD,    DIFF,      ALL,      (BURST+RTRS)  );  // assuming RTRS > OST
  set(Cmd_RDAP,  Cmd_RD,    SAME,      DIFF,      std::max(BURST, CCD)  );
  set(Cmd_RDAP,  Cmd_RD,    DIFF,      ALL,      (BURST+RTRS)  );  // assuming RTRS > OST
  set(Cmd_WR,    Cmd_RD,    SAME,      ALL,      (CWD+BURST+WTR  -1)  );  // -1 to compensate for CWD/BURST overlap
  set(Cmd_WR,    Cmd_RD,    DIFF,      ALL,      (CWD+BURST+RTRS-CAS  -1)  );  // assuming RTRS > OST // -1 to compensate for CWD/BURST overlap
  set(Cmd_WRAP,  Cmd_RD,    SAME,      DIFF,      (CWD+BURST+WTR  -1)  );  // -1 to compensate for CWD/BURST overlap
  set(Cmd_WRAP,  Cmd_RD,    DIFF,      ALL,      (CWD+BURST+RTRS-CAS  -1)  );  // assuming RTRS > OST // -1 to compensate for CWD/BURST overlap
  // constraints on RD: ACT on the same rank/bank, RD/WR on the same bus

  set(Cmd_ACT,    Cmd_RDAP,  SAME,      SAME,      (RCD-AL)    );
  set(Cmd_RD,    Cmd_RDAP,  SAME,      ALL,      std::max(BURST, CCD)  );
  set(Cmd_RD,    Cmd_RDAP,  DIFF,      ALL,      (BURST+RTRS)  );  // assuming RTRS > OST
  set(Cmd_RDAP,  Cmd_RDAP,  SAME,      DIFF,      std::max(BURST, CCD)  );
  set(Cmd_RDAP,  Cmd_RDAP,  DIFF,      ALL,      (BURST+RTRS)  );  // assuming RTRS > OST
  set(Cmd_WR,    Cmd_RDAP,  SAME,      ALL,      (CWD+BURST+WTR  -1)  );  // -1 to compensate for CWD/BURST overlap
  set(Cmd_WR,    Cmd_RDAP,  DIFF,      ALL,      (CWD+BURST+RTRS-CAS  -1)  );  // assuming RTRS > OST // -1 to compensate for CWD/BURST overlap
  set(Cmd_WRAP,  Cmd_RDAP,  SAME,      DIFF,      (CWD+BURST+WTR  -1)  );  // -1 to compensate for CWD/BURST overlap
  set(Cmd_WRAP,  Cmd_RDAP,  DIFF,      ALL,      (CWD+BURST+RTRS-CAS  -1)  );  // assuming RTRS > OST // -1 to compensate for CWD/BURST overlap
  // constraints on RDAP: ACT on the same rank/bank, RD/WR on the same bus
  // Note: RDAP is the same as RD

  set(Cmd_ACT,    Cmd_WR,    SAME,      SAME,      (RCD-AL)    );
  set(Cmd_RD,    Cmd_WR,    ALL,      ALL,      (CAS+BURST+RTRS-(AL+CWD)  -1)  );  // assuming RTRS > OST // -1 to compensate for CWD/BURST overlap
  set(Cmd_RDAP,  Cmd_WR,    SAME,      DIFF,      (CAS+BURST+RTRS-(AL+CWD)  -1)  );  // assuming RTRS > OST // -1 to compensate for CWD/BURST overlap
  set(Cmd_RDAP,  Cmd_WR,    DIFF,      ALL,      (CAS+BURST+RTRS-(AL+CWD)  -1)  );  // assuming RTRS > OST // -1 to compensate for CWD/BURST overlap
  set(Cmd_WR,    Cmd_WR,    SAME,      ALL,      std::max(BURST, CCD)  );
  set(Cmd_WRAP,  Cmd_WR,    SAME,      DIFF,      std::max(BURST, CCD)  );
  set(Cmd_WR,    Cmd_WR,    DIFF,      ALL,      (BURST+OST)    );
  set(Cmd_WRAP,  Cmd_WR,    DIFF,      ALL,      (BURST+OST)    );
  // constraints on WR: ACT on the same rank/bank, RD/WR on the same bus

  set(Cmd_ACT,    Cmd_WRAP,  SAME,      SAME,      (RCD-AL)    );
  set(Cmd_RD,    Cmd_WRAP,  ALL,      ALL,      (CAS+BURST+RTRS-(AL+CWD)  -1)  );  // assuming RTRS > OST // -1 to compensate for CWD/BURST overlap
  set(Cmd_RDAP,  Cmd_WRAP,  SAME,      DIFF,      (CAS+BURST+RTRS-(AL+CWD)  -1)  );  // assuming RTRS > OST // -1 to compensate for CWD/BURST overlap
  set(Cmd_RDAP,  Cmd_WRAP,  DIFF,      ALL,      (CAS+BURST+RTRS-(AL+CWD)  -1)  );  // assuming RTRS > OST // -1 to compensate for CWD/BURST overlap
  set(Cmd_WR,    Cmd_WRAP,  SAME,      ALL,      std::max(BURST, CCD)  );
  set(Cmd_WRAP,  Cmd_WRAP,  SAME,      DIFF,      std::max(BURST, CCD)  );
  set(Cmd_WR,    Cmd_WRAP,  DIFF,      ALL,      (BURST+OST)    );
  set(Cmd_WRAP,  Cmd_WRAP,  DIFF,      ALL,      (BURST+OST)    );
  // constraints on WRAP: ACT on the same rank/bank, RD/WR on the same bus
  // Note: WRAP is the same as WR

  set(Cmd_ACT,    Cmd_PRE,  SAME,      SAME,      RAS          );
  set(Cmd_RD,    Cmd_PRE,  SAME,      SAME,      (AL+BURST+RTP-CCD)  );  // BURST and CCD cancel each other, so it's simply RTP
  set(Cmd_WR,    Cmd_PRE,  SAME,      SAME,      (AL+CWD+BURST+WR  -1)    );  // -1 to compensate for CWD/BURST overlap
  // constraints on PRE: ACT, RD, and WR on the same rank/bank

  set(Cmd_ACT,    Cmd_PREA,  SAME,      SAME,      RAS          );
  set(Cmd_RD,    Cmd_PREA,  SAME,      SAME,      (AL+BURST+RTP-CCD)  );  // BURST and CCD cancel each other, so it's simply RTP
  set(Cmd_WR,    Cmd_PREA,  SAME,      SAME,      (AL+CWD+BURST+WR  -1)    );  // -1 to compensate for CWD/BURST overlap
  // constraints on PREA: same as PRE

  set(Cmd_REF,    Cmd_REF,  SAME,      ALL,      RFC          );
  set(Cmd_PRE,    Cmd_REF,  SAME,      ALL,      RP          );
  set(Cmd_PREA,  Cmd_REF,  SAME,      ALL,      RP          );
  set(Cmd_RDAP,  Cmd_REF,  SAME,      ALL,      (AL+BURST+RTP-CCD  + RP)       );  // same as RDAP to ACT
  set(Cmd_WRAP,  Cmd_REF,  SAME,      ALL,      (AL+CWD+BURST+WR  -1  + RP)   );  // same as WRAP to ACT
  // constraints on REF: REF and PRE on the same rank

  // (mjeong) constraints on power-down entry
  set(Cmd_ACT,   Cmd_PDN,  SAME,     ALL,      1);  // TODO(mjeong) is it 1 or 0?
  set(Cmd_PRE,   Cmd_PDN,  SAME,     ALL,      1);
  set(Cmd_PREA,  Cmd_PDN,  SAME,     ALL,      1);
  set(Cmd_RD,    Cmd_PDN,  SAME,     ALL,      CAS+BURST+1);
  set(Cmd_RDAP,  Cmd_PDN,  SAME,     ALL,      CAS+BURST+1);
  set(Cmd_WR,    Cmd_PDN,  SAME,     ALL,      CWD+BURST+WR);
  set(Cmd_WRAP,  Cmd_PDN,  SAME,     ALL,      CWD+BURST+WR+1);
  set(Cmd_REF,   Cmd_PDN,  SAME,     ALL,      1);
//
  // constraints on power-down exit
  set(Cmd_PDN,   Cmd_PUP,  SAME,     ALL,      CKEMIN);

  set(Cmd_PUP,   Cmd_ACT,  SAME,     ALL,      XP);
  set(Cmd_PUP,   Cmd_PRE,  SAME,     ALL,      XP);
  set(Cmd_PUP,   Cmd_PREA, SAME,     ALL,      XP);
  set(Cmd_PUP,   Cmd_RD,   SAME,     ALL,      XPDLL);
  set(Cmd_PUP,   Cmd_RDAP, SAME,     ALL,      XPDLL);
  set(Cmd_PUP,   Cmd_WR,   SAME,     ALL,      XP);
  set(Cmd_PUP,   Cmd_WRAP, SAME,     ALL,      XP);
  set(Cmd_PUP,   Cmd_REF,  SAME,     ALL,      XP);

  // complete time -- mem cont needs completion times ofRD, RDP, WR, WRP
  m_iCompleteTime[ Cmd_ACT   ] = RCD;
  m_iCompleteTime[ Cmd_RD   ] = CAS+BURST;
  m_iCompleteTime[ Cmd_RDAP ] = CAS+BURST;  // Note: it's read complete time, not including precharge time
  m_iCompleteTime[ Cmd_WR   ] = CWD+BURST;
  m_iCompleteTime[ Cmd_WRAP ] = CWD+BURST;  // Note: it's write complete time, not including precharge time
  m_iCompleteTime[ Cmd_PRE   ] = RP;
  m_iCompleteTime[ Cmd_PREA ] = RP;
  m_iCompleteTime[ Cmd_REF   ] = RFC;
  m_iCompleteTime[ Cmd_NOP   ] = 0;
}


UINT32 cProtocol::getMinTime( const eMemCmd prevCmd, const eMemCmd currCmd,
                              const bool same_rank, const bool same_bank ) const {
  // can't handle these yet
  ASSERT(currCmd != Cmd_PREA);

  return m_iMinTime[ prevCmd ][ currCmd ][ same_rank ][ same_bank ];
}
UINT32 cProtocol::getCompleteTime(const eMemCmd currCmd) const {
  return m_iCompleteTime[ currCmd ];
}
/********************************************************************/

/********************************************************************
 *  cMemXact class
 ********************************************************************/
cMemReq::cMemReq(void)
    : m_eMemCmd( Cmd_NOP), m_uiRank(0), m_uiBank(0), m_time(-1000) {  // don't care much on the initial value of m_DRAMAddr
}
cMemReq::cMemReq(eMemCmd mem_cmd, UINT32 rk, UINT32 bk, TIME time)
    : m_eMemCmd(mem_cmd), m_uiRank(rk), m_uiBank(bk), m_time(time) {
}
cMemReq::~cMemReq(void) {
}
/********************************************************************/

/********************************************************************
 * cBank class
 ********************************************************************/
cBank::cBank(void) {
  m_eBankState = BANK_CLOSE;
  m_addrRow = -1;
}
cBank::~cBank(void) {
}
bool cBank::isBankOpen(void) const {
  return (m_eBankState == BANK_OPEN);
}
ADDRESS cBank::getRow(void) const {
  return m_addrRow;
}
void cBank::xClose(const bool bDontCheckOpen) {
  ASSERT(bDontCheckOpen || isBankOpen());
  m_eBankState = BANK_CLOSE;
  m_addrRow = -1;
}
void cBank::xOpen(const ADDRESS row) {
  ASSERT(isBankOpen() == false);
  m_eBankState = BANK_OPEN;
  m_addrRow = row;
}
void cBank::xIssueReq(const cMemReq &curr_req) {
  if (curr_req.getCmd() == Cmd_ACT) {
    ASSERT(isBankOpen() == false);
    m_cPrevACT = curr_req;
  } else if ( curr_req.getCmd() == Cmd_PRE || curr_req.getCmd() == Cmd_PREA ||
              curr_req.getCmd() == Cmd_RDAP || curr_req.getCmd() == Cmd_WRAP ) {
    // ASSERT(isBankOpen());  // DRAM spec says that issuing PRE on close bank is ok!
    m_cPrevPRE = curr_req;
  } else if (curr_req.getCmd() == Cmd_REF) {
    ASSERT(isBankOpen() == false);
    m_cPrevPRE = curr_req;
//    m_cPrevREF = curr_req;  // Refresh is a kind of Precharge since it ends up with a bank closed
  }

  if ( curr_req.getCmd() == Cmd_RD || curr_req.getCmd() == Cmd_WR ||
       curr_req.getCmd() == Cmd_RDAP || curr_req.getCmd() == Cmd_WRAP ) {
    if (isBankOpen() == false) {
      using namespace std;
      cout << "Curr Req: " << curr_req.getCmd() << endl;
      ASSERT(0);
    }
    m_cPrevRW = curr_req;
  }
}
const cMemReq& cBank::getPrevACT(void) const {
  return m_cPrevACT;
}
const cMemReq& cBank::getPrevPRE(void) const {
  return m_cPrevPRE;
}
//const cMemReq& cBank::getPrevREF(void) const {
//  return m_cPrevREF;
//}
const cMemReq& cBank::getPrevRW(void) const {
  return m_cPrevRW;
}
/********************************************************************/

/********************************************************************
 *  cRank class
 ********************************************************************/
cRank::cRank(void) {
  m_uiTotalBanks = 8;  // default 8 banks per rank
  m_uiNumOpenBanks = 0;
  m_atimeFourActs[0] = -1000;
  m_atimeFourActs[1] = -1000;
  m_atimeFourActs[2] = -1000;
  m_atimeFourActs[3] = -1000;
}
cRank::cRank(const UINT32 total_banks) {
  m_uiTotalBanks = total_banks;
  m_uiNumOpenBanks = 0;
  m_atimeFourActs[0] = -1000;
  m_atimeFourActs[1] = -1000;
  m_atimeFourActs[2] = -1000;
  m_atimeFourActs[3] = -1000;
}
cRank::~cRank(void) {
}
void cRank::xOpen(const UINT32 bk) {
  ASSERT(m_uiNumOpenBanks < m_uiTotalBanks);

  m_uiNumOpenBanks++;
}
void cRank::xClose(const UINT32 bk) {
  m_uiNumOpenBanks--;
  ASSERT(m_uiNumOpenBanks >= 0 && m_uiNumOpenBanks < m_uiTotalBanks);
}
void cRank::xIssueReq(const cMemReq &curr_req) {
  if (curr_req.getCmd() == Cmd_ACT) {
    m_cPrevACT = curr_req;
  } else if ( curr_req.getCmd() == Cmd_PRE || curr_req.getCmd() == Cmd_PREA ||
              curr_req.getCmd() == Cmd_RDAP || curr_req.getCmd() == Cmd_WRAP ) {
    m_cPrevPRE = curr_req;
  } else if (curr_req.getCmd() == Cmd_REF) {
    m_cPrevPRE = curr_req;
//    m_cPrevREF = curr_req;  // Refresh is a kind of Precharge since it ends up with a bank closed
  } else if (curr_req.getCmd() == Cmd_PDN) {
    m_cPrevPDN = curr_req;
  } else if (curr_req.getCmd() == Cmd_PUP) {
    m_cPrevPUP = curr_req;
  }

  if (curr_req.getCmd() == Cmd_ACT) {
    for (INT32 i = 2; i >= 0; i--) {
      m_atimeFourActs[i+1] = m_atimeFourActs[i];
    }
    m_atimeFourActs[0] = curr_req.getTime();
  }
}
bool cRank::violateFAW(const TIME curr_time) const {
  if (m_atimeFourActs[3] < 0) return false;  // have not observed 4 ACTs yet

  TIME diff = curr_time - m_atimeFourActs[3];
  if (diff >= DRAMTiming::FAW) {
    return false;
  }

  return true;
}
const cMemReq& cRank::getPrevACT(void) const {
  return m_cPrevACT;
}
const cMemReq& cRank::getPrevPRE(void) const {
  return m_cPrevPRE;
}
const cMemReq& cRank::getPrevPDN(void) const {
  return m_cPrevPDN;
}
const cMemReq& cRank::getPrevPUP(void) const {
  return m_cPrevPUP;
}
//const cMemReq& cRank::getPrevREF(void) const {
//  return m_cPrevREF;
//}

UINT32 cRank::getRefCounter(void) const {
  return m_uiRefCounter;
}
void cRank::setRefCounter(UINT32 refCounter) {
  m_uiRefCounter = refCounter;
}
void cRank::incrementRefCounter(void) {
  m_uiRefCounter++;
}

/********************************************************************/

/********************************************************************
 *  cDBus class
 ********************************************************************/
cDBUS::cDBUS(void) {
}
cDBUS::~cDBUS(void) {
}
void cDBUS::xIssueReq(const cMemReq &curr_req) {
  if ( curr_req.getCmd() == Cmd_RD || curr_req.getCmd() == Cmd_RDAP ||
       curr_req.getCmd() == Cmd_WR || curr_req.getCmd() == Cmd_WRAP ) {
    m_cPrevRW = curr_req;
  }
}
const cMemReq& cDBUS::getPrevRW(void) const {
  return m_cPrevRW;
}
/********************************************************************/

/********************************************************************
 *  cTimingModel class
 ********************************************************************/
bool cTimingModel::s_bInitialized = false;
cTimingModel::cTimingModel( const UINT32 channel_id,
                            const UINT32 num_ranks,
                            const UINT32 num_banks,
                            const UINT32 num_subranks,
                            const UINT32 num_ABUS,
                            const UINT64 dram_freq )
    : m_uiChannelID(channel_id),
    m_uiNumRanksPerChannel(num_ranks),
    m_uiNumBanksPerRank(num_banks),
    m_uiNumSubranksPerRank(num_subranks),
    m_uiNumABUS(num_ABUS),
    m_cProtocol( dram_freq) {
  // allocate states for timing model

  m_bpSubRankAvailability = new bool[num_subranks];
  for (UINT32 a = 0; a < num_ABUS; a++) {
    m_bpSubRankAvailability[a] = true;
  }

  m_sDBUS = new cDBUS[num_subranks];
  m_rsRank = new cRank *[num_ranks];
  m_rbsBank = new cBank **[num_ranks];
  for (UINT32 rk = 0; rk < num_ranks; rk++) {
    m_rsRank[rk] = (cRank *) ::operator new(num_subranks * sizeof(cRank));  // allocate just space
    for (UINT32 sr = 0; sr < num_subranks; sr++) {
      new(&m_rsRank[rk][sr]) cRank(num_banks);  // initialize each cRAnk with num_banks argument
      m_rsRank[rk][sr].setRefCounter(((DRAMTiming::REFI)*rk)/num_ranks);
      // initialize ref counter; spread rank refresh timing uniformly
      // TODO(dhyoon): may spread refresh timing with subranks also, but not yet...
    }
    m_rbsBank[rk] = new cBank *[num_banks];
    for (UINT32 bk = 0; bk < num_banks; bk++) {
      m_rbsBank[rk][bk] = new cBank[num_subranks];
    }  // banks
  }  // ranks
#ifdef LOG_ENABLE
  if (s_bInitialized == false) {
    DBG_LOG_NOCLK(TRACE_TM, "TIME \t   " << "CMD \t" << "[ch:rk:bk:sr{");
    for (UINT32 sr = 0; sr  <m_uiNumSubranksPerRank; sr++) {
      DBG_LOG_NOCLK(TRACE_TM, sr);
      if (sr < (m_uiNumSubranksPerRank-1)) {
        DBG_LOG_NOCLK(TRACE_TM, ":");
      }
    }
    DBG_LOG_NOCLK(TRACE_TM, "}:row:col] \t");
    if (m_uiNumSubranksPerRank >= 4) {
      DBG_LOG_NOCLK(TRACE_TM, "\t");
    }
    DBG_LOG_NOCLK(TRACE_TM, "complete \t");
    for (UINT32 sr = 0; sr  <m_uiNumSubranksPerRank; sr++) {
      DBG_LOG_NOCLK(TRACE_TM, "sr" << sr << " data\t");
    }
    DBG_LOG_NOCLK(TRACE_TM, std::endl);
    s_bInitialized = true;
  }
#endif
}  // constructor

cTimingModel::~cTimingModel(void) {
  // deallocate states for timing model

  for (UINT32 rk = 0; rk < m_uiNumRanksPerChannel; rk++) {
    for (UINT32 bk = 0; bk < m_uiNumBanksPerRank; bk++) {
      delete [] m_rbsBank[rk][bk];
    }
    for (UINT32 sr = 0; sr < m_uiNumSubranksPerRank; sr++) {
      m_rsRank[rk][sr].~cRank();  // call descructor without deallocation for each cRank
    }  // banks
    ::operator delete(m_rsRank[rk]);  // just deallocate space without destructor
    delete [] m_rbsBank[rk];
  }  // ranks
  delete [] m_sDBUS;
  delete [] m_rsRank;
  delete [] m_rbsBank;

  delete [] m_bpSubRankAvailability;
}  // destructor

bool cTimingModel::checkViolation(const cMemReq &prev_req, const cMemReq &curr_req) const {
  eMemCmd prevCmd = prev_req.getCmd();
  eMemCmd currCmd = curr_req.getCmd();
  bool same_rank = prev_req.sameRank(curr_req);
  bool same_bank = prev_req.sameBank(curr_req);
  TIME prevTime = prev_req.getTime();
  TIME currTime = curr_req.getTime();

  UINT32 minTime = m_cProtocol.getMinTime(prevCmd, currCmd, same_rank, same_bank);

#if LOG_ENABLE
  if (minTime > 0 ) {
    // level 10 -- this will generate huge amount of print outs
    DBG_LOG_NOCLK(10, "prev: " << prevCmd << ", curr: " << currCmd << " r/b: " << same_rank << "/" << same_bank << std::endl);
    DBG_LOG_NOCLK(10, "curr time: " << currTime << ", prev time: " << prevTime << ", min time: " << minTime << std::endl);
  }
#endif

  if ((currTime - prevTime) >= minTime) {
    return false;  // ok
  } else {
    DBG_LOG_NOCLK(10, "prev: " << prevCmd << ", curr: " << currCmd << " r/b: " << same_rank << "/" << same_bank << std::endl);
    return true;  // violates timing constraints
  }
}

void cTimingModel::xCycle(void) {
  // per-cycle initialization here!
  for (UINT32 rk = 0; rk < m_uiNumRanksPerChannel; rk++) {
    for (UINT32 sr = 0; sr < m_uiNumSubranksPerRank; sr++) {
      m_rsRank[rk][sr].incrementRefCounter();
#ifndef IGNORE_REFRESH
      if (m_rsRank[rk][sr].getRefCounter() > (DRAMTiming::refresh_defer_allowance)*(DRAMTiming::REFI)) {
        std::cout << "Refresh Violation: rank-" << rk << ", sr-" << sr << ", RefCounter-" << m_rsRank[rk][sr].getRefCounter();
        std::cout << " is larger than 8xREFI (" << (DRAMTiming::refresh_defer_allowance)*(DRAMTiming::REFI) << ")";
        std::cout << " @ DRAM time:" << clock::get_dram_clk() << std::endl;
        ASSERT(0);
      }
#endif // IGNORE_REFRESH
    }
  }

  // initialize subrank availability vector
  for (UINT32 a = 0; a < m_uiNumABUS; a++) {
    m_bpSubRankAvailability[a] = true;
  }

#ifndef IGNORE_REFRESH // TODO(dhyoon): may use DBG_LOG_NOCLK(TRACE_REFRESH, ... );
  bool at_least_one = false;
  for (UINT32 rk = 0; rk < m_uiNumRanksPerChannel; rk++) {
    for (UINT32 sr = 0; sr < m_uiNumSubranksPerRank; sr++) {
      if ((m_rsRank[rk][sr].getRefCounter() % DRAMTiming::REFI) == 0) {
        if (at_least_one == false) {
          std::cout << "Need Refresh at ";
        }
        at_least_one = true;
        std::cout << "[" << rk << "][" << sr << "] " << m_rsRank[rk][sr].getRefCounter() << "@" << clock::get_dram_clk() << " : ";
      }
    }
  }
  if (at_least_one) {
    std::cout << std::endl;
  }
#endif
}

bool cTimingModel::checkCommand(const eMemCmd cmd, const DRAMAddr daddr) const {
  ChipMask chip_issuable;
  return checkCommand(cmd, daddr, false, &chip_issuable);
}

// FIXME: do some thing with this function
// check whether a command is ok to be issued at this cycle
bool cTimingModel::checkCommand(const eMemCmd cmd, const DRAMAddr daddr, bool split, ChipMask* chip_issuable) const {
  // construct a currCmd
  UINT32 rk = daddr.rank;
  UINT32 bk_ = daddr.bank;
  const ChipMask subrank_vector = daddr.chip_cmd_count;//subrank_mask;

  UINT32 bk_start = bk_;
  UINT32 bk_end = bk_ + 1;
  if (cmd == Cmd_PREA || cmd == Cmd_REF) {
    bk_start = 0;
    bk_end = m_uiNumBanksPerRank;
  }  // violation check of PREA should be done for all banks

  UINT32 uiViolation = 0;
  UINT32 uiViolation_sr = 0;
  chip_issuable->SetAll(0);
  for (UINT32 bk = bk_start; bk < bk_end; bk++) {
    cMemReq curr_req(cmd, rk, bk, clock::get_dram_clk());

    // For each chip
    for (UINT32 sr = 0; sr < m_uiNumSubranksPerRank; sr++) {
      if (subrank_vector[sr]) {
        if (m_bpSubRankAvailability[sr/(m_uiNumSubranksPerRank/m_uiNumABUS)] == false) {
          uiViolation_sr = 1000;  // A command to that chip has been scheduled this cycle.
        } else {
          // Check DBUS availability.
          uiViolation_sr += checkViolation(m_sDBUS[sr].getPrevRW(), curr_req);

          uiViolation_sr += checkViolation(m_rsRank[rk][sr].getPrevACT(), curr_req);
          uiViolation_sr += checkViolation(m_rsRank[rk][sr].getPrevPRE(), curr_req);
          uiViolation_sr += checkViolation(m_rsRank[rk][sr].getPrevPDN(), curr_req);
          uiViolation_sr += checkViolation(m_rsRank[rk][sr].getPrevPUP(), curr_req);
  //         uiViolation_sr += checkViolation(m_crsRank[rk][sr].getPrevREF(), curr_req);
          if (curr_req.getCmd() == Cmd_ACT) {
            uiViolation_sr += m_rsRank[rk][sr].violateFAW(clock::get_dram_clk());
          }

          uiViolation_sr += checkViolation(m_rbsBank[rk][bk][sr].getPrevACT(), curr_req);
          uiViolation_sr += checkViolation(m_rbsBank[rk][bk][sr].getPrevPRE(), curr_req);
  //         uiViolation_sr += checkViolation(m_crbsBank[rk][bk][sr].getPrevREF(), curr_req);
          uiViolation_sr += checkViolation(m_rbsBank[rk][bk][sr].getPrevRW(), curr_req);
        }
        uiViolation += uiViolation_sr;
        if (uiViolation_sr == 0) { (*chip_issuable)[sr] = 1; }
        uiViolation_sr = 0;
      }  // if (subrank_vector[sr])
    }  // for subrank_vector
  }  // bk

  return (uiViolation == 0) || (split && !chip_issuable->Empty());
}

//void cTimingModel::xReserveCommand(const eMemCmd cmd, const DRAMAddr daddr) {
//  ASSERT(cmd == Cmd_RD || cmd == Cmd_WR);
//
//  TIME repeat_interval = m_cProtocol.getMinTime(cmd, cmd, true, true);
//  DBG_LOG_NOCLK(TRACE_MC, g_stringMemCmd[cmd] << " is reserved to be issued after " << repeat_interval << "dclk" << std::endl);
//
//  cMemReq curr_req( cmd, daddr.rank, daddr.bank, clock::get_dram_clk() + repeat_interval);
//
//  for (UINT32 sr = 0; sr < m_uiNumSubranksPerRank; sr++) {
//    if (daddr.chip_cmd_count[sr]) {
//      m_sDBUS[sr].xIssueReq(curr_req);
//      m_rsRank[daddr.rank][sr].xIssueReq(curr_req);
//      m_rbsBank[daddr.rank][daddr.bank][sr].xIssueReq(curr_req);
//    }
//  }
//}

TIME cTimingModel::xIssueCommand(const eMemCmd cmd, const DRAMAddr daddr) {
  UINT32 n = 1;
  for (INT32 i = 0; i < daddr.chip_cmd_count.num_chip; i++) {
    if (daddr.chip_cmd_count[i] > n)
      n = daddr.chip_cmd_count[i];
  }

  return xIssueCommand(cmd, daddr, n);
}

// issue a command to update the timing model
TIME cTimingModel::xIssueCommand(const eMemCmd cmd, const DRAMAddr daddr, UINT32 n) {
  // make sure that the command is ok to be issued
  ASSERT(checkCommand(cmd, daddr));

  // construct a currCmd
  UINT32 rk = daddr.rank;
  UINT32 bk_  = daddr.bank;
  ADDRESS row = daddr.row;
  const ChipMask subrank_vector = daddr.chip_cmd_count;//subrank_mask;

//  TRACE_CLK(TRACE_TM);
  DBG_LOG_NOCLK(TRACE_TM, g_stringMemCmd[cmd] << "\t");
  DBG_LOG_NOCLK(TRACE_TM, daddr << "\t");
  DBG_LOG_NOCLK(TRACE_TM, std::setw(2) << std::setfill('0') << clock::get_dram_clk()+m_cProtocol.getCompleteTime(cmd)-1 << " \t");

  UINT32 bk_start = bk_;
  UINT32 bk_end = bk_ + 1;
  if (cmd == Cmd_PREA || cmd == Cmd_REF) {
    bk_start = 0;
    bk_end = m_uiNumBanksPerRank;
  }  // violation check of PREA should be done for all banks

  // NOTE(mjeong): This is based on an assumption that RD/WR can be repeated back-to-back
  ASSERT(n == 1 || cmd == Cmd_RD || cmd == Cmd_WR || cmd == Cmd_RDAP || cmd == Cmd_WRAP );
  TIME repeat_interval = 0;
  if (cmd == Cmd_RD || cmd == Cmd_RDAP)
    repeat_interval = m_cProtocol.getMinTime(Cmd_RD, Cmd_RD, true, true);
  else if (cmd == Cmd_WR || cmd == Cmd_WRAP)
    repeat_interval = m_cProtocol.getMinTime(Cmd_WR, Cmd_WR, true, true);
  if (n > 1)
    DBG_LOG_NOCLK(TRACE_MC, g_stringMemCmd[cmd] << "x" << n << " is issued with repeat_interval of " << repeat_interval << std::endl);

  for (UINT32 bk = bk_start; bk < bk_end; bk++) {
    cMemReq curr_req( cmd, rk, bk, clock::get_dram_clk() + (n-1)*repeat_interval);

    // for each subrank
    for (UINT32 sr = 0; sr < m_uiNumSubranksPerRank; sr++) {
      if (subrank_vector[sr]) {
        // update the prevCmd with the currCmd at rk, b
        // to DBus, Rank, and Bank
        m_sDBUS[sr].xIssueReq(curr_req);
        m_rsRank[rk][sr].xIssueReq(curr_req);  // this pushes ACT timing into the rank structure
        m_rbsBank[rk][bk][sr].xIssueReq(curr_req);

        // update bank rank states, other command specific stuffs here
        switch (cmd) {
        case Cmd_ACT:
          m_rsRank[rk][sr].xOpen(bk);
          m_rbsBank[rk][bk][sr].xOpen(row);  // open bank
          break;

        case Cmd_PRE:
        case Cmd_PREA: // operations for PREA is the same as those of PRE, but the outermost for loop will apply PRE for all banks
          if (isBankOpen(rk, sr, bk)) { // well, it's ok to issue PRE for closed bank... but why does this happen?
            m_rsRank[rk][sr].xClose(bk);
          }
          m_rbsBank[rk][bk][sr].xClose(true);  // close bank -- don't check open
          break;

        case Cmd_RDAP:
          DBG_LOG_NOCLK(TRACE_TM, "\t" << "D:" << clock::get_dram_clk()+DRAMTiming::CAS << "--" << clock::get_dram_clk()+DRAMTiming::CAS+DRAMTiming::BURST-1 << " ");
          m_rsRank[rk][sr].xClose(bk);
          m_rbsBank[rk][bk][sr].xClose( );  // close bank
          break;

        case Cmd_WRAP:
          DBG_LOG_NOCLK(TRACE_TM, "\t" << "D:" << clock::get_dram_clk()+DRAMTiming::CWD << "--" << clock::get_dram_clk()+DRAMTiming::CWD+DRAMTiming::BURST-1 << " ");
          m_rsRank[rk][sr].xClose(bk);
          m_rbsBank[rk][bk][sr].xClose( );  // close bank
          break;

        case Cmd_RD:
          DBG_LOG_NOCLK(TRACE_TM, "\t" << "D:" << clock::get_dram_clk()+DRAMTiming::CAS << "--" << clock::get_dram_clk()+DRAMTiming::CAS+DRAMTiming::BURST-1 << " ");
          break;
        case Cmd_WR:
          DBG_LOG_NOCLK(TRACE_TM, "\t" << "D:" << clock::get_dram_clk()+DRAMTiming::CWD << "--" << clock::get_dram_clk()+DRAMTiming::CWD+DRAMTiming::BURST-1 << " ");
          break;

        case Cmd_REF: // the outermost loop executes on all banks
          // all banks should be closed
          ASSERT(isBankOpen(rk, sr, bk) == false);
          if (bk == bk_start) { // should be done just once
            UINT32 ref_counter = m_rsRank[rk][sr].getRefCounter();
            if (ref_counter > DRAMTiming::REFI) { ref_counter -= DRAMTiming::REFI; }
            else                                 { ref_counter = 0; }
            m_rsRank[rk][sr].setRefCounter( ref_counter);
            // TODO(dhyoon): DBG_LOG_NOCLK(TRACE_REFRESH, .....);
          }
          break;

        case Cmd_NOP:
        default:
          // do nothing: bank/rank states aren't changed with these commands
          break;
        }  // switch(cmd)
      } else { // if (subrank_vector[sr])
        DBG_LOG_NOCLK(TRACE_TM, "\t\t");
      }  // if (subrank_vector[sr])
    }  // for subrank_vector
  }  // bk

  DBG_LOG_NOCLK(TRACE_TM, std::endl);

  // update subrank availability vector
  UINT32 step = m_uiNumSubranksPerRank / m_uiNumABUS;
  ASSERT((m_uiNumSubranksPerRank % m_uiNumABUS) == 0);
  for (UINT32 sr = 0, a =0; sr < m_uiNumSubranksPerRank; sr += step, a++) {
    bool bTouched = false;
    for (UINT32 i = sr; i < (sr+step); i++) {
      if (subrank_vector[i]) {
        bTouched = true;
      }
    }
    if (bTouched) {
      ASSERT(m_bpSubRankAvailability[a]);
      m_bpSubRankAvailability[a] = false;
    }
  }

  return ((n-1)*repeat_interval + m_cProtocol.getCompleteTime(cmd) + DRAMTiming::ABUS_delay );  // return the projected completion time of this command
  // Note: it's a trick to model RDIMM
  // timing model operates without ABUS_delay, but it's added to latency returned to the memory controller
  // so, the system will observe additional latency due to ABUS_delay (if it's nonzero)
}

UINT32 cTimingModel::canIssueMoreCommand(void) const {
  UINT32 num_available_commands = 0;
  for (UINT32 a = 0; a < m_uiNumABUS; a++) {
    if (m_bpSubRankAvailability[a]) {
      num_available_commands++;
    }
  }
  ASSERT(num_available_commands <= m_uiNumABUS);
  return num_available_commands;
}

// functions that return whether a bank state is opne or close
bool cTimingModel::isBankOpen(const UINT32 rank, const UINT32 subrank, const UINT32 bank) const {
  return m_rbsBank[rank][bank][subrank].isBankOpen();
}
bool cTimingModel::isBankOpen(const UINT32 rank, const ChipMask sr_vector, const UINT32 bank) const {
  bool bOpen = true;
  for (UINT32 sr = 0; sr < m_uiNumSubranksPerRank; sr++) {
    if (sr_vector[sr]) {
      bOpen &= m_rbsBank[rank][bank][sr].isBankOpen();
    }
  }
  return bOpen;
}
bool cTimingModel::isBankOpen(const DRAMAddr daddr) const {
  UINT32 rk = daddr.rank;
  UINT32 bk = daddr.bank;
  const ChipMask subrank_vector = daddr.chip_cmd_count;//subrank_mask;

  return isBankOpen(rk, subrank_vector, bk);
}

bool cTimingModel::getRowStatusChipMask(DRAMAddr* daddr, ReqRowStatus status, bool split) const {
  bool one_matches = false;
  bool all_matches = true;

  UINT32 rank = daddr->rank;
  UINT32 bank = daddr->bank;
  UINT32 row = daddr->row;
  for (UINT32 c = 0; c < m_uiNumSubranksPerRank; c++) {
    if (daddr->chip_cmd_count[c]) {
      daddr->chip_cmd_count[c] = (getBankStatus(rank, c, bank, row) == status);
      one_matches = one_matches || daddr->chip_cmd_count[c];
      all_matches = all_matches && daddr->chip_cmd_count[c];
    }
  }

  return (split ? one_matches : all_matches);
}

// functions that return a bank status of a specific row: hit, miss, empty
ReqRowStatus cTimingModel::getBankStatus( UINT32 rank,
                                          UINT32 subrank,
                                          UINT32 bank,
                                          UINT32 row ) const {
  if (m_rbsBank[rank][bank][subrank].isBankOpen() == false) {
    return ROW_EMPTY;
  } else if (m_rbsBank[rank][bank][subrank].getRow() == row) {
    return ROW_HIT;
  } else {
    return ROW_MISS;
  }
}
// this one returns the status as hit,miss,empty when all subranks have identical status
// otherwise, banks status of subranks are different, then return miss
ReqRowStatus cTimingModel::getBankStatus(const DRAMAddr daddr) const {
  UINT32 rank = daddr.rank;
  UINT32 bank = daddr.bank;
  UINT32 row = daddr.row;
  ReqRowStatus row_state, row_state_new;
  row_state = row_state_new = INVALID_ROW_STATUS;
  bool bFirst = true;
  for (UINT32 sr = 0; sr < m_uiNumSubranksPerRank; sr++) {
    if (daddr.chip_cmd_count[sr]) {
      row_state_new = getBankStatus(rank, sr, bank, row);
      if (bFirst) {
        row_state = row_state_new;
        bFirst = false;
      } else {
        if (row_state != row_state_new) {
          return ROW_MIX;
        }
      }
    }
  }
  return row_state;
}

bool cTimingModel::isRankOpen(UINT32 rank, UINT32 subrank) {
  bool rOpen = false;

  for (UINT32 bk = 0; bk < m_uiNumBanksPerRank && !rOpen; bk++) {
    rOpen = rOpen || isBankOpen(rank, subrank, bk);
  }

  return rOpen;
}

}  // namespace drsim
/********************************************************************/
