#!/usr/bin/perl
# This script provides a set of perl subroutines to 
# 1. run a trace-driven Dr.Sim simulation
# 2. calculate metrics
# 3. grep statistics
# 
# Author: Doe Hyun Yoon and Min Kyu Jeong

# global options
$sim_cycle = 0;
$dram_freq = 800;
$sys_freq = 4000;
$scheduler = "FR_FCFS";
$closed_page = 1;
$rq_size = 64; # 32;
$wq_size = 64; # 32;
$wm_enter = 58; #28;
$wm_exit = 6;
$wm_allow_read_CAS = 1;
$channel = 1;
$subrank_map = "M1";
$rank_interleave = 1;
$rank = 1;
$subrank = 8;
$bank = 8;
$num_ABUS = 1;
$ABUS_delay = 0;
$subrank_xor = 0;
$channel_xor = 1;

$drsim = "../drsim";
$trace_dir = "../../trace/";
#$trace_dir = "../tools/tracegen/";
$templog = ".templog";
$num_transactions = 4096;
$element_size = 64;
$data_width = 8;
$BW_peak = $dram_freq*2*$data_width/1000;

sub print_header {
	($result_fn) = @_;
    open (RESULT, ">>", $result_fn) or die("Can't open result file: $result_fn");
    print RESULT "subrank_map, subrank, trace_filename: ";
    print RESULT "sys_clk,\t AU,\t DU,\t throughput,\t PH\n ";
    close RESULT;
}

sub run($){
	($trace_filename, $subrank, $rank, $rq_size, $wq_size_, $subrank_map, $num_ABUS, $result_filename) = @_;

	if( $wq_size_ != $wq_size ) {
		$wm_exit = int(0.3*$wq_size_+0.5);
		$wm_enter = int(0.8*$wq_size_+0.5);
    $wq_size = $wq_size_;
	}

	$command = "$drsim --sim_cycle $sim_cycle --dram_freq $dram_freq --sys_freq $sys_freq ";
	$command = $command . "--scheduler $scheduler --closed_page $closed_page --rq_size $rq_size ";
	$command = $command . "--wq_size $wq_size --wm_enter $wm_enter --wm_exit $wm_exit ";
	$command = $command . "--wm_allow_read_CAS $wm_allow_read_CAS --channel $channel ";
	$command = $command . "--subrank_map $subrank_map --rank_interleave $rank_interleave ";
	$command = $command . "--rank $rank --subrank $subrank --bank $bank --num_ABUS $num_ABUS ";
	$command = $command . "--ABUS_delay $ABUS_delay --sr_xor_hash $subrank_xor --ch_xor_hash $channel_xor ";
	$command = $command . "--age_promo_threshold 5000 ";
#	$command = $command . "$trace_dir/$trace_filename >> $templog";
	$command = $command . "$trace_dir/$trace_filename > $templog";

#  print $command . "\n";
#  return;
#  exit(0);
	system($command);

	my @results = process_log_file( $templog ); # sys_clk, AU, DU, PH, POWER
	my @throughput = calculate_throughput( $results[0] );

	my $debug = 0;
	if( $debug ) {
		print "sys_clk: $results[0], AU: $results[1], DU: $results[2], PH: $results[3]\n";
		print "throughput: $throughput[0] GB/s, $throughput[1]\% -- peak: $BW_peak GB/s \n";
	}

  open(RESULT, ">>", $result_filename) or die("Can't open result file: $result_filename");
  print RESULT "$subrank_map, $subrank, $trace_filename: ";
  print RESULT "$results[0]\t $results[1]\t $results[2]\t $throughput[0]\t $results[3]%\t\n";
  close RESULT;

  return $results[0];
#	print "$results[3]\t";
}

sub calculate_throughput($$$$)
{
	my ($sys_clk) = @_; 

	my $throughput = $sys_freq * $element_size * $num_transactions / $sys_clk / 1000;
	my $percent = 100*$throughput /$BW_peak;
		  
	return $throughput, $percent;
}

use Text::ParseWords;

sub process_log_file($) 
{
	my ($log_filename) = @_; 
	my $sys_clk = -1;
	my $AU = -1;
	my $DU = -1;
	my $PH = -1;
  my $POWER = -1;
	my $debug = 0;

	open(LOG, "<", $log_filename) or die("Can't open $log_filename");
	while( <LOG> ) { 
		if( m/^Simulation Finished/ ) { 
			my @words = &quotewords('\s+', 0, $_);
			$sys_clk = $words[6];
			if( $debug ) {
				print "sys_clk: $sys_clk\n";
			}
		}   
# Note: in a multi-channel configuration, AU/DU will have the last values
		if( m/^ABUS utilization:/ ) { 
			my @words = &quotewords('\s+', 0, $_);
			my $AU_ = $words[$#words-1];
			$AU_ =~ m/(\d+.?\d*)/;  
			$AU = $1;
			if( $debug ) {
				print "ABUS utilization: $AU\n"; 
			}
		}  
		if( m/^DBUS utilization:/ ) { 
			my @words = &quotewords('\s+', 0, $_);
			my $DU_ = $words[$#words-1];
			$DU_ =~ m/(\d+.?\d*)/;  
			$DU = $1;
			if( $debug ) {
				print "DBUS utilization: $DU\n"; 
			}
		}  
		if( m/^Page hit rate:/ ) { 
			my @words = &quotewords('\s+', 0, $_);
			my $PH_ = $words[$#words-1];
			$PH_ =~ m/(\d+\.?\d*)/;  
			$PH = $1;
			if( $debug ) {
				print "Page Hitrate: $PH\n"; 
			}
		}  
# NOTE(mjeong): power calc removed from the drsim
##		if( m/^allchannel_power/ ) { 
##			my @words = &quotewords('\s+', 0, $_);
##			my $POWER_ = $words[$#words-1];
##			$POWER_ =~ m/(\d+\.?\d*)/;  
##			$POWER = $1;
##			if( $debug ) {
##				print "allchannel_power: $POWER\n"; 
##			}
##		}  
	}
	close(LOG);
	
	return $sys_clk, $AU, $DU, $PH, $POWER;
}
