/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#include <iostream>
#include "base/output.hh"
#include "mem/drsim/drsim.hh"

using std::cout;
using std::endl;

/* 
 * A processor simulator should provide the LLC tag bits starting position.
 * It points to the physical address bits to XOR the bank index with, as suggested in 
 *
 * Z. Zhang, Z. Zhu, and X. Zhang. A permutation-based page interleaving scheme
 * to reduce row-buffer conﬂicts and exploit data locality. In Proc. the 33rd
 * Ann. IEEE/ACM Int’l Symp. Microarchitecture (MICRO), Dec. 2000
 *
 */
int llc_tag_shift = -1;

DrSim::DrSim(const Params* p)
    : PhysicalMemory(p),
      numReqSent(0),
      responseDelay(p->response_delay),
      use_ingress_q(p->use_ingress_q),
      stat_print_period(p->stat_print_period),
      tickEvent(this), statEvent(this), exitEvent(this, p->output_file)
{
    // using the old style param passing for now
    char buf[65536];
    strncpy(buf, p->drsim_config.c_str(), 65535);
    char* str1 = strtok(buf, ":");
    char* argv[256];
    int argc = 0;
    argv[argc++] = (char*)"drsim";
    while (str1) {
        argv[argc++] = str1;
        str1 = strtok(NULL, ":");
    }
    args.parse_input_args(argc, argv);

    // instantiate a decoder
    decoder = new drsim::AddrDecoder
        (args.num_channel(),  // uint32_t num_channel,
         args.module_config().num_rank, // uint32_t num_rank,
         args.module_config().num_chip, // uint32_t num_chip,
         args.module_config().num_bank, // uint32_t num_bank,
         args.module_config().num_row,    // uint32_t num_row,
         args.module_config().num_column, // uint32_t num_column,
         args.module_config().chip_width, // uint32_t chip_width,
         args.module_config().burst_length, // uint32_t burst_length,
         64,    // uint32_t channel_interleaving_granularity
         false, // bool bank_interleaving
         true,  // bool rank_interleaving
         8,     // uint32_t subrank_width,
         64,    // uint32_t subrank_interleaving_granularity
         true,  // bool bank_xor_hashing,        
         llc_tag_shift, // uint32_t bank_xor_hashing_offset,
         true,  // bool subrank_xor_hashing,
         false); // bool channel_xor_hashing)

    // instantiate memory channels
    mc = new drsim::MemoryChannel*[args.num_channel()];
    for (uint32_t i = 0; i < args.num_channel(); i++) {
        mc[i] = new drsim::MemoryChannel(args,
                                         args.module_config(),
                                         args.read_queue_size(),
                                         args.write_queue_size(),
                                         i);
    }

    // initialize clock -- DRAM clock and Sys clock all in MHz
    drsim::clock::xInitialize(args.dram_freq(), args.dram_freq());

    // print out the settings
    cout << "*** drsim args ***" << endl;
    args.print(cout);
    drsim::DRAMTiming::print_dram_timing(cout);
    cout << endl << "[Address Decoding Map]" << endl;
    decoder->PrintMask(cout);
}

DrSim::~DrSim()
{
  for (int i = 0; i < args.num_channel(); i++) {
    delete mc[i];
  }
  delete [] mc;
  delete decoder;
}

bool
DrSim::enqueueable(PacketPtr pkt)
{
    bool is_read = pkt->isRead();
    ADDRESS addr = pkt->getAddr();
    int channel = decoder->GetChannel(addr);

    if (use_ingress_q)
        return !(mc[channel]->isIngressQueueFull());
    else
        return ((is_read && !mc[channel]->isReadQueueFull()) ||
                (!is_read && !mc[channel]->isWriteQueueFull()));
}

void
DrSim::enqueue(PacketPtr pkt, int port_id)
{
    bool is_read = pkt->isRead();
    ADDRESS addr = pkt->getAddr();
    ADDRESS line_addr = (addr >> 6) << 6;
    int linesize = pkt->getSize();
    // FIXME(mjeong) : overriding the size with 64B
    linesize = 64;

    // code copied from DRAMsim wrapper
    if (pkt->isReadWrite()) {
        fatal("I don't know how to handle ReadWrites!");
    } else if (pkt->isRead() || pkt->isWrite()) {
        //readwriteAccesses++;
    } else {
        doAtomicAccess(pkt);
        ports[port_id]->schedSendTiming(pkt, curTick() + 1);
        //otherAccesses++;
        return;
    }

    // decode begin
    DRAMAddr dram_addr;
    uint32_t channel = 0;
    channel = decoder->GetChannel(line_addr);
    decoder->Decode(line_addr, linesize, &dram_addr);

    Gem5Req* req = new Gem5Req(this,
                               numReqSent++,
                               is_read,
                               line_addr,
                               linesize,
                               dram_addr,
                               0,
                               pkt,
                               port_id
                               );
    // do the functional operation now to ensure ordering
    doAtomicAccess(pkt);

    // send the request to the DrSim
    if (use_ingress_q)
        mc[channel]->sendRequestInQ(req);
    else
        mc[channel]->sendRequest(req);
}

void
DrSim::tick()
{
    do {
      drsim::clock::xTick();
      if (drsim::clock::is_dram_ticked()) {
        for (int i = 0; i < args.num_channel(); i++)
          mc[i]->cycle();
      }
    } while (!drsim::clock::is_sys_ticked());

    // TODO(mjeong): tick every cycle no matter if there's work to do
    // Power-down work also needs to be accounted for.
    schedule(tickEvent, curTick() + ticks_per_dram_cycle);
}

void
DrSim::printStat(ostream &os)
{
    os << "*** dsim stats ***" << endl;
    os << "********************************************************" << endl;
    os << "Simulation Finished at " <<
          "master_clk: " << drsim::clock::get_master_clk() <<
          ", sys_clk: "  << drsim::clock::get_sys_clk() <<
          ", dram_clk: " << drsim::clock::get_dram_clk() << endl;
    os << "--------------------------------------------------------" << endl;
    drsim::Stat::printall(os);
    os << "--------------------------------------------------------" << endl;
    drsim::MemoryChannel::printChannelStats(os);
    os << "--------------------------------------------------------" << endl;
}

void
DrSim::periodicStatPrint()
{
    static int period_i = 0;
    string fn = csprintf("drsim_stat_period_%d.txt", period_i++);
    ostream* os = simout.create(fn);
    drsim::Stat::printall_periodic(*os);
    simout.close(os);
    schedule(statEvent, curTick() + stat_print_period);
}

void
DrSim::printStatus()
{
    // NOTE(mjeong):
    // there is no standard way to convert FILE* to ofstream
    // we can write MemoryChannel::printStatus(FILE* fp), but seems
    // to be an overkill, so just print out to the stderr for now
    for (uint32_t i = 0; i < args.num_channel(); i++) {
        mc[i]->printStatus(cout);
    }
}

void
DrSim::startup()
{
    // FIXME(mjeong): assuming 1ps gem5 tick.
    ticks_per_dram_cycle = 1000000 / args.dram_freq();
    schedule(tickEvent, curTick() + ticks_per_dram_cycle);

    if (stat_print_period > 0)
        schedule(statEvent, curTick() + stat_print_period);
}

void
DrSim::sendResponse(PacketPtr pkt, int port_id)
{
    DPRINTF(DrSim, "DRAM access completed: %s 0x%x\n",
            pkt->cmdString(), pkt->getAddr());

    if (pkt->isResponse())
        ports[port_id]->schedSendTiming(pkt, curTick() + 1 + responseDelay);
    else
        delete pkt;
}

void
DrSim::wrapUpStats()
{
  for (int i = 0; i < args.num_channel(); i++) {
    mc[i]->wrapUpStats();
  }
}

void
DrSim::dumpStates()
{
  for (int i = 0; i < args.num_channel(); i++) {
    mc[i]->printStatus(std::cerr);
  }
}

void 
DrSim::init()
{
    if (ports.size() == 0)
        fatal("PhysicalMemory object %s is unconnected!", name());

    for (vector<DrSimPort*>::iterator pi = ports.begin();
         pi != ports.end(); ++pi) {
        if (*pi)
            (*pi)->sendRangeChange();
    }
}

Port*
DrSim::getPort(const std::string &if_name, int idx)
{
    if (if_name == "functional")
        return new DrSimPort(name() + "-functional", this, -1);

    if (if_name != "port")
        panic("Octopus::getPort: unknown port '%s' requested", if_name);

    if (idx >= ports.size())
        ports.resize(idx+1);

    if (ports[idx] != NULL)
        panic("DrSim::getPort: port %d already assigned", idx);

    DrSimPort *port =
        new DrSimPort(csprintf("%s-port%d", name(), idx), this, idx);

    ports[idx] = port;

    return port;
}

void
DrSim::regStats()
{
    using namespace Stats;

    /*
    totalAccesses
        .name(name() + ".totalAccesses")
        .desc("total accesses sent to DRAMsim")
        ;
    readwriteAccesses
        .name(name() + ".readnwriteAccesses")
        .desc("total number of accesses sent to DRAM module")
        ;
    otherAccesses
        .name(name() + ".otherAccesses")
        .desc("total number of accesses not sent to DRAM module")
        ;
    totalAccesses = 0;
    readwriteAccesses = 0;
    otherAccesses = 0;
    */
    registerExitCallback(&exitEvent);
    PhysicalMemory::regStats();
}

void
Gem5Req::completed_callback()
{
    // Trans finished
    drsim_->sendResponse(pkt_, port_id_);
}

void
Gem5Req::critical_word_callback(int cw_size)
{
    // No critical word first for Gem5 yet
//    if (pkt_->isResponse())
//        ports[port_id_]->schedSendTiming(pkt_, curTick() + 1);
}

void
Gem5Req::PrintExtra(ostream &os) const
{
    os << " m5tick:" << curTick();
}

DrSim* DrSimParams::create()
{
    return new DrSim(this);
}

DrSim::DrSimPort::DrSimPort(string name, DrSim* d, int i)
    : MemoryPort(name, d), id(i), drsim(d)
{
}

bool
DrSim::DrSimPort::recvTiming(PacketPtr pkt)
{
    if (pkt->memInhibitAsserted()) {
        // C&P from SimpleTimingPort::recvTiming()
        // snooper will supply based on copy of packet
        // still target's responsibility to delete packet
        delete pkt;
        return true;
    }

    if (drsim->enqueueable(pkt)) {
        drsim->enqueue(pkt, id);
        return true;
    }
    return false;
}

void
DrSim::ExitCallbackEvent::process()
{
    drsim->wrapUpStats();
    ostream* os = simout.create(output_fn);
    drsim->printStat(*os);
    simout.close(os);
}
